#ifndef controlledprocess_H
#define controlledprocess_H
//
// Copyright 2002,2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

///@file
/// This file deines class ControlledProcess which allows you to launch
/// a child process in such a way that you can write to its input stream
/// and read from its output stream.

#include <sys/types.h>
#include <string>
#include <stdio.h>
#include <cxxtls/thread.h>

namespace cxxtls
{

class ControlledProcess
    ///
    ///   A controlled process is a wrapper around a child process
    ///   that allows the owner to write to the child process's stdin
    ///   and read from it's stdout -- unlinke popen() that only lets
    ///   you do one or the other.
    ///
    ///   For example uses, see
    ///   @li tests/test_controlledprocess.cxx
    ///   @li bin/q.cxx
    ///
    ///   To create an object of type ControlledProcess, you must call
    ///   @ref ControlledProcess::start() -- which returns a pointer.
    ///
    ///   You can then presumably write to ptr->process_stdin(), and read from
    ///   ptr->process_stdout();
    ///
    ///   You can stop the process arbitrarily by calling ptr->stop().
    ///
    ///   The destructor calls stop(), automatically.
    ///
    ///   You can read characters from the process_stdout() stream by calling
    ///   the readStdoutByte() function.  It will return EOF on end of file or error
    ///   or the first available output byte if no error has occurred.
    ///
    ///
{
  ControlledProcess(ControlledProcess const &);              // not implemented
  ControlledProcess &operator=(ControlledProcess const &);   // not implemented

  ControlledProcess()
  :  process_(0)
  ,  command_invocation_()
  ,  status_(general_failure)
  ,  process_stdin_fh_(-1)
  ,  process_stdout_fh_(-1)
  ,  process_stdin_(0)
  ,  process_stdout_(0)
  ,  input_buffer_mode_(0)
  ,  output_buffer_mode_(0)
  {
  }

  pid_t        process_;
  std::string  command_invocation_;

  int          status_;  // 0 is good, all else is bad

  int          process_stdin_fh_;
  int          process_stdout_fh_;

  FILE*        process_stdin_;   // opened for WRITE!
  FILE*        process_stdout_;  // opened for READ!

  int          input_buffer_mode_;
  int          output_buffer_mode_;
  
  
  public:

  enum statusConstants
  {
    good                 = 0,
    general_failure      =-1,
    can_not_create_pipes =-2,
    fork_error           =-3,
    closed               =-4,

  };


  enum bufferingConstants
  {
    BLOCK_BUFFERED       = 0, // 8192 byte blocks or so
    LINE_BUFFERED        = 1, // until \n character is read
    CHARACTER_BUFFERED   = 2  // one character at a time
  };

  int status() const 
      ///  Get the status.  See ControlledProcess::constants for a list
      ///   of possible return values.  When the status() is 0, all is well.
      ///   Othewise, some error occurred.
  { 
      return status_; 
  }

  pid_t  process_id() const 
  { 
      if(status_ != good)
          return -1;
      return process_; 
  }

  FILE*  process_stdin()    { return process_stdin_;  }  // opened for non-buffered WRITE!
  FILE*  process_stdout()   { return process_stdout_; }  // opened for non-buffered READ! 

  static ControlledProcess* start(std::string commandLine, 
                                  int         inputBuffering =  LINE_BUFFERED,
                                  int         outputBuffering = LINE_BUFFERED
                                 );  
      ///< Create a controlled process.  Note that this function will
      ///  NOT return null on error.  You must check the status() method.
      ///
      ///  @param command_line -- the command to execute.  Remember to include
      ///                         stderr redirection, but you probably don't want
      ///                         to include stdin or stdout redirection as that
      ///                         would defeat the purpose of using this class.
      ///
      ///  @param inputBuffering -- the buffering mode of the stdin file.  See the
      ///                           following symbols:  @ref BLOCK_BUFFERED,
      ///                           @ref LINE_BUFFERED, @ref CHARACTER_BUFFERED.
      ///
      ///  @param outputBuffering -- similar to inputBuffering but specific to
      ///                            the stdout file handle.
      ///    
      ///  @return A pointer to a controlled process which you will need to
      ///          destroy when you are done with.  Note that this function will always
      ///          return a non-zero value.  To determine whether or the function was
      ///          successful, call the function @ref status(), and compare its result
      ///          to @ref good.  You must destroy the object even if the status is not
      ///          good.

  void stop()   
      /// closes the output and input files.  This should kill the process.
  {
      if(status_ == good)
      {
         fclose(process_stdin_);
         fclose(process_stdout_);  // should kill the process

         // do not null out these file handles because other threads
         // need to use them to find out that the file handles are
         // closed!

         status_ = closed;

      }
  }

  int readStdoutByte()
      // read a byte from the stdout file handle from the controlled
      // process in a thread safe manner.   At least it is thread safe
      // with respect to the "stop" function
  {
     if(status_ == good)
         return fgetc(process_stdout());

     return EOF;
  }


  ~ControlledProcess()
  {
     stop();
  }


}; // ControlledProcess

} // cxxtls

#endif   // controlledprocess_H
