#ifndef options_h_included
#define options_h_included

// this file is deprecated.  See commandLineOptions.h instead.


//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.




///@file
///
///   Defines a set of tools used to implement and check program command line options.
///

#include <string>
#include <vector>
#include <map>

namespace cxxtls
{

class ProgramOptions
    ///
    /// Class ProgramOptions represents the command line options and
    /// environment passed to a program by the operating system.  A
    /// ProgramOptions object can be viewed as a map of environment variables
    /// as well as a list of non-option command parameters, and finally a
    /// map of options and their parameters.
    ///
    /// While it is possible to have more than one ProgramOptions active at
    /// any one time, in general the use of this class is to represent _the_
    /// program options.  Thus there is a global variable holding this single
    /// program options declared by this header.  See ProgramOptions::global_options
    /// below.  If this pointer is null, then no options are set.
    ///
    /// For example, if a program is invoked in the following way:
    /// @code
    ///   program -h -I. -I .. -I ../.. -stuff file file2 file3 ...
    /// @endcode
    /// The ProgramOptions are described like this:
    ///
    ///   @li non-optional parameters:  file, file2, file3, etc
    ///
    ///   @li options without parameters:  -h, -stuff
    ///
    ///   @li options with parameters:   -I has ., .., ../..
    ///
    /// And of course the environment comes in 'behind' the command
    /// line, as it were.
    ///
    /// Once constructed, a ProgramOptions object provides the
    /// following members:
    ///
    ///   @li argv        a vector<string> containing the non-optional arguments
    ///   @li getenv(n)   a way of getting the program's environment variables
    ///   @li option(n)   a way of getting an option's value
    ///   @li setenv(n,v) a way of setting environment variables
    ///
    /// Note that there is no connection between a ProgramOptions object's
    /// copy of the program environment and the real program environment --
    /// except that one is copied from the other at the time of construction.
    ///
    /// Here is an example of constructing and using a ProgramOptions object:
    ///@code
    ///   int main(int argc, char **argv, char **environ)
    ///   {
    ///
    ///     // Parse the command line options and arguments and environment --
    ///     // AND store the arguments in the default GLOBAL place so that
    ///     // all parts of the program can have access to them.
    ///
    ///     ProgramOptions::global_options_ = new ProgramOptions("-x,-y,-z",
    ///                                                          argc,
    ///                                                          argv,
    ///                                                          environ);
    ///
    ///     ProgramOptions &op = *ProgramOptions::global_options_;  
    ///       // a local variable to make the accesses easier to understand
    ///
    ///     // Use the parsed options and arguments
    ///
    ///     if( op.option("-x") )  // operator int() tells you if option is set
    ///      {
    ///         string s = op.option("-x").value(); 
    ///          // operator string() converts value to string but some compilers
    ///          // this is buggy, so use .value()
    ///      }
    ///
    ///     // here is the easy to check for environment variables and to set them
    ///
    ///     if( op.getenv("PATH") )
    ///     {
    ///         string path = op.getenv("PATH");  // string conversion
    ///     }
    ///     else
    ///     {
    ///         op.setenv("PATH", "stuff");
    ///     }
    ///
    ///     // Here's how you access non-option arguments
    ///
    ///     for(int i=0; i < op.argv.size(); ++i)
    ///       cout << op.argv[i] << endl;
    ///
    ///     // Here's how you iterate over all environment items
    ///
    ///     ProgramOptions::environment const& env = op.get_environment();
    ///
    ///     ProgramOptions::environment::const_iterator f = env.begin();
    ///     ProgramOptions::environment::const_iterator l = env.end();
    ///
    ///     while(f != l)
    ///     {
    ///        cout << f->first << "=" << f->second << endl;
    ///        ++f;
    ///     }
    ///   }
    ///@endcode
    ///
    /// The 'value' of an option is an array of string objects.  Even if there
    /// is no value, or if there is only 1 allowed value, an option is stored
    /// as an array of strings.  Thus you can iterate over the values in an
    /// option using the same algorithm -- regardless of the kind of option it
    /// is.  The option member, operator string(), converts all the values into
    /// a comma separated string and returns that.  You can iterate over the
    /// individual values of an option like this:
    ///
    ///@code
    ///   ProgramOptions::value_iterator first = op.option("-x").begin();
    ///   ProgramOptions::value_iterator last  = op.option("-x").end();
    ///
    ///   while(first != last)
    ///   {
    ///     string v = *first++;
    ///   }
    ///@endcode
    ///
    /// Note, if the double lookup of op.option("-x") bothers you, do this
    ///
    ///@code
    ///   ProgramOptions::Option const& option = op.option("-x");
    ///   ProgramOptions::value_iterator first = option.begin();
    ///   ProgramOptions::value_iterator last  = option.end();
    ///   ...
    ///@endcode
    ///
    /// Note: if an option does not take values, then it is viewed as a
    /// boolean flag.  If the user supplies it multiple times on the
    /// command line, the value will be toggled with each use.
    ///
    /// When one character options take values, these values can be supplied
    /// as part of the option -- as occurs with the -I option to compilers,
    /// but multiple character option names must be separated from their
    /// values:
    ///
    ///@code
    ///   -I../../junk.h    // works
    ///   -stuff../file.h   // wrong will not be recognized.
    ///@endcode
    ///
    /// When constructing an options object, you may optionally
    /// choose to instruct it not to print an error to stdout and exit if
    /// any options are not recognized -- as is the default behaviour.  If
    /// you choose this mode of operation, use ProgramOptions::error() to
    /// determine if an error in processing the options has occurred.  It
    /// should return false to indicate that no error occurred.  If it returns
    /// true, use ProgramOptions::error_text() to find out what the error is.
    ///
    /// When constructing ProgramsOptions and supplying a description of the
    /// options as a string, use the following constructor syntax:
    ///
    ///@code
    ///   ProgramOptions op(option_declaration_string,
    ///                     argc, argv, environment,
    ///                     fail_on_error_flag);
    ///@endcode
    ///
    /// Note that the fail_on_error_flag is optional and defaults to true;
    ///
    /// When definiting an options object and passing it a string describing
    /// the options, rather than array of descriptors, the string must conform
    /// to the following:
    ///
    ///@code
    ///
    ///   The string is a collection of parameter names followed by
    ///   terminator that indicate the number parameters for the option.
    ///   A ":" means 1 parameter, a ";" means multiple parameters, and a ","
    ///   means no parameters.  For example:
    ///
    ///     "-a,-queue,-bee:-c;-d"
    ///
    ///   Means:
    ///
    ///      Accept an option 'a' that takes no arguments
    ///      accept an option 'queue' that takes no arguments
    ///      accept an option, 'bee', that takes 1 argument
    ///      accept an option, 'c', that takes multiple arguments
    ///        (meaning that you can use it multiple times an suply a
    ///         new option on each use)
    ///      accept an option 'd' that takes no arguments.
    ///@endcode
    ///
    /// Here's how to process options which allow multiple values to be set:
    ///
    ///@code
    ///
    ///  if(options.option("-I"))
    ///  {
    ///    ProgramOptions::value_iterator f, l;
    ///
    ///    f = options.option("-I").begin();
    ///    l = options.option("-I").end();
    ///
    ///    while(f !=l)
    ///    {
    ///      useString(*f);
    ///      ++f;
    ///    }
    ///
    ///
    ///  }
    ///@endcode
        
{

public:

 // types

    class                                  Descriptor;
    class                                  Descriptors;
    class                                  Option;
    class                                  EnvVar;

    typedef std::string                    string;

    typedef std::vector<string>            option_values;  ///< storage type for option values
    typedef option_values::const_iterator  value_iterator; ///< iterator for option values
    typedef std::map<string,string>        Environment;    ///< fast access to program env.

    typedef std::map<string,   Option*>    Options;
        //
        ///  list of options and their values
        ///  if any.  The map key points
        ///  to the string name_ field in the
        ///  Option object.


    static ProgramOptions*                 global_options_; ///< _the_ program's options

    enum allowed_parms_
        ///  Symbolic names for the parms parameters for add_options and
        ///  Descriptors::Descriptors()
    {
      no_values,   ///<  This is a boolean option.
      one_value,   ///< This option takes exactly one value.
      many_values  ///< This option can be used multiple time and each
                   ///< use requires a value.
    };

 // Public Variables

    option_values   argv;           ///< non-option args are public
                                    ///  arg0 is program name as usual

 // Methods

    ProgramOptions(std::string          const &option_descriptor_string,
                   int                  argc,
                   void const *         argvp,
                   void const *         environp,
                   bool                 exit_on_error=true
                  );
       ///<
       ///  Construct program options and an environment from those passed
       ///  in by the operating system.  See the class comments for details
       ///  of using this constructor.  This constructor
       //

    ProgramOptions(Descriptors const &  desc,
                   int                  argc,
                   void const *         argvp,
                   void const *         environp,
                   bool                 exit_on_error=true
                  )
    : exit_on_error_(exit_on_error),
      error_(false)
    {
      constructor(desc, argc, argvp, environp, exit_on_error);
       //
       ///  Construct program options and an environment from those passed
       ///  in by the operating system.  See the class comments for details
       ///  of using this constructor.  This constructor requires that
       ///  you supply a vector of descriptors.  The next constructor
       ///  accepts a character string describing the options.
       //
    }

    ProgramOptions()
    : exit_on_error_(true),
      error_(false)
    {
       //
       /// Default constructor in case you need to make the options from scratch.
       //
    }

    private:
        ProgramOptions(ProgramOptions const &); // not implemented
    public:

    ~ProgramOptions();       ///< destroys the options map and everything else
                                
    Option &add_option(string  const &name);
       ///<
       /// Create a new option with no values in it.  That is, it pretends
       /// that an option was set in the argv passed in at construction
       /// time.  This kind of option does not need to appear in the
       /// option descriptors used at construction time.
       //
    void remove_option(string const &name);

       /// Remove an option and its values
                                
    Option const &option(string const &name) const;
       ///<
       /// Find the option or return a reference a static option that
       /// will tell you that it doesn't exist.
       //

    bool error() const { return error_; }
        ///<
        ///  Indicates that an error occurred while processing the options.
        ///  Unless you construct the options with the exit_on_error parameter
        ///  set to false, this function will never return anything but false.
        
    string const& error_text() const { return errmsg_; }
        ///<
        ///  Returns the text of the error message, if any, corresponding to
        ///  the error that occured when parsing the options.  This will never
        ///  return a non-zero string unless you construct the options with
        ///  exit_on_error=false.
        
    EnvVar  getenv(string const &name) const;
        ///<
        ///  Get an environment variable given its name.  EnvVar is a wrapper
        ///  class that lets you tell whether or not the variable is set or
        ///  not as well as get its values.  See ProgramOptions::EnvVar for
        ///  details of using the return value.
        
    void setenv(string const &name, string const & value);
        ///<
        ///  Set a new value in the evironment or change an old one.  Note
        ///  that this does not update the real program environment.  That
        ///  ::getenv() won't get you the value you just set.

    Environment const& get_environment() const { return environ_; }
        ///<
        ///  Get the actual environment so the user can iterate over the
        ///  variable, not just check them out, one by one.  Rarely done,
        ///  but useful in testing.
        
    Options const& get_options() const { return options_; }
        ///<
        ///  Get a list of the options that are set so that you can
        ///  iterate over it in some way -- mainly for testing.
        
    std::string expand(std::string const &) const;
        ///<
        ///   Return a copy of the input string with environment variable
        ///   references expanded.  An environment variable reference
        ///   is a substring of the form '$name' where the name can take
        ///   either of the following forms:
        ///   @code
        ///
        ///      $[0-9]                    -- command line parameter expansion
        ///      $[a-zA-Z_][0-9a-zA-Z_]*   -- environment variable expansion
        ///   @endcode
        /// 
        ///   Here is an example expansion:
        ///   @code
        ///     static char *args[]= { "zero", "one", "two", 0 };
        /// 
        ///     options.setenv("fred", "freddy boy");
        /// 
        ///     std::string rv = options.expand("$1-$fred-");
        ///   @endcode
        ///   After expansion, rv will contain "one-freddy boy-".
        /// 
        ///   Obviously, only the first 9 program parameters can be included
        ///   in expansions of this form.  This restriction is kept in order to
        ///   be consistent with the behavior of most shell interpreters.
        /// 
        
    class Option
       //
       /// Interface to a given option.  See documentation for it in the
       /// ProgramOptions class.
    {

    public:

        Option(string const &name);             ///< normal constructor
        
        void append_value(string const &value); ///< add a new value to this option
        
        value_iterator begin() const { return values_.begin(); }
          ///<  first value in option's list
        
        value_iterator end()   const { return values_.end(); }
          ///< last value therein
        
        operator bool() const { return set_; }
          ///< returns true if the option is set

        string value() const;
          ///< convert values to a string (comma separated if needed)
        
        operator string() const { return value(); } 
          ///<  convert values to a string (comma separated if needed)
        
        string const& name() const { return name_; }  ///<  get the option's name
        
    private:

        friend class ProgramOptions;
        
        Option(); ///< default constructor creates an option that is empty of
                  ///  values and claims not to be set.

        option_values values_;  ///< values associated with this option if any
        
        string        name_;    ///< ProgramOptions::members point directly
                                ///  at this object
                        
        bool          set_;     ///< flag indicating that the option is actually set

        
    };


    class Descriptor
        //
        ///  The description of a single option's type -- not its values,
        ///  for that see ProgramOptions::Option.
        //
    {
      public:

        string  name_;   ///< of the option
        int     parms_;  ///< allowed parameter count, see ProgramOptions::allowed_parms_


        Descriptor(string const &name, int parms=no_values)
            ///  Construct an Option Descriptor given a name
            ///  for the option and specification of how many, if any,
            ///  values are accepted by this option.  Use one of the
            ///  ProgramOptions::allowed_parms_ names.
            //
        : name_(name),
          parms_(parms)
        {
        }

    };


    class EnvVar
        //
        ///  An EnvVar object exists to provide information about environment
        ///  variables stored in the map of environment variables without having
        ///  the user know about the implementation.  Here is how an EnvVar
        ///  is used:
        ///  @code
        ///    if( op.getenv("PATH") )
        ///    { // environment variable is set
        ///  @endcode
        ///  Or you can do this:
        ///  @code
        ///    string s = op.getenv("PATH");        // gets the value of PATH
        ///    string n = op.getenv("PATH").name()  // gets 'PATH'
        ///  @endcode
        ///  If you want to print one of these things, convert it to a std::string
    {

        bool          set_;    ///< is this variable actually defined?
        string        name_;   ///< the name of the variable you are talking about
        string        value_;  ///< the value of it (or empty string if not set)
        
      public:

        EnvVar(string const &name)
        : set_(false),
          name_(name)
        {
           ///  Construct from only a name.  This leaves the value unset.
        }
        
        EnvVar(string const& name, string const &value)
        :  set_(true),
           name_(name),
           value_(value)
        {
           ///  Construct with both name and value.
        };

        EnvVar &operator=(std::string const &rhs)
           ///  Replace the value of the variable and mark as being defined.
        {
           set_ = true;
           value_ = rhs;

           return *this;
        }
        
        operator bool() const { return set_; }
           ///<
           ///  convert this object to a bool so you can find out if
           ///  the variable is set or not.
        
        string name() const { return name_; }
            ///<  get the name of the variable this object represents.  The
            ///   function returns a constructed string to account for the
            ///   most common use of an EnvVar -- as a function return value.

        string value() const { return value_; }
           ///<  get the value of the variable.  This function returns a 
           ///   constructed string rather than a reference to account for
           ///   the most common use of an EnvVar -- as a function return value.


        // warning, do not return a pointer to the value's c_str().  Due to common
        // usage mistakes the address of destructed string will be given out.
        //
        // Also, do not add an operator string() method.  This confuses the string
        // template on g++.  Some kind of odd bug in the compiler, I'm guessing.

    };


    class Descriptors

        ///  A collection of Descriptor's of options.  This is
        ///  a thin wrapper around a vector<Descriptor>  Its main
        ///  purpose is to define the add_option method so that
        ///  the interface to array of descriptors becomes simpler
        ///  to use.
    {
    public:

        std::vector<Descriptor> options_; ///< the list of allowed options
        
        void add_option(string const & name, int parms = no_values);

    };

private:

    static Option  null_option_;  ///< static option which indicates that it isn't set.

    Options        options_;      ///< list of options and their values
                                  ///  if any.  The map key points
                                  ///  to the string name_ field in the
                                  ///  Option object.

    bool   exit_on_error_;        ///< flag saying user wants control of error handling
    string errmsg_;               ///< return by error_text (text parsing options error)
    bool   error_;                ///< error occurred while parsing options

    Environment environ_;         ///< map of environment variables

    Option* get_option(string const&); ///< get a writeable pointer to the named option.

    void constructor(Descriptors const &  desc,
                     int                  argc,
                     void const           *argv_ptr,    // any ptr can be converted to this
                     void const           *environ_ptr, // ditto
                     bool                 exit_on_error=true
                    );
        ///<
        ///  constructor() is the 'real' constructor.  It is a separate
        ///  function so that it can be called as a function by
        ///  the constructor that takes a character string defining the
        ///  descriptors.


};

} // namespace cxxtls

#endif

