#ifndef _threadSafeHandle_header_included_p
#define _threadSafeHandle_header_included_p
//
// Copyright 2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

#include <cxxtls/thread.h>
#include <new>

///@file
///
///  This file defines ThreadSafeHandle.
///

namespace cxxtls
{

template<class T>
struct ThreadHandleImpl
    /// the "body" of a thread handle.  The body has the mutex guard and
    /// knows how to get to the counter and knows how to get to the thing
    /// being refered to by a handle.
    ///
    /// the ThreadHandleImpl exists so as to ensure that nothing happens
    /// to the thread handle unless it is truly guarded by a mutex.  Note
    /// that mutex locking is not necessary for dereferencing the mutex,
    /// just copying and deleting.
    ///
    /// You can construct these objects but you cannot destroy them directly.
    /// when you call the dec() method, if the count goes to zero, the
    /// dec method destroys it for you.
{
    T* value_;
    int count_;

    Mutex guard_;

    ThreadHandleImpl(T* v)
    :  value_(v)
    ,  count_(1)
    {
    }

    void inc()
    {
       ScopedMutexLock lock(guard_);

       ++count_;
    }

    void dec()
    {
       bool deleted = false;

       {
            ScopedMutexLock lock(guard_);
           
            --count_;
           
            if(count_ < 1)
            {
               deleted = true;
            }
           
       }

       if(deleted)
          delete this;
    }

    int count() 
    {
       ScopedMutexLock lock(guard_);
       return count_;
    }

private:

     ~ThreadHandleImpl()
     {
         delete value_;
     }

     ThreadHandleImpl &operator=(ThreadHandleImpl const &); // not implemented

     ThreadHandleImpl(ThreadHandleImpl const &rhs);  // not implemented
};




template<class T>
class ThreadSafeHandle
    /// A ThreadSafeHandle is a wrapper around a thread safe counter and a potentially
    /// non-thread safe pointer to a new-ed up object which will be deleted by the 
    /// destructor of the handle, if their are no other references to the object.
    ///
    /// It is an error to put the same pointer in multiple ThreadSafeHandles
    ///
    /// @warning, And I mean, seriously, WARNING:  do not add new member functions
    /// to this class or you will create a maintenance nightmare.  Do not  add functions
    /// like:
    ///
    ///   get() -- instead, use (&*handle)
    ///   reset() -- this is not an auto-ptr.  If you make it act like auto-pointer, it will be
    ///              serious bitch to convert code that uses autho-pointer into code that uses
    ///              handles.
    ///   release() -- this function makes no sense fo rthis class.  The destructor releases it
    ///                for you.  You will just confuse maintainers if you add this member.
    ///
{
    mutable ThreadHandleImpl<T>  *body_;


    T* get() 
    {
       if(body_)
       {
          return body_->value_;
       }
       else
          return 0;
    }


    T const * get() const 
    {
       if(body_)
       {
          return body_->value_;
       }
       else
          return 0;
    }

public:

    ThreadSafeHandle()
    :  body_(0)
    {
    }

    explicit ThreadSafeHandle(T* newbie)
        ///  it is an error to put the same pointer in multiple ThreadSafeHandles
    {
       if(newbie)
       {
           body_ = new ThreadHandleImpl<T>(newbie);
       }
       else
       {
           body_ = 0;
       }
    }

    ThreadSafeHandle(ThreadSafeHandle<T> const &rhs)
       /// Copy construct and point to the same object as rhs.  Note that if rhs is really const, in your application
       /// this fact is lost in the construction process.  You must assert constness for the copied object yourself.
       ///
       /// Sorry, this is hard enough to get to work as it is.
    {
        body_ = rhs.body_;

        if (body_)
            body_->inc();
    }

    ThreadSafeHandle &operator=(ThreadSafeHandle<T> const &rhs)
       /// note that if your rhs of the assignment is const, this object, which is non-const, will still point to it.
       /// you must write your code not to violate const given the situation.
       ///
       /// Sorry, this is hard enough to get to work as it is.
    {
       if(this == &rhs)
          return *this;

       if(this->body_ == rhs.body_)
          return *this;


        if(body_)
           body_->dec();

        body_ = rhs.body_;

        if (body_)
            body_->inc();

        return *this;
    }


    ~ThreadSafeHandle()
    {
       if(body_)
          body_->dec();
    }



    T* operator->() 
    {
       return get();
    }

    T & operator*()
    {
       return *get();
    }




    T const * operator->() const
    {
       return get();
    }

    T const& operator*() const
    {
       return *get();
    }

    operator bool() const 
    {
       return 0 != get();
    }


    int referenceCount() const  
    {
        ThreadHandleImpl<T>* body = const_cast<ThreadHandleImpl<T>*>(body_);

        if(body)
           return body->count();
        else
           return 0;
    }


}; // class ThreadSafeHandle

} // namespace cxxtls

#endif
