//
// Copyright 2002-2016, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include "BrashCommands.h"
#include "TestExpression.h"

#include <cxxtls/stringexpressionparser.h>
#include <cxxtls/file.h>
#include <cxxtls/ArithmeticExpressionParser.h>
#include <cxxtls/strtool.h>
#include <cxxtls/getopt.h>
#include <cxxtls/lastErrno.h>
#include <cxxtls/stlstream.h>
#include <cxxtls/commandLineOptions.h>
#include <cxxtls/dateTools.h>
#include <cxxtls/cutset.h>
#include <cxxtls/longestCommonSubsequence.h> // for .diff
#include <cxxtls/GaussianLpSolver.h>  // for .solve
#include <cxxtls/timeseries.h>        // for .linreg


#include <fstream>
#include <sstream>

using namespace cxxtls;
using namespace std;

///@file
///
///  This file implements the builtin functions for the brash language.  


// Static initialization is
//  used to populate the list of builtin functions.  This doesn't happen just because you add a new
//  builtin function, you must add a line to the BuiltinsInitializer::BuiltinsInitializer() 
//  function to get your builtin added to the language.


// file name pattern searches should be fully qualified directory names, though we will not always report them as such,
// because this is a multithreaded application and another thread can change the name of the current directory
// at any time.


extern BrashScriptCommandHandler *brashGlobalScriptCommandHandler;  // owner should actually be passed in to the execute methods.  fix later
                                                                    // note that brashGlobalScriptCommandHandler is derived from JobController,
                                                                    // which is passed to all functions, so theoretically, we should be able to
                                                                    // cast the job controller back into a BrashScriptCommandHandler, which is
                                                                    // the owner.  I haven't tried this yet but it should work.

//==================================================================================================
//==================================================================================================
//==================================================================================================

static string emptyString;
static string nl("\r\n");
static string tab("\t");
static string space(" ");
static string equalSign("=");
static string is(" is ");
static string singleQuote("'");
static string alias("alias");
static string builtin("builtin");
static string unknown("unknown");
static string semicolon(" ; ");      // keep the spaces here to reduce the number of string concatenations
static string functionText("function");
static string dollar("$");
static string slash("/");   // works on windows too most of the time
static string colon(":");
static string comma(",");
static string BRASH_RESULT_VAR_PREFIX("BRASH_PIPE_RESULTS_");
static string funcParmsString("()");
static string dueTo(" due to ");
static string TEsyntaxError(":\r\n  [test]-statement: error:  Syntax Error Parsing Expression:\r\n ");
static string apostrophe("'");
static string doubleQuote("\"");
static string error("error:  ");
static string dashH("-h");
static string dashdashH("--help");

bool isBrashBuiltin(std::string const &name);

static string findDirInPath(string const &dirname, string const &dirpath, string const &startCurdir);  
    ///<  find a directory by searching the specified @b dirpath.
    ///
    ///   @param dirname The name of a directory which you wish to find.  Given a full pathname,
    ///                  the @b dirpath is ignored, but if not, it is searched.
    ///
    ///   @param dirpath A list of directories which are separated by the BASH_PATH_SEPARATOR variable,
    ///                  which will either be ; or : depending on the OS.
    ///
    ///   @param startCurdir The presumed current directory from which to start searches if the 
    ///                      dirpath elements are not absolute paths.
    ///
    ///   @return        Either the dirname modified by the appropriate dirpath
    ///                  member, or just just the dirname if none is found.


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .attrib command
//


struct    BrashBuiltinAttribExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         // split the command line options into 3 types:
         //  file names
         //  valid options
         //  invalid options

         vector<string> files;
         vector<string> options;

         vector<string> invalidArgs;

         for(size_t i = 1; i < arguments.argc; ++i)
         {
            string arg(arguments.args_[i]);

            if(arg == "-h" || arg == "--help")
            {
                workingContext.fileHandles()->writeBytes(2, string(
                                                                    "\r\n"
                                                                    ".attrib:  list/change Windows file attribute\r\n"
                                                                    "\r\n"
                                                                    "Usage:\r\n"
                                                                    "    .attrib [ options ]  files;\r\n"
                                                                    "Options:\r\n"
                                                                    "    +H     set the hidden flag\r\n"
                                                                    "    -H     remove the hidden flag\r\n"
                                                                    "    +R     set the read-only flag\r\n"
                                                                    "    -R     remove the read-only flag\r\n"
                                                                    "    +S     set the system flag\r\n"
                                                                    "    -S     remove the system flag\r\n"
                                                                    "    +I     set the non-content-indexed flag\r\n"
                                                                    "    -I     remove the non-conten-indexed flag\r\n"
                                                                    "    +A     set the archive flag\r\n"
                                                                    "    -A     remove the archive flag\r\n"
                                                                    "Description\r\n"
                                                                    "  With no - or + options, list the attributes of the named files.\r\n"
                                                                    "\r\n"
                                                                    "  The - operator means to remove the attribute from the files\r\n"
                                                                    "\r\n"
                                                                    "  The + operator means to add the attribute to the files.\r\n"
                                                                    "\r\n"
                                                                    "  Changable Attributes:\r\n"
                                                                    "    H -- hidden\r\n"
                                                                    "    R -- read only\r\n"
                                                                    "    S -- system\r\n"
                                                                    "    I -- not indexed\r\n"
                                                                    "    A -- archived\r\n"
                                                                    "  Fixed Attributes:\r\n"
                                                                    "    d -- directory\r\n"
                                                                    "    L -- reparse point (or symbolic link)\r\n"
                                                                    "    C -- compressed\r\n"
                                                                    "\r\n"
                                                                  )
                                                        );

                workingContext.executionContext()->variables().setErrorCode(1);
                return ScriptFunctions::SRC_Success;

                
            }

            if(arg.empty())
               invalidArgs.push_back(arg);
            else
            {
               if(arg[0] == '+' || arg[0] == '-')
               {
                  if(arg.size() != 2)
                     invalidArgs.push_back(arg);
                  else
                  switch(arg[1])
                  {
                     case 'I':
                     case 'A':
                     case 'R':
                     case 'H':
                     case 'S':
                         options.push_back(arg);
                         break;

                     default:
                         invalidArgs .push_back(arg);

                  }
               }
               else
               {
                   // treat the rest of the arguments as file names

                   while(i < arguments.argc)
                   {
                       if(!arguments.args_[i].empty())
                           files.push_back(arguments.args_[i]);

                       ++i;
                   }
               }
            }

         }

         if(!invalidArgs.empty())
         {
            workingContext.fileHandles()->writeBytes(2, commandLine->sourceFile_ + colon + lexical_cast<string>(commandLine->sourceLine_) + 
                                                        string(":  Error, invalid arguments to attrib:  ")
                                                    );

            for(size_t i = 0; i < invalidArgs.size(); ++i)
                workingContext.fileHandles()->writeBytes(2, invalidArgs[i] + space);

            workingContext.fileHandles()->writeBytes(2, nl);

               

            workingContext.executionContext()->variables().setErrorCode(1);
            return ScriptFunctions::SRC_Success;
         }

         // if we get here, the +/- args are all good

         if(options.empty())
         {
             // print existing options

             for(size_t i = 0; i < files.size(); ++i)
             {
                 
                 DWORD atts = GetFileAttributes(files[i].c_str());

                 if(atts == INVALID_FILE_ATTRIBUTES)
                 {
                    workingContext.fileHandles()->writeBytes(2, string("Error:  cannot read file attributes for:  ") + files[i] + nl);
                    continue;
                 }

                 string txt;

                 if(atts & FILE_ATTRIBUTE_ARCHIVE)
                    txt += 'A';
                 else
                    txt += ' ';

                 if(atts & FILE_ATTRIBUTE_COMPRESSED)
                    txt += 'C';
                 else
                    txt += ' ';

                 if(atts & FILE_ATTRIBUTE_DIRECTORY)
                    txt += 'd';
                 else
                    txt += ' ';

                 if(atts & FILE_ATTRIBUTE_HIDDEN)
                    txt += 'H';
                 else
                    txt += ' ';

                 if(atts & FILE_ATTRIBUTE_READONLY)
                    txt += 'R';
                 else
                    txt += ' ';

                 if(atts & FILE_ATTRIBUTE_REPARSE_POINT)
                    txt += 'L';
                 else
                    txt += ' ';

                 if(atts & FILE_ATTRIBUTE_SYSTEM)
                    txt += 'S';
                 else
                    txt += ' ';

                 if(atts & FILE_ATTRIBUTE_NOT_CONTENT_INDEXED)
                    txt += 'I';
                 else
                    txt += ' ';


                 txt += ":  ";

                 workingContext.fileHandles()->writeBytes(1, txt);

                 workingContext.fileHandles()->writeBytes(1, files[i] + nl);


             }
         }
         else
         {
             // modify options

             for(size_t i = 0; i < files.size(); ++i)
             {
                 
                 DWORD atts = GetFileAttributes(files[i].c_str());

                 if(atts == INVALID_FILE_ATTRIBUTES)
                 {
                    workingContext.fileHandles()->writeBytes(2, string("Error:  cannot read file attributes for:  ") + files[i] + nl);
                    continue;
                 }

                 for(size_t j = 0; j < options.size(); ++j)
                 {
                     int bit = options[j][1];   // ARHSI
                     int mode= options[j][0];   // +, or -

                     switch(bit)
                     {
                         case 'h':
                         case 'H':
                             {
                                 if(mode == '+')
                                     atts |= FILE_ATTRIBUTE_HIDDEN;
                                 else
                                     atts &= ~(FILE_ATTRIBUTE_HIDDEN);
                             }
                             break;

                         case 'r':
                         case 'R':
                             {
                                 if(mode == '+')
                                     atts |= FILE_ATTRIBUTE_READONLY;
                                 else
                                     atts &= ~(FILE_ATTRIBUTE_READONLY);
                             }
                             break;

                         case 's':
                         case 'S':
                             {
                                 if(mode == '+')
                                     atts |= FILE_ATTRIBUTE_SYSTEM;
                                 else
                                     atts &= ~(FILE_ATTRIBUTE_SYSTEM);
                             }
                             break;

                         case 'a':
                         case 'A':
                             {
                                 if(mode == '+')
                                     atts |= FILE_ATTRIBUTE_ARCHIVE;
                                 else
                                     atts &= ~(FILE_ATTRIBUTE_ARCHIVE);
                             }
                             break;

                         case 'i':
                         case 'I':
                             {
                                 if(mode == '+')
                                     atts |= FILE_ATTRIBUTE_NOT_CONTENT_INDEXED;
                                 else
                                     atts &= ~(FILE_ATTRIBUTE_NOT_CONTENT_INDEXED);
                             }
                             break;

                     }

                 } // modify option flags

                 DWORD ok = SetFileAttributes(files[i].c_str(), atts);

                 if(!ok)
                 {
                    workingContext.fileHandles()->writeBytes(2, string("Error:  set read file attributes for:  ") + files[i] + nl);
                 }

             } // for each file

         }  // modifying file options



         return ScriptFunctions::SRC_Success;
     }

     virtual string abstract() const
     {
         return "Set/List file attributes:  .attrib [-h | --help]  [-/+{A,H,R,S}] [file] ...";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// NOOP command
//

struct    BrashBuiltinNOOPExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
            // this command really does 1 thing -- it sets the $_ variable to indicate the final parameter.
            //
            // this is useful because it lets find the last parameter of the current script or function
            //
            //     function f
            //     {
            //         : "$@"
            //
            //         echo The last parameter was $_
            //     }

            CommandLineArguments arguments(commandLine, executionContext_in, skipArg0);

            return ScriptFunctions::SRC_Success;
     }

     virtual string abstract() const
     {
         return "Do Nothing, '<colon> commmand':  : ...;  Note:  sets $_ to be the last parameter value.";
     }
};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// CD command
//

struct    BrashBuiltinCdExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         string oldPWD = workingContext.executionContext()->variables().find("PWD");
         

         if(arguments.argc < 2)
         {
            // cd with no arguments means go home

            string home = workingContext.executionContext()->variables().find("HOME");

            if(home.empty())
            {
#ifdef _MSC_VER
                string userprofile= workingContext.executionContext()->variables().find("USERPROFILE");

                if(userprofile.empty())
                   home="\\";
                else
                   home=userprofile;
#else
                home="/tmp";
#endif
            }

            int err = !workingContext.fileHandles()->setCurrentDirectory(home.c_str());

            if(err)
            {
                FileName tmp(commandLine->sourceFile_);
                tmp.convert_to_absolute_path();

                workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) + 
                                                            string(":  Error, cd could not change to directory, '") + home + apostrophe + nl);

                bool   throwExceptionOnError = workingContext.executionContext()->variables().errorMode();   // by default, cd failure is a major error

                if(throwExceptionOnError)
                {
                    throw BadCdTargetException("directory", home);
                }

                return ScriptFunctions::SRC_Success;
            }

            workingContext.executionContext()->variables().assign("PWD", workingContext.fileHandles()->currentDirectory());
            workingContext.executionContext()->variables().assign("OLDPWD", oldPWD);

            workingContext.executionContext()->variables().setErrorCode(0);
            return ScriptFunctions::SRC_Success;


         }

         if(arguments.argc > 2)
         {
                FileName tmp(commandLine->sourceFile_);
                tmp.convert_to_absolute_path();

                workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +  
                                                           string(":  Error:  cd has too many parameters!\r\n"));

                bool   throwExceptionOnError = workingContext.executionContext()->variables().errorMode();   // by default, cd failure is a major error

                if(throwExceptionOnError)
                {
                    workingContext.executionContext()->variables().setErrorCode(1);
                    throw BadCdTargetException("error type", "too many parameters");
                }
                workingContext.executionContext()->variables().setErrorCode(1);
                return ScriptFunctions::SRC_Success;
         }

         string target = findDirInPath(arguments.args_[1],   workingContext.executionContext()->variables().find("CDPATH"), oldPWD);

         int err = !workingContext.fileHandles()->setCurrentDirectory(target.c_str());
         
         if(err)
         {
             FileName tmp(commandLine->sourceFile_);
             tmp.convert_to_absolute_path();

             workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) + 
                                                         string(":  Error, cd could not change to directory, '") + target + apostrophe + nl);
         
             bool   throwExceptionOnError = workingContext.executionContext()->variables().errorMode();   // by default, cd failure is a major error
         
             if(throwExceptionOnError)
             {
                 workingContext.executionContext()->variables().setErrorCode(1);
                 throw BadCdTargetException("target directory", target);
             }
         
             workingContext.executionContext()->variables().setErrorCode(1);
             return ScriptFunctions::SRC_Success;
         }
         
         workingContext.executionContext()->variables().assign("PWD", workingContext.fileHandles()->currentDirectory());
         workingContext.executionContext()->variables().assign("OLDPWD", oldPWD);

         workingContext.executionContext()->variables().setErrorCode(0);



         return ScriptFunctions::SRC_Success;
     }

     virtual string abstract() const
     {
#ifdef _MSC_VER
         return "Change directory:  cd [word];  $USERPROFILE is used if word is missing";
#else
         return "Change directory:  cd [word];  $HOME is used if word is missing";
#endif
     }
};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .stopwatch command
//

static string dashS("-s");
static string dashC("-c");
static string dashP("-p");

struct    BrashBuiltinStopwatchExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);


         if(   ( (arguments.argc != 3) && (arguments.argc != 4)  )
            || !(
                        arguments.args_[1] == dashS  // -s
                    ||  arguments.args_[1] == dashC  // -c
                    ||  arguments.args_[1] == dashP  // ?Hmmm.  What could it be????
                )
            ||  (arguments.args_[2].find_first_not_of("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz") < arguments.args_[2].size())
            ||  (arguments.args_[1] == dashC && arguments.argc != 4)
           )
         {
             workingContext.fileHandles()->writeBytes(2, string("\r\n<brash builtin> .stopwatch:  Save, Print, or Check the Time in Milliseconds\r\n"));
             workingContext.fileHandles()->writeBytes(2, string("\r\nUsage:\r\n"));
             workingContext.fileHandles()->writeBytes(2, string("   .stopwatch -s varName # stores the time in milliseconds into variable varName\r\n"));
             workingContext.fileHandles()->writeBytes(2, string("   .stopwatch -p varName # prints the elapsed time as defined by variable varName\r\n"));
             workingContext.fileHandles()->writeBytes(2, string("   .stopwatch -c varName maxElapsed # sets $? based on whether or not a maximim interval\r\n"));
             workingContext.fileHandles()->writeBytes(2, string("                                    # has elapsed.  True, 0,  means NO.\r\n\r\n"));

             workingContext.executionContext()->variables().setErrorCode(1);

             return ScriptFunctions::SRC_Success;

         }

         string const &varname = arguments.args_[2];

         if(arguments.argv[1] == dashS)
         {
             // start the watch in the named variable

             ULONGLONG curTicks = GetTickCount64();

             char buffer[40];

             snprintf(buffer, 40, "%lld", (long long)curTicks);

             executionContext_in->variables().assign(varname, buffer);

         }
         else
         if(arguments.argv[1] == dashP)
         {
             // print the current time

             ULONGLONG curTicks = GetTickCount64();

             char buffer[40];

             long long oldTicks=0;

             sscanf(executionContext_in->variables().find(varname).c_str(), "%lld", &oldTicks);

             curTicks -= oldTicks;

             snprintf(buffer, 40, "%lld", (long long)curTicks);

             workingContext.fileHandles()->writeBytes(1, string(buffer) + nl);

         }
         else
         {
            // presumbably -c

            long long count = 0;  // count for comparison

            sscanf(arguments.args_[3].c_str(), "%lld", &count);

            ULONGLONG curTicks = GetTickCount64();

            char buffer[40];

            long long oldTicks=0;

            sscanf(executionContext_in->variables().find(varname).c_str(), "%lld", &oldTicks);

            curTicks -= oldTicks;

            if(curTicks > count)
            {
               // max count exceeded, so set error code

               workingContext.executionContext()->variables().setErrorCode(1);

               return ScriptFunctions::SRC_Success;
            }


         }



         workingContext.executionContext()->variables().setErrorCode(0);

         return ScriptFunctions::SRC_Success;
     }

     virtual string abstract() const
     {
         return "Timer control:  .timer [ -s | -c ] variableName ; start or print the elapsed time";
     }
};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// DIRS command
//


struct    BrashBuiltinDirsExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);
         
         static string blank(" ");
         
         // for the moment, ignore the options

         string dirstack = workingContext.variables().find("DIRSTACK");

         StrTool::trim(dirstack);


         if(!dirstack.empty())
         {
             workingContext.fileHandles()->writeBytes(1, dirstack);
             workingContext.fileHandles()->writeBytes(1, nl);
         }

         workingContext.variables().setErrorCode(0);

         // redirectedFileHandles, if any will auto destroy/close

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "List DIRSTACK members:  dirs;  See also pushd, popd";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// POPD command
//

struct    BrashBuiltinPopDExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         string oldPWD = workingContext.executionContext()->variables().find("PWD");
         

         if(arguments.argc > 1)
         {
                workingContext.fileHandles()->writeBytes(2, string("popd:  Error:  too many parameters!\r\n"));

                bool   throwExceptionOnError = workingContext.executionContext()->variables().errorMode();   // by default, cd failure is a major error

                if(throwExceptionOnError)
                {
                    workingContext.executionContext()->variables().setErrorCode(1);
                    throw BadCdTargetException("error type", "too many parameters");
                }
                workingContext.executionContext()->variables().setErrorCode(1);
                return ScriptFunctions::SRC_Success;
         }

         string target;

         {
             // pop left most node of $DIRSTACK into target

             string dirstack = workingContext.executionContext()->variables().find("DIRSTACK");

             StrTool::remove_leading(dirstack, '\n');

             size_t eol = dirstack.find_first_of('\n');

             if(eol >= dirstack.size())
             {
                 target = dirstack;
                 dirstack.resize(0);
             }
             else
             {
                 target = dirstack.substr(0, eol);
                 dirstack.erase(0,eol+1);
             }

             workingContext.executionContext()->variables().assign("DIRSTACK", dirstack);

         }

         int err = !workingContext.fileHandles()->setCurrentDirectory(target.c_str());
         
         if(err)
         {
             workingContext.fileHandles()->writeBytes(2, string("popd:  Error, could not change to directory:  ") + target + nl);
         
             bool   throwExceptionOnError = workingContext.executionContext()->variables().errorMode();   // by default, cd failure is a major error
         
             if(throwExceptionOnError)
             {
                 workingContext.executionContext()->variables().setErrorCode(1);
                 throw BadCdTargetException("target directory", target);
             }
         
             workingContext.executionContext()->variables().setErrorCode(1);
             return ScriptFunctions::SRC_Success;
         }
         
         workingContext.executionContext()->variables().assign("PWD", workingContext.fileHandles()->currentDirectory());
         workingContext.executionContext()->variables().assign("OLDPWD", oldPWD);

         workingContext.executionContext()->variables().setErrorCode(0);



         return ScriptFunctions::SRC_Success;
     }

     virtual string abstract() const
     {
         return "Pop the front of $DIRSTACK and go to that directory.";
     }
};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// PUSHD command
//

struct    BrashBuiltinPushDExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         string oldPWD = workingContext.executionContext()->variables().find("PWD");
         

         if(arguments.argc < 2)
         {
            // pushd with no arguments means go home

            string home = workingContext.executionContext()->variables().find("HOME");

            if(home.empty())
            {
#ifdef _MSC_VER
                string userprofile= workingContext.executionContext()->variables().find("USERPROFILE");

                if(userprofile.empty())
                   home="\\";
                else
                   home=userprofile;
#else
                home="/tmp";
#endif
            }

            int err = !workingContext.fileHandles()->setCurrentDirectory(home.c_str());

            if(err)
            {
                workingContext.fileHandles()->writeBytes(2, string("push:  Error, could not change to directory:  ") + home + nl);

                bool   throwExceptionOnError = workingContext.executionContext()->variables().errorMode();   // by default, cd failure is a major error

                if(throwExceptionOnError)
                {
                    throw BadCdTargetException("directory", home);
                }

                return ScriptFunctions::SRC_Success;
            }

            workingContext.executionContext()->variables().assign("PWD", workingContext.fileHandles()->currentDirectory());
            workingContext.executionContext()->variables().assign("OLDPWD", oldPWD);

            {
               // now push the current directory on the stack
   
               string dirstack = workingContext.executionContext()->variables().find("DIRSTACK");
   
               StrTool::remove_leading(dirstack,'\n');
   
               dirstack = oldPWD + nl + dirstack;
   
               workingContext.executionContext()->variables().assign("DIRSTACK", workingContext.fileHandles()->currentDirectory());
            }

            workingContext.executionContext()->variables().setErrorCode(0);
            return ScriptFunctions::SRC_Success;


         }

         if(arguments.argc > 2)
         {
                workingContext.fileHandles()->writeBytes(2, string("pushd:  Error:  too many parameters!\r\n"));

                bool   throwExceptionOnError = workingContext.executionContext()->variables().errorMode();   // by default, cd failure is a major error

                if(throwExceptionOnError)
                {
                    workingContext.executionContext()->variables().setErrorCode(1);
                    throw BadCdTargetException("error type", "too many parameters");
                }
                workingContext.executionContext()->variables().setErrorCode(1);
                return ScriptFunctions::SRC_Success;
         }

         string target = arguments.args_[1];

         int err = !workingContext.fileHandles()->setCurrentDirectory(target.c_str());
         
         if(err)
         {
             workingContext.fileHandles()->writeBytes(2, string("pushd:  Error, could not change to directory:  ") + target + nl);
         
             bool   throwExceptionOnError = workingContext.executionContext()->variables().errorMode();   // by default, cd failure is a major error
         
             if(throwExceptionOnError)
             {
                 workingContext.executionContext()->variables().setErrorCode(1);
                 throw BadCdTargetException("target directory", target);
             }
         
             workingContext.executionContext()->variables().setErrorCode(1);
             return ScriptFunctions::SRC_Success;
         }
         
         workingContext.executionContext()->variables().assign("PWD", workingContext.fileHandles()->currentDirectory());
         workingContext.executionContext()->variables().assign("OLDPWD", oldPWD);

         {
            // now push the current directory on the stack

            string dirstack = workingContext.executionContext()->variables().find("DIRSTACK");

            StrTool::remove_leading(dirstack,'\n');

            dirstack = oldPWD + nl + dirstack;

            workingContext.executionContext()->variables().assign("DIRSTACK", dirstack);
         }


         workingContext.executionContext()->variables().setErrorCode(0);



         return ScriptFunctions::SRC_Success;
     }

     virtual string abstract() const
     {
#ifdef _MSC_VER
         return "Change directory:  pushd [word];  $USERPROFILE is used if word is missing";
#else
         return "Change directory:  pushd [word];  $HOME is used if word is missing";
#endif
     }
};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// PWD command
//


struct    BrashBuiltinPwdExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         workingContext.fileHandles()->writeBytes(1, workingContext.executionContext()->variables().find("PWD") + nl );


         return ScriptFunctions::SRC_Success;
     }

     virtual string abstract() const
     {
         return "Print the current directory name:  pwd";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// BREAK command
//



struct    BrashBuiltinBreakExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
         if(executionContext_in->loggingExecution())
         {
             executionContext_in->fileHandles_->writeBytes(2, string("+ break\r\n") );
         }

         return ScriptFunctions::SRC_Break;
     }

     virtual string abstract() const
     {
         return "Break out of the current loop  or case statement;  break";
     }

};

//
// CONTINUE command
//


struct    BrashBuiltinContinueExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         if(executionContext_in->loggingExecution())
         {
             executionContext_in->fileHandles_->writeBytes(2, string("+ continue\r\n") );
         }

         return ScriptFunctions::SRC_Continue;
     }

     virtual string abstract() const
     {
         return "Skip immediately to the next iteration of the current loop;  continue";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// BASENAME command
//

static string blank(" ");


struct    BrashBuiltinBasenameExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);
         
         bool handleEscapeSequences = false;

         bool processingOptions = true;
         bool printNewline = true;
         
         #ifdef DebugBuild
             CXXTLS_LOG("BrashBuiltinBasename:  executing:  " << arguments.args_);
         #endif

         if(arguments.args_.size() < 2 || arguments.args_.size() > 3 )
         {
             workingContext.fileHandles()->writeBytes(1, string("Error:  basename parameters invalid, use:  basename path [suffix]\r\n") );

             workingContext.variables().setErrorCode(0);
             
         }
         else
         {
             FileName tmp(arguments.args_[1]);

             string basepart = tmp.basename();

             if(arguments.args_.size() == 3)
             {
                string suffix = arguments.args_[2];

                if(basepart.size() >= suffix.size())
                {
                    size_t index = basepart.size() - suffix.size();

                    if(0 == memcmp(basepart.data() + index, suffix.data(), suffix.size()))
                    {
                        basepart = basepart.substr(0,index);
                    }
                }
             }
             
             workingContext.fileHandles()->writeBytes(1, basepart + nl);
             
             workingContext.variables().setErrorCode(0);
             
         }
         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Print the basename part of a full pathname;  basename path [suffix];  Optionally, remove the suffix too.";
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// DIRNAME command
//

struct    BrashBuiltinDirnameExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);
         
         bool handleEscapeSequences = false;

         bool processingOptions = true;
         bool printNewline = true;
         
         #ifdef DebugBuild
             CXXTLS_LOG("BrashBuiltinBasename:  executing:  " << arguments.args_);
         #endif

         if(arguments.args_.size() != 2)
         {
             workingContext.fileHandles()->writeBytes(1, string("Error:  dirname parameters invalid, use:  dirname path\r\n") );

             workingContext.variables().setErrorCode(0);
             
         }
         else
         {
             FileName tmp(arguments.args_[1]);

             string basepart = tmp.dirname();

             // remove the trailing /

             if(basepart.size() > 1 )
             {
                char c = basepart[ basepart.size()-1 ];

                if(c == '/' || c == '\\')
                {
                    basepart = basepart.substr(0, basepart.size()-1);
                }
             }

             #ifdef _MSC_VER
             if (basepart.size() == 2 && basepart[1] == ':')
                 basepart += '\\';
             #endif

             workingContext.fileHandles()->writeBytes(1, basepart + nl);
             
             workingContext.variables().setErrorCode(0);
             
         }
         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Print the directory part of a pathname";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// ECHO command
//



struct    BrashBuiltinEchoExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);
         
         bool handleEscapeSequences = false;

         bool processingOptions = true;
         bool printNewline = true;
         
         int dataStartOption = 1;

         #ifdef DebugBuild
             CXXTLS_LOG("BrashBuiltinEcho:  executing:  " << arguments.args_);
         #endif

         for(int parmNo = 1; parmNo < arguments.argc; ++parmNo)
         {
             // check options here

             string curArg = arguments.args_[parmNo];

             if(curArg.empty() )
                processingOptions = false;
             else
             if (processingOptions)
             {
                if(curArg[0] == '-')
                {
                   dataStartOption = parmNo + 1;    // when deciding to print a leading space before outputing a pattern, below
                                                    // do not include any  parameters up to this point

                   // process options in this group

                   string::const_iterator f = curArg.begin(),
                                          l = curArg.end()
                               ;
                   ++f;  // skip leading -

                   bool skipThisOne = false;

                   while(f != l  && processingOptions)
                   {
                       switch(*f++)
                       {
                          case 'e':  
                            handleEscapeSequences = true;
                            break;
                          case 'n':
                            printNewline = false;
                            break;
                          case '-':
                            if(*f == 0)
                            {
                                // -- stops options processing, but "-- fred" does not

                                skipThisOne = true;
                            }
                            else
                                --dataStartOption;  // count this one as a non-option
                            processingOptions = false;
                            break;
                          default:
                            {
                                --dataStartOption;  // count this one as a non-option
                                processingOptions = false;
                            }
                            break;
                       }
                   }

                   if(skipThisOne || processingOptions)
                      continue;

                }
                else
                   processingOptions = false;
             }


             if(parmNo > dataStartOption)
             {
                 workingContext.fileHandles()->writeBytes(1, blank);
             }

             if(handleEscapeSequences)
             {
                 char const *arg = arguments.argv[parmNo];
                 
                 string tmp;
                 
                 
                 if(arg)
                 {
                     while(*arg)
                     {
                        char c = *arg++;
                 
                        if(c != '\\')
                           tmp += c;
                        else
                        {
                           c = *arg++;
                 
                           if(c == 0)
                           {
                              tmp += '\\';
                              break;
                           }
                 
                           switch(c)
                           {
                              default:  
                                        tmp += '\\';  // leave the backslash unless we recognize it
                                        tmp += c; 
                                        break;
                 
                              case 'r':  tmp += '\r'; break;
                              case 'n':  tmp += '\n'; break;
                              case 't':  tmp += '\t'; break;
                              case 'a':  tmp += '\a'; break;
                              case 'b':  tmp += '\b'; break;
                              case 's':  tmp += ' ' ; break;
                              case 'e':  tmp += 0x1a; break;

                             case '0':
                                 {
                                     // parse octal

                                     int v = 0;


                                     while(c != 0 )  // nul in this case
                                     {
                                         c = *arg;

                                         if(c < '0' || c >= '8')
                                            break;

                                         ++arg;

                                         v <<= 3;
                                         
                                         v += c - '0';
                                     }

                                     tmp += v;

                                 }
                                 break;

                           }
                 
                        }
                 
                     }
                 }
                 
                 
                 workingContext.fileHandles()->writeBytes(1, tmp) ;
             }
             else
             {
                 workingContext.fileHandles()->writeBytes(1, arguments.args_[parmNo]) ;
             }
         }

         if(printNewline)
             workingContext.fileHandles()->writeBytes(1, nl);

         workingContext.variables().setErrorCode(0);

         // redirectedFileHandles, if any will auto destroy/close

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Print text to stdout:  echo [options] [word] ...;  Options:  -e:  interpret escape sequences, -n:  print no newline";
     }

};




//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .LS command
//

static string dashL("-l");

string longEntry(FileStatus const &fstat)
{
    string mode = fstat.mode;
    string time = fstat.time;
    string size = lexical_cast<string>(fstat.size, fmtspec(10,'d'));

    return mode + space + time + space + size;
}



struct    BrashBuiltinLSExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);
         
         static string blank(" ");

         bool longListing = false;
         bool lookingForOptions = true;
         int  foundFiles = 0;               // count of file name patterns found on the command line

         string curdir = workingContext.fileHandles()->currentDirectory();

         // file name pattern searches should be fully qualified directory names, though we will not always report them as such,
         // because this is a multithreaded application and another thread can change the name of the current directory
         // at any time.

         using namespace CommandLine;

         static OptionDeclaration allowedOptions[]=
         {
             OptionDeclaration("l",    0, 0, "print long listing"),
             OptionDeclaration("a",    0, 0, "list all files"),
             OptionDeclaration("r",    0, 0, "reverse the sort order"),
             OptionDeclaration("R",    0, 0, "recursive directory processing"),
             OptionDeclaration("Q",    0, 0, "wrap names with double quotes"),
             OptionDeclaration("h",    0, 0, "get help"),
             OptionDeclaration("help", 0, 1, "get help"),
             OptionDeclaration("d",    0, 0, "do not print directory contents"),
         };


         try 
         {
             CommandLine::Options options(arguments.args_.begin(),
                                          arguments.args_.end(),
                                          allowedOptions,
                                          sizeof(allowedOptions) / sizeof(allowedOptions[0])
                                         );

             bool listAllFiles= options.isSet("a");
             bool reverseSort = options.isSet("r");
             bool doubleQuotes= options.isSet("Q");
             bool printDirectoryContents = !options.isSet("d");

             int  findMatchingOptions = 1;

             if(options.isSet("R"))
                findMatchingOptions |= 0x20;   // turn on recursive.


             if(options.isSet("h") || options.isSet("help") )
             {
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\n<brash builtin> .ls:  list directories\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\nUsage:\r\n\r\n    .ls [options] [file ...]\r\n\r\n"));

                 stringstream  tmp;

                 options.printHelp(tmp);

                 string helpText = tmp.str();

                 workingContext.fileHandles()->writeBytes(2, helpText + nl);

             }
             else
             {
                 //
                 //  Gather the file names from the command line, or infer them if none are found
                 //

                 vector<string> working;

                 if(options.size() == 1)
                 {
                    // user has not specified any files to list, so use the current directory

                    FileName::sorted_names_t  matches; // matching file names
                                
#ifdef _MSC_VER
                    FileName::find_matching(curdir + "\\*", &matches, findMatchingOptions);   // always include full pathnames in result
#else
                    FileName::find_matching(curdir + "/*", &matches, findMatchingOptions);   // always include full pathnames in result
#endif

                    vector<string> order;

                    order.reserve(working.size());

                    if(!reverseSort)
                       CXXTLS_FOREACH(string const &cur, matches)  
                       {
#ifdef _MSC_VER
                           string tmp;

                           CXXTLS_FOREACH(char c, cur)
                              if(c == '/')
                                 tmp += '\\';
                              else
                                 tmp += c;

                           order.push_back(tmp);
#else
                           order.push_back(cur);
#endif
                       }
                    else
                    {
                        FileName::sorted_names_t::iterator f = matches.begin(), 
                                                           l = matches.end();

                        while( l != f )
                        {
                           --l;

#ifdef _MSC_VER

                           string tmp;

                           CXXTLS_FOREACH(char c, *l)
                              if(c == '/')
                                 tmp += '\\';
                              else
                                 tmp += c;


                           order.push_back(tmp);
#else
                           order.push_back(*l);
#endif
                        }
                    }


                    CXXTLS_FOREACH(string const &cur, order)
                    {
                        if(listAllFiles)
                        {
                            working.push_back(cur);
                        }
                        else
                        {
                            size_t lastSlash= cur.find_last_of("\\/");

                            bool printit = true;

                            if(   lastSlash < cur.size() 
                               && lastSlash < cur.size()-1   // lasSlash is unsigned, so you can't conflate the two test expressions
                               && cur[lastSlash+1] =='.'
                              )
                            {
                                printit = false;
                            }


                            if(printit)
                              working.push_back(cur);
                        }
                    }

                 }
                 else
                 for(size_t i = 1; i < options.size(); ++i)
                 {
                     // handle arguments explicitly stated by the user

                     string const &cur = options.arg(i);

                     FileName tmp(cur);

                     tmp.convert_to_absolute_path();

                     if(printDirectoryContents && tmp.is_dir())
                     {
                         // include its contents

                         FileName::sorted_names_t  matches; // matching file names
                                     
#ifdef _MSC_VER          
                         FileName::find_matching(string(tmp) + "\\*", &matches, findMatchingOptions);   // always include full pathnames in result
#else                    
                         FileName::find_matching(string(tmp) + "/*", &matches, findMatchingOptions);   // always include full pathnames in result
#endif                   

                         vector<string> order;

                         order.reserve(working.size());

     
                         if(!reverseSort)
                            CXXTLS_FOREACH(string const &cur, matches)  
                            {
#ifdef _MSC_VER
                               string tmp;
                      
                               CXXTLS_FOREACH(char c, cur)
                                  if(c == '/')
                                     tmp += '\\';
                                  else
                                     tmp += c;
                      
                               order.push_back(tmp);
#else                 
                               order.push_back(cur);
#endif
                            }
                         else
                         {
                             FileName::sorted_names_t::iterator f = matches.begin(), 
                                                                l = matches.end();
     
                             while( l != f )
                             {
                                --l;
     
#ifdef _MSC_VER

                              string tmp;
                             
                              CXXTLS_FOREACH(char c, *l)
                                 if(c == '/')
                                    tmp += '\\';
                                 else
                                    tmp += c;
                             
                             
                              order.push_back(tmp);
#else                        
                              order.push_back(*l);
#endif
                                
                             }
                         }

                         CXXTLS_FOREACH(string const &cur, order)
                         {
                             if(listAllFiles)
                             {
                                 working.push_back(cur);
                             }
                             else
                             {

                                 size_t lastSlash= cur.find_last_of("\\/");
                                 
                                 bool printit = true;
                                 
                                 if(   lastSlash < cur.size() 
                                    && lastSlash < cur.size()-1   // lasSlash is unsigned, so you can't conflate the two test expressions
                                    && cur[lastSlash+1] =='.'
                                   )
                                 {
                                     printit = false;
                                 }
                                 
                                 
                                 if(printit)
                                   working.push_back(cur);
                            
                             }
                         }

                     }
                     else
                     {
#ifdef _MSC_VER
                         string copy;
                         CXXTLS_FOREACH(char c, tmp)
                             copy += (c == '/' ? '\\': c);
                         tmp = copy;
#endif
                         if (tmp.exists())
                             working.push_back(tmp);
                         else
                         {
                             workingContext.fileHandles()->writeBytes(2, error + doubleQuote + string(tmp) + doubleQuote + string(" does not exist\r\n"));
                         }

                     }


                 }

                 // at this point, working is populated with the names of all files to be printed.
                 // on windows, all / have been coverted to \\.


                 // now print the results

                 longListing = options.isSet("l");


                 int outputRows = 0;

                 CXXTLS_FOREACH(string const &cur, working)
                 {
                     if( (outputRows % 100) == 0 )
                     {
                        if(!consoleStillRunning())
                        {
                            workingContext.fileHandles()->writeBytes(2, "^C\r\n", 4);
                            break;
                        }

                     }

                     ++outputRows; // only used for delaying checks between calls to consoleStillRunning()


                     string tmp(cur.substr(0,curdir.size()));

#ifdef _MSC_VER
                     string copy;
                     CXXTLS_FOREACH(char c, tmp)
                         copy += (c == '/' ? '\\': c);
                     tmp = copy;
#endif

                     if(tmp == curdir)
                     {
                        tmp = cur;
                        tmp.erase(0, curdir.size() + 1);
                     }
                     else
                        tmp = cur;

                     string printName = tmp;

                     FileName fname(cur);

                     cxxtls::ScriptThread::ManagedFileHandles::IO_ERROR  ioError;
                     
                     if(longListing)
                     {
                     
                         FileStatus status;
                     
                         if(fname.file_stat(&status))
                         {
                             // error reading status
                     
                             if(doubleQuotes)
                                 ioError = workingContext.fileHandles()->writeBytes(1, doubleQuote + printName+ doubleQuote + nl);
                             else
                                 ioError = workingContext.fileHandles()->writeBytes(1, printName+nl);
                         }
                         else
                         {
                             // found data
                     
                             string tmp = longEntry(status);  // convert status to string
                     
                             if(doubleQuotes)
                                 ioError = workingContext.fileHandles()->writeBytes(1, tmp + space + doubleQuote + printName + doubleQuote + nl);
                             else
                                 ioError = workingContext.fileHandles()->writeBytes(1, tmp + space + printName + nl);
                     
                         }
                     
                     }
                     else
                     {
                         if(doubleQuotes)
                             ioError = workingContext.fileHandles()->writeBytes(1, doubleQuote + printName + doubleQuote + nl);
                         else
                             ioError = workingContext.fileHandles()->writeBytes(1, printName+nl);
                     }

                     if(ioError)
                        break; // something happened while we are writing, just give it up

                 }  // write all results


                 workingContext.variables().setErrorCode(0);

                 return ScriptFunctions::SRC_Success;

             }



         }
         catch(exception &e)
         {
             workingContext.variables().setErrorCode(1);
    
             workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .ls Error:  ") + e.what() + nl);

    
             return ScriptFunctions::SRC_Success;
             

         }



         workingContext.variables().setErrorCode(0);

         // redirectedFileHandles, if any will auto destroy/close

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "list file names and unix style attributes:  .ls [options] [files]; -h for help"; 
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .CAT command
//

struct    BrashBuiltinCatExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         if(arguments.args_.size() < 2)
         {
             char buffer[32768];

             for(;;)
             {
                 size_t bytesRead;


                 if(!consoleStillRunning())
                 {
                     workingContext.fileHandles()->writeBytes(2, "^C\r\n", 4);
                     break;
                 }

                 ScriptThread::ManagedFileHandles::IO_ERROR rc = workingContext.fileHandles()->readBytes(0, sizeof(buffer), buffer, &bytesRead);


                 if(bytesRead == 0 || rc != ScriptThread::ManagedFileHandles::IOERR_OK)
                    break;

                 workingContext.fileHandles()->writeBytes(1, buffer, bytesRead);

             }
         }
         else
         {
             for(size_t i = 1; i < arguments.args_.size(); ++i)
             {
                 if(!consoleStillRunning())
                 {
                     workingContext.fileHandles()->writeBytes(2, "^C\r\n", 4);
                     break;
                 }

                 if(arguments.args_[i] == "-")
                 {
                    char buffer[32768];

                    for(;;)
                    {
                       size_t bytesRead = 0;

                       if(!consoleStillRunning())
                       {
                           workingContext.fileHandles()->writeBytes(2, "^C\r\n", 4);
                           break;
                       }

                       workingContext.fileHandles()->readBytes(0, sizeof(buffer), buffer, &bytesRead);

                       if(bytesRead == 0)
                          break;

                       workingContext.fileHandles()->writeBytes(1, buffer, bytesRead);
                    }
                 }
                 else
                 {
                     
                     ifstream file(arguments.args_[i], std::ifstream::binary);
                     
                     
                     if(!file.good())
                     {
                         workingContext.fileHandles()->writeBytes(2, arguments.args_[i] + ": read error\r\n");
                         continue;
                     }

                     while(file)
                     {
                         char buffer[32768];
                     
                         if(!consoleStillRunning())
                         {
                             workingContext.fileHandles()->writeBytes(2, "^C\r\n", 4);
                             break;
                         }


                         file.read(buffer, sizeof(buffer));
                     
                         size_t readBytes = file.gcount();
                     
                         if(readBytes)
                             workingContext.fileHandles()->writeBytes(1, buffer, readBytes);
                     }
                 }

             }
         }

         workingContext.variables().setErrorCode(0);

         // redirectedFileHandles, if any will auto destroy/close

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Copy files to stdout:  .cat [file ...]";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// EXIT command
//


struct    BrashBuiltinExitExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         cxxtls::ScriptThread::ThreadContext        *executionContext = workingContext.executionContext();
         cxxtls::ScriptThread::ManagedFileHandles   *workingFileHandles = workingContext.fileHandles();


         vector<string> args;

         evaluateCommandLine(commandLine, executionContext, args, skipArg0);

         if(executionContext->loggingExecution())
         {
             string logMessage("+");

     
             CXXTLS_FOREACH(std::string const &cur, args)
             {
                 logMessage += ' ';
                 logMessage += cur;
             }

             logMessage += nl;

             executionContext->fileHandles_->writeBytes(2, logMessage);
               
         }


         if(args.size() <= 2)
         {

              int parmNo=0;
              
              CXXTLS_FOREACH(string const &cur, args)
              {
     
                  if(parmNo == 0)
                  {
                     ++parmNo;
                     continue; // ignore the word "exit"
                  }
     
                  // at parm 1
     
                  try
                  {
                      int exitcode = lexical_cast<int>(cur);
     
                      BrashScriptCommandHandler::setExitCode(exitcode);

                      executionContext_in->variables().setErrorCode(exitcode);
     
                  }
                  catch(...)
                  {
                      BrashScriptCommandHandler::setExitCode(1);
                      workingFileHandles->writeBytes(1, (string("exit: error:  invalid exit statement parameter:  ") + cur + nl));
                  }
     
                  return ScriptFunctions::SRC_Return;  // signal to the BashStatementList::execute() function to trigger a shutdown
     
     
                  // should never get here
              
                  ++parmNo;
              }

         }
         else
         {
              workingFileHandles->writeBytes(1, (string("exit: error: too many exit statement parameters\r\n")));
         }

         return ScriptFunctions::SRC_Return; // signal to the BashStatementList::execute() function to trigger a shutdown
     }

     virtual string abstract() const
     {
         return "Exit brash:  exit [exitCode];  exit occurs when all other jobs are finished";
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// GETOPTS command
//


struct    BrashBuiltinGetOptsExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {


         vector<string> args;

         evaluateCommandLine(commandLine, executionContext, args, skipArg0);

         if(executionContext->loggingExecution())
         {
             string logMessage("+");

     
             CXXTLS_FOREACH(std::string const &cur, args)
             {
                 logMessage += ' ';
                 logMessage += cur;
             }

             logMessage += nl;

             executionContext->fileHandles_->writeBytes(2, logMessage);
               
         }


         if(args.size() < 3)
         {
            executionContext->fileHandles_->writeBytes(1, (string("getopts: error, usage: getopts optionsList varname [arg ...]\r\n") ));
            executionContext->variables().setErrorCode(2);   // bad parms, etc
         }
         else
         try
         {
             string optlist = args[1];
             string varname = args[2];
             string optind  = executionContext->variables().find("OPTIND");
             string optoff  = executionContext->variables().find("OPTOFF");

             vector<char const *> arguments;


             if(args.size() > 3)
             {
                 // the arguments to be processed are on the command line,

                 arguments.reserve(args.size());

                 arguments.push_back("unused");  

                 for(size_t i = 3; i < args.size(); ++i)
                    arguments.push_back(args[i].c_str());
             }
             else
             {
                 // the arguments to be processed are found in "$@"

                 arguments.reserve(executionContext->variables().arguments().size());

                 CXXTLS_FOREACH(string const &cur, executionContext->variables().arguments())
                    arguments.push_back(cur.c_str());
             }

             int argc = int(arguments.size());

             if (argc == 0)
             {
                 executionContext->variables().assign("OPTIND", "1");
                 executionContext->variables().assign("OPTOFF", "0");
                 executionContext->variables().setErrorCode(1);   // Error:  no data to work on

                 return ScriptFunctions::SRC_Success;
            }

             char const **argv = &arguments[0];

             if(optind.empty())
                optind="1";

             if(optoff.empty())
                optoff="0";

             int startIndex = 0;
             int startOffset= 0;

             sscanf(optind.c_str(), "%d", &startIndex);
             sscanf(optoff.c_str(), "%d", &startOffset);

             if(startIndex < 1)
             {
                 startIndex = 1;
             }

             GetOpts optProc(argc, (char *const *&)argv, optlist);

             optProc.optind = startIndex;
             optProc.startOffset = startOffset; 

             int optionValue;

             optionValue = optProc.getNextOption();

             if(optionValue < 0)
             {
                executionContext->variables().setErrorCode(1);   // end of list
                return ScriptFunctions::SRC_Success;
             }
            
             executionContext->variables().assign("OPTIND", lexical_cast<string>(optProc.optind));
             executionContext->variables().assign("OPTOFF", lexical_cast<string>(optProc.startOffset));


             if(optionValue == '?')
             {
                 string invopt;

                 invopt += char(optProc.optopt);

                 executionContext->variables().assign("OPTARG", invopt);
             }
             else
             if(optionValue == ':')
             {
                 // missing parameter to an option

                 string optvalue;

                 optvalue += char(optProc.optopt);

                 executionContext->variables().assign("OPTARG", optvalue);
             }
             else
                 executionContext->variables().assign("OPTARG", optProc.optarg ? optProc.optarg : "");

             string optvalue;

             optvalue += char(optionValue);

             executionContext->variables().assign(varname, optvalue);

             executionContext->variables().setErrorCode(0);   // successful return
         }
         catch(...)
         {
             executionContext->fileHandles_->writeBytes(1, (string("getopts:  error: invalid optons or OPTIND variable\r\n") ));
             executionContext->variables().setErrorCode(2);   // end of list
         }


         return ScriptFunctions::SRC_Success; 
     }

     virtual string abstract() const
     {
         return "Parse options from $@:  getopts  optlst var;  var is the variable to hold parsed option, optlist='ab:cd:e'";
     }

};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// RETURN command
//


struct    BrashBuiltinReturnExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         cxxtls::ScriptThread::ThreadContext        *executionContext = workingContext.executionContext();
         cxxtls::ScriptThread::ManagedFileHandles   *workingFileHandles = workingContext.fileHandles();


         vector<string> args;

         evaluateCommandLine(commandLine, executionContext, args, skipArg0);

         if(executionContext->loggingExecution())
         {
             string logMessage("+");

     
             CXXTLS_FOREACH(std::string const &cur, args)
             {
                logMessage += ' ';
                logMessage += cur;
             }

             logMessage += nl;
             executionContext->fileHandles_->writeBytes(2, logMessage);
               
         }


         if(args.size() <= 2)
         {

              int parmNo=0;
              
              CXXTLS_FOREACH(string const &cur, args)
              {
     
                  if(parmNo == 0)
                  {
                     ++parmNo;
                     continue; // ignore the word "return"
                  }
     
                  // at parm 1
     
                  try
                  {
                      int exitcode = lexical_cast<int>(cur);
     
                      BrashScriptCommandHandler::setExitCode(exitcode);
                      executionContext->variables().setErrorCode(exitcode);
     
                  }
                  catch(...)
                  {
                      workingFileHandles->writeBytes(1, (string("return: error:  invalid return statement parameter:  ") + cur + nl));
                  }
     
                  return ScriptFunctions::SRC_Return;  // signal to the BashStatementList::execute() function to trigger a shutdown
     
     
                  // should never get here
              
                  ++parmNo;
              }

         }
         else
         {
              workingFileHandles->writeBytes(1, (string("return: error: too many return statement parameters\r\n")));
         }


         return ScriptFunctions::SRC_Return; // signal to the BashStatementList::execute() function to trigger a shutdown
     }

     virtual string abstract() const
     {
         return "Return from the current function:  return [exitCode]";
     }


};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// SLEEP command
//


struct    BrashBuiltinSleepExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         CXXTLS_DEBUG("BrashSleepExecutor");

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         cxxtls::ScriptThread::ThreadContext        *executionContext = workingContext.executionContext();
         cxxtls::ScriptThread::ManagedFileHandles   *workingFileHandles = workingContext.fileHandles();


         vector<string> args;

         evaluateCommandLine(commandLine, executionContext, args, skipArg0);

         if(executionContext->loggingExecution())
         {
             string logMessage("+");

     
             CXXTLS_FOREACH(std::string const &cur, args)
             {
                 logMessage += ' ';
                 logMessage += cur;
             }

             logMessage += nl;

             executionContext->fileHandles_->writeBytes(2, logMessage);
               
         }

         if(args.size() != 2)
         {
             FileName tmp(commandLine->sourceFile_);
             tmp.convert_to_absolute_path();

             workingFileHandles->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                               string("sleep:  error, expected a sleep time in seconds (1.05, for example)\r\n"));

             executionContext->variables().setErrorCode(1);

             return ScriptFunctions::SRC_Success;

         }

         static string blank(" ");

         int parmNo=0;
         
         CXXTLS_FOREACH(string const &cur, args)
         {

             if(parmNo == 0)
             {
                ++parmNo;
                continue; // ignore the word "exit"
             }


             // at parm 1

             try
             {
                 double sleepLength = lexical_cast<double>(cur);

#ifdef _MSC_VER
                 long sleepParm = 1000 * sleepLength;  

                 // sleep in 1/10'th second increments, check for ^C each loop
                 // until the time is up

                 while(sleepParm != 0)
                 {
                     
                     if(!consoleStillRunning())
                     {
                        throw ScriptThread::UserIOCancellation();
                     }
                 
                 
                     if(sleepParm >= 100)
                     {
                         Sleep(100); // sleep 0.1 seconds

                         sleepParm -= 100;
                     }
                     else
                     {
                        Sleep(sleepParm);
                        break;
                     }
                     
                     
                 }
#else
 
                 long sleepParm = 1000000 * sleepLength;  
                 usleep(sleepParm);
#endif

             }
             catch(...)
             {
                 workingFileHandles->writeBytes(1, (string("sleep: error:  invalid sleep statement parameter:  ") + cur + nl));
                 executionContext->variables().setErrorCode(1);

                 if (!consoleStillRunning())
                 {
                     throw ScriptThread::UserIOCancellation();
                 }

                 return ScriptFunctions::SRC_Success;
                 
             }

              return ScriptFunctions::SRC_Success;


             // should never get here
         
             ++parmNo;
         }

         executionContext->variables().setErrorCode(0);


         // redirectedFileHandles, if any will auto destroy/close

         return ScriptFunctions::SRC_Success;
     }

     virtual string abstract() const
     {
         return "Wait for a count of seconds:  sleep [word];  word is a floating point number of seconds ";
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// ENV command
//

struct    BrashBuiltinEnvExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         cxxtls::ScriptThread::ThreadContext        *executionContext = workingContext.executionContext();
         cxxtls::ScriptThread::ManagedFileHandles   *workingFileHandles = workingContext.fileHandles();


         vector<string> args;

         evaluateCommandLine(commandLine, executionContext, args, skipArg0);

         if(executionContext->loggingExecution())
         {
             string logMessage("+");

     
             CXXTLS_FOREACH(std::string const &cur, args)
             {
                 logMessage += ' ';
                 logMessage += cur;
             }

             logMessage += nl;

             executionContext->fileHandles_->writeBytes(2, logMessage);
               
         }


         // ignore the command line for the moment, we'll need to fix this 


          cxxtls::StringExpr::ExecutionContext   &context = executionContext->variables();
          cxxtls::StringExpr::ExecutionVariables &env = context.variables();

          const cxxtls::StringExpr::ExecutionVariables &supers = context.superseeds();
   

          map<string,string>  tmpCopy;   // a copy of the named variables which will be merged with the
                                         // superseeds.
   

          // copy the exported variables to the tmpCopy

          cxxtls::StringExpr::ExecutionVariables::const_iterator f = env.begin(), l = env.end();
   
          while(f != l)
          {
              cxxtls::StringExpr::ExecutionVariables::VariablesMap::value_type const &cur = *f++;

              if(!cur.second.exportFlag())
              {
                 continue;
              }

              if(cur.first[0] == 'B' && 
                 isdigit(cur.first[cur.first.size()-1])
                 && cur.first.substr(0,BRASH_RESULT_VAR_PREFIX.size()) == BRASH_RESULT_VAR_PREFIX
                )
              {
                  // do not include BRASH_PIPE_RESULTS_ variables in the report
              
                  continue;
              }

              tmpCopy[cur.first] = cur.second;
   
          }

          // perform needed superseeds


          f = supers.begin();
          l = supers.end();

          while(f != l)
          {
             const string &name = f->first;
             const string &value= f->second;

             ++f;

             map<string,string>::iterator where = tmpCopy.find(name);

             if(where != tmpCopy.end())
             {
                 where->second = value;   // change the existing value
             }
             else
                 tmpCopy[name] = value;   // ad a new one
          }

          // now write out the variables list


          typedef map<string,string>::value_type value_type;

          CXXTLS_FOREACH(value_type const &var, tmpCopy)
          {
              string value = var.first + equalSign + var.second + nl;

              workingFileHandles->writeBytes(1, value);

          }




   
          executionContext->variables().setErrorCode(0);
          return ScriptFunctions::SRC_Success;


     }

     virtual string abstract() const
     {
         return "Print the exported environment variables";
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// BUILTINS command
//

map<string, BrashBuiltinCommandExecutor*> builtinCommands; // this is a constant after threading starts

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// BUILTIN command
//



struct    BrashBuiltinBuiltinExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
         std::list<cxxtls::StringExpression*> const &args = *commandLine->arguments_;

         if(args.size() < 2)
         {
             FileName tmp(commandLine->sourceFile_);
             tmp.convert_to_absolute_path();

             executionContext_in->fileHandles_->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                              string("brash:  Error:  builtin requires a command name\r\n"));

             executionContext_in->variables().setErrorCode(1);

             return ScriptFunctions::SRC_Success;

         }


         StringArray output;

         std::list<cxxtls::StringExpression*>::const_iterator cur = args.begin();

         ++cur;
      
         (*cur)->execute(&executionContext_in->variables(), output);

         string commandName = output.combine();

         if(commandName == "builtin")
         {
             FileName tmp(commandLine->sourceFile_);
             tmp.convert_to_absolute_path();

             executionContext_in->fileHandles_->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                              string("brash:  Error:  builtin cannot execute itself\r\n"));

             executionContext_in->variables().setErrorCode(1);

             return ScriptFunctions::SRC_Success;
         }

         map<string, BrashBuiltinCommandExecutor*>::iterator where = builtinCommands.find(commandName);

         if(where ==  builtinCommands.end())
         {
             FileName tmp(commandLine->sourceFile_);
             tmp.convert_to_absolute_path();

             executionContext_in->fileHandles_->writeBytes(2,  tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                               string("brash:  Error:  ") + commandName + string(" is not a builtin.  Use 'builtins' for a list\r\n"));

             executionContext_in->variables().setErrorCode(1);

             return ScriptFunctions::SRC_Success;
         }

         StatementReturnCode rc;

         rc =  where->second->execute(commandLine,
                                       executionContext_in,
                                       functionScope,
                                       jobController,
                                       true
                                      );

         return rc;

     }
     virtual string abstract() const
     {
         return "Execute builtin command:  builtin cmdName [args]";
     }
};



struct    BrashListBuiltinsExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {


         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         cxxtls::ScriptThread::ThreadContext        *executionContext = workingContext.executionContext();
         cxxtls::ScriptThread::ManagedFileHandles   *workingFileHandles = workingContext.fileHandles();


         vector<string> args;

         evaluateCommandLine(commandLine, executionContext, args, skipArg0);

         if(executionContext->loggingExecution())
         {
             string logMessage("+");

     
             CXXTLS_FOREACH(std::string const &cur, args)
             {
                 logMessage += ' ';
                 logMessage += cur;
             }

             logMessage += nl;

             executionContext->fileHandles_->writeBytes(2, logMessage);
               
         }

         typedef map<string, BrashBuiltinCommandExecutor*>::value_type  value_type;

         CXXTLS_FOREACH( value_type const &cur, builtinCommands )
         {
             char buffer[400];

             snprintf(buffer, 400, "%-12s", cur.first.c_str());

             workingFileHandles->writeBytes(1, buffer, strlen(buffer));
             workingFileHandles->writeBytes(1, "    ", 4);

             workingFileHandles->writeBytes(1, cur.second->abstract() );

             workingFileHandles->writeBytes(1, nl);
         }

         workingFileHandles->writeBytes(1, "\r\n\r\nTestExpr syntax:\r\n");
         workingFileHandles->writeBytes(1, "    [ -f file ] -- check for file or directory existence\r\n");
         workingFileHandles->writeBytes(1, "    [ -e file ] -- confirm that file exists and is not a directory\r\n");
         workingFileHandles->writeBytes(1, "    [ -r file ] -- check for file exists and is readable\r\n");
         workingFileHandles->writeBytes(1, "    [ -w file ] -- check for file exists and is writeable\r\n");
         workingFileHandles->writeBytes(1, "    [ -x file ] -- check for file exists and is executable\r\n");
         workingFileHandles->writeBytes(1, "    [ -s file ] -- check for file exists and has non-zero size\r\n");

         workingFileHandles->writeBytes(1, "\r\n");
         workingFileHandles->writeBytes(1, "    [ exprA -a exprB ] -- true if both exprA and exprB are true\r\n");
         workingFileHandles->writeBytes(1, "    [ exprA -o exprB ] -- true if either exprA and exprB are true\r\n");
         workingFileHandles->writeBytes(1, "    [ ! exprA        ] -- true if exprA is false\r\n");

         workingFileHandles->writeBytes(1, "\r\n");
         workingFileHandles->writeBytes(1, "    [ -n string           ] -- true if the string is not empty\r\n");
         workingFileHandles->writeBytes(1, "    [ -z string           ] -- true if the string IS empty\r\n");
         workingFileHandles->writeBytes(1, "    [ stringA = stringB   ] -- true if the strings are equal\r\n");
         workingFileHandles->writeBytes(1, "    [ stringA != stringB  ] -- true if the strings are NOT equal\r\n");

         workingFileHandles->writeBytes(1, "\r\n");
         workingFileHandles->writeBytes(1, "    [ integerA -eq integerB ] -- true if the integers are equal numerically\r\n");
         workingFileHandles->writeBytes(1, "    [ integerA -ne integerB ] -- true if the integers are NOT equal numerically\r\n");
         workingFileHandles->writeBytes(1, "    [ integerA -gt integerB ] -- true if integerA is greater than integerB\r\n");
         workingFileHandles->writeBytes(1, "    [ integerA -ge integerB ] -- true if integerA is greater than or equal to integerB\r\n");
         workingFileHandles->writeBytes(1, "    [ integerA -lt integerB ] -- true if integerA is less than integerB\r\n");
         workingFileHandles->writeBytes(1, "    [ integerA -le integerB ] -- true if integerA is less than or equal to integerB\r\n");

   
         executionContext->variables().setErrorCode(0);
         return ScriptFunctions::SRC_Success;


     }

     virtual string abstract() const
     {
         return "list the builtin commands and their options";
     }


};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// JOBS command
//

struct    BrashBuiltinJobsExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {


         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         cxxtls::ScriptThread::ThreadContext        *executionContext = workingContext.executionContext();
         cxxtls::ScriptThread::ManagedFileHandles   *workingFileHandles = workingContext.fileHandles();


         list<JobDefinition>   jobs;

         BrashScriptCommandHandler::jobList(jobs);

         CXXTLS_FOREACH(JobDefinition const &cur, jobs)
         {
             string        jobId = lexical_cast<string>(cur.jobId_);
             string const &args = cur.args_;

             workingFileHandles->writeBytes(1,jobId);
             workingFileHandles->writeBytes(1, "    ", 4);
             workingFileHandles->writeBytes(1,args);

             if(args.empty() || args[args.size()-1] != '\n')
                  workingFileHandles->writeBytes(1, nl);


         }

   
         executionContext->variables().setErrorCode(0);
         return ScriptFunctions::SRC_Success;


     }

     virtual string abstract() const
     {
         return "List the currently running jobs;  Jobs are started using the & operator.";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// TRUE command
//

struct    BrashBuiltinTrueExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {


         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         cxxtls::ScriptThread::ThreadContext        *executionContext = workingContext.executionContext();
         cxxtls::ScriptThread::ManagedFileHandles   *workingFileHandles = workingContext.fileHandles();

         if(workingContext.executionContext()->loggingExecution())
         {
             workingContext.fileHandles()->writeBytes(2,"true\r\n", 6);
         }


         // workingFileHandles->writeBytes(1, string("[builtin][true]\r\n"));

   
         executionContext->variables().setErrorCode(0);  // yes, zero is true
         return ScriptFunctions::SRC_Success;


     }

     virtual string abstract() const
     {
         return "Set the current exit to true (0)";
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// FALSE command
//

struct    BrashBuiltinFalseExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {


         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         cxxtls::ScriptThread::ThreadContext        *executionContext = workingContext.executionContext();
         cxxtls::ScriptThread::ManagedFileHandles   *workingFileHandles = workingContext.fileHandles();

         if(workingContext.executionContext()->loggingExecution())
         {
             workingContext.fileHandles()->writeBytes(2,"true\r\n", 6);
         }


         executionContext->variables().setErrorCode(1); // yes, 1 is false.
         return ScriptFunctions::SRC_Success;


     }

     virtual string abstract() const
     {
         return "set the current exitCode to false (1)";
     }
};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// TEST command  (aka  [ ... ] )
//

struct    BrashBuiltinTestExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         cxxtls::ScriptThread::ThreadContext      * executionContext   = workingContext.executionContext();
         cxxtls::ScriptThread::ManagedFileHandles * workingFileHandles = workingContext.fileHandles();

         
         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);


         if(arguments.args_.size() < 3)
         {
             FileName tmp(commandLine->sourceFile_);
             tmp.convert_to_absolute_path();

             workingFileHandles->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                               string("[test]-statement: error:  Invalid Test Expression, arguments :"));

             CXXTLS_FOREACH(string const &cur, arguments.args_)  
             {
                 workingFileHandles->writeBytes(2,cur);
                 workingFileHandles->writeBytes(2,space);
             }

             workingFileHandles->writeBytes(2,nl);

             workingContext.variables().setErrorCode(1);  // failed -- may be a problem but who knows

             // should we throw an exception?  probably depends on the "set -e"

         }

         int exitCode = 0; 

         try
         {
             vector<string>::const_iterator f = arguments.args_.begin(), 
                                            l = arguments.args_.end();

             ++f; 

             auto_ptr<BrashTestExpression::TestExpression>  expression(BrashTestExpression::parseTestExpression(f,l));

             // if we get here, there is not parsing error

             bool success = expression->execute(executionContext_in);

             if(!success)
                exitCode = 1;


         }
         catch(std::exception const &e)
         {

             string expression("[test]-statement:  Error expression:");

             vector<string>::const_iterator f = arguments.args_.begin(), 
                                            l = arguments.args_.end();


             if(f != l)
             {
                 while(f != l)
                 {
                     expression += "   ";
                     expression += *f++;
                 }
             }

             FileName tmp(commandLine->sourceFile_);
             tmp.convert_to_absolute_path();

             workingFileHandles->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                               TEsyntaxError + e.what() + nl + expression + nl);


             exitCode = 1;
             workingContext.variables().setErrorCode(exitCode);

             throw ProgramStructureException();   //  just terminates the current command
            
         }


         workingContext.variables().setErrorCode(exitCode);

         // redirectedFileHandles, if any will auto destroy/close

         return ScriptFunctions::SRC_Success;
     }

     virtual string abstract() const
     {
         return "Test for file attributes and/or compare strings:  if [ TestExpr ]; then passed ; fi";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// READ command
//




struct    BrashBuiltinReadExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {



         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         cxxtls::ScriptThread::ThreadContext      * executionContext   = workingContext.executionContext();
         cxxtls::ScriptThread::ManagedFileHandles * workingFileHandles = workingContext.fileHandles();

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         string inputLine;

         if(arguments.argc == 3)
         {
            if(arguments.args_[1] == "-f")
            {
                //  read the entire file into a variable as a giant string

                char   buffer[30000];
                size_t bytes_read;

                string rv;

                bool found = false;

                for(;;)
                {
                    bytes_read = 0;

                    workingFileHandles->readBytes(0, sizeof(buffer), buffer, &bytes_read);

                    if(bytes_read == 0)
                       break;

                    found = true;

                    rv.append(buffer, buffer + bytes_read);

                }

                if(found)
                {
                    executionContext->variables().assign(arguments.args_[2], rv);
                    workingContext.variables().setErrorCode(0); // return useful data!
                }
                else
                {
                   // second call to read will not do anything -- the first call should have
                   // emptied the input stream, so there should be no obvious ill-effects
                   // from simply ignoring

                   workingContext.variables().setErrorCode(1);   // return 1 to indicate end of file

                }

                return ScriptFunctions::SRC_Success;


            }
         }


         if(!workingFileHandles->readLine(0, inputLine) )
         {
             CXXTLS_DEBUG("BrashReadExecutor(" << std::hex << (long )(this) << ") -- END OF INPUT");  

             workingContext.variables().setErrorCode(1);   // return 1 to indicate end of file
             return ScriptFunctions::SRC_Success;
         }

#if 1
#ifdef _MSC_VER
         size_t eof_offset = inputLine.find_first_of(char(0x1a));

         if(eof_offset < inputLine.size())
         {
            inputLine.erase(eof_offset);

            if(inputLine.empty())
            {
                 CXXTLS_DEBUG("BrashReadExecutor(" << std::hex << (long )(this) << ") -- END OF FILE CHARACTER ON INPUT");  

                 workingContext.variables().setErrorCode(1);   // return 1 to indicate end of file
                 return ScriptFunctions::SRC_Success;
            }
         }
#endif
#endif


         CXXTLS_DEBUG("BrashReadExecutor(" << std::hex << (long )(this) << ") -- input line is '" << inputLine << "'" );



         vector<string>::const_iterator f = arguments.args_.begin(),
                                        l = arguments.args_.end();

         if(f != l)
         {
           ++f;  // skip the command name
         }

         string IFS = workingContext.variables().find("IFS");

         if(IFS.empty())
            IFS=" \r\t\n";


         size_t startOffset = inputLine.find_first_not_of(IFS);
         size_t foundOffset = startOffset;

         if(!inputLine.empty())
         {

             size_t endOffset = inputLine.find_last_not_of(IFS);

             if(endOffset < inputLine.size()-1)
             {
                 inputLine.erase(endOffset+1);
             }
         }


         if(f == l)
         {
            // no variable names where given on the command line
   
            executionContext->variables().assign("REPLY", inputLine.substr(startOffset));

         }
         else
         {
             // split up the words on the input line using the interfield separator


             while(f != l)
             {
             
                 if(startOffset >= inputLine.size())
                 {
                    break;
                 }
             
                 string varname = *f++;
             
                 foundOffset = inputLine.find_first_of(IFS, startOffset);

                 if(foundOffset >= inputLine.size() || f == l)
                 {
                     executionContext->variables().assign(varname, inputLine.substr(startOffset));
                     break;
                 }
                 else
                 {
                     executionContext->variables().assign(varname, inputLine.substr(startOffset, foundOffset - startOffset));
                     startOffset = inputLine.find_first_not_of(IFS,foundOffset + 1);


                 }
             
             }

             // if there are any non-populated variables, make them all empty strings
             
             while(f != l)
             {
                string varname = *f++;
                executionContext->variables().assign(varname, emptyString);
             }
         }

         CXXTLS_DEBUG("BrashReadExecutor(" << std::hex << (long )(this) << ") -- found data");  


         workingContext.variables().setErrorCode(0);

         return ScriptFunctions::SRC_Success;
     }


     virtual string abstract() const
     {
         return "Read text from stdin into environment variables:  read [-f] word [word]...;"
                "\r\n                    Split on $IFS;  last word gets rest of line"
                "\r\n                    -f means read the whole file into a single variable as a string.";
     }


};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// SET command
//


struct    BrashBuiltinSetExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         vector<string> argsToSet;

         argsToSet.push_back(workingContext.executionContext()->variables().scriptName());

       bool stillParsingTrueArgs = true;
       bool onlyOptions = false;

       for(size_t arg=1; arg < arguments.args_.size(); ++arg)
       {
           std::string const &s = arguments.args_[arg];

           if(stillParsingTrueArgs)
           {
               if(s.empty())
                  stillParsingTrueArgs = false;
               else
               if(s == "-x")
               {
                  workingContext.executionContext()->setExecutionLogging(true);
                  onlyOptions = true;
               }
               else
               if(s == "+x")
               {
                  workingContext.executionContext()->setExecutionLogging(false);
                  onlyOptions = true;
               }
               else
               if(s == "-f")
               {
                  workingContext.executionContext()->variables().setGlobbing(false); // keeping -f consistent with bash where it means "noglob"
                  onlyOptions = true;
               }
               else
               if(s == "+f")
               {
                  workingContext.executionContext()->variables().setGlobbing(true);
                  onlyOptions = true;
               }
               else
               if(s == "--")
               {
                  stillParsingTrueArgs = false;
                  onlyOptions = false;
                  continue;  // don't insert into output
               }
               else
               if(s == "-h" || s == "--help")
               {
                    workingContext.fileHandles()->writeBytes(2, string(
                                                                         "\r\n"
                                                                         "<brash builtin> set:  List the environment or Set arguments and brash modes\r\n"
                                                                         "\r\n"
                                                                         "Usage:\r\n\r\n"
                                                                         "    set [-h|--help|-f|+f|-x|+x|--] [arg ...]\r\n\r\n"
                                                                         "Description\r\n"
                                                                         "    With no parameters, list all the environment variables and functions\r\n"
                                                                         "\r\n"
                                                                         "    The set command is used to change operating parameters and \r\n"
                                                                         "    to superseed the command line parameters.  For example: \r\n"
                                                                         "\r\n"
                                                                         "       set -f one two three\r\n"
                                                                         "\r\n"
                                                                         "    Will do two things:  it turns off globbing (see below) and also\r\n"
                                                                         "    replace the command line arguments with 'one', 'two', and 'three'.\r\n"
                                                                         "\r\n"
                                                                         "Options\r\n"
                                                                         "    -x    turns on logging\r\n"
                                                                         "    +x    turns it off\r\n"
                                                                         "    -f    turns off globbing\r\n"
                                                                         "    +f    turns globbing on\r\n"
                                                                         "    --    ends the option list and allows the rest of the line to be treated\r\n"
                                                                         "          as it were all normal parameters.  This lets you use -h as parameter\r\n"
                                                                         "          instead of an option.\r\n"
                                                                         "\r\n"
                                                                         "Globbing\r\n"
                                                                         "    Globbing is the act of expanding file name patterns to the names of the\r\n"
                                                                         "    files matching that pattern.  For example:\r\n"
                                                                         "        *.c\r\n"
                                                                         "    Will expand from '*.c' to the list of names of file whose names end in \r\n"
                                                                         "    '.c'.  Generally this is desired, but some algorithms need to have this \r\n"
                                                                         "    behavior turned off temporarily.  Use -f to turn it off, and +f to turn it\r\n"
                                                                         "    back on.\r\n"
                                                                         "\r\n"

                                                                      ));

                    workingContext.variables().setErrorCode(1);

                    return ScriptFunctions::SRC_Success;
                   
               }
               else
               {
                  stillParsingTrueArgs = false;
             
               }
           }


           if (!stillParsingTrueArgs)
           {
           
               argsToSet.push_back(s);
               onlyOptions = false;
           }
       }

       if (!onlyOptions)
         workingContext.executionContext()->variables().setArguments(argsToSet);
         
         workingContext.variables().setErrorCode(0);

         if(!onlyOptions && argsToSet.size() == 1)
         {
             cxxtls::StringExpr::ExecutionContext   &context = workingContext.executionContext()->variables();
             cxxtls::StringExpr::ExecutionVariables &env = context.variables();
      
      
             map<string,string>  tmpCopy;  // of the variables -- this lets us merge the superseeds into the vars for
                                           // printing purposes

             cxxtls::StringExpr::ExecutionVariables::const_iterator f = env.begin(), l = env.end();
      
             while(f != l)
             {
                 cxxtls::StringExpr::ExecutionVariables::VariablesMap::value_type const &cur = *f++;

                 string const &name = cur.first;

                 if(!name.empty())
                 {
                    if(name[0] == 'B' && 
                       isdigit(name[name.size()-1])
                       && name.substr(0,BRASH_RESULT_VAR_PREFIX.size()) == BRASH_RESULT_VAR_PREFIX
                      )
                    {
                        // do not include BRASH_PIPE_RESULTS_ variables in the report

                        continue;
                    }

                    tmpCopy[cur.first] = cur.second;   // copy the variable into the tmpCopy

                 }
      
             }

             const cxxtls::StringExpr::ExecutionVariables &supers = context.superseeds();

             f = supers.begin();
             l = supers.end();

             while (f != l)
             {
                  string const &name = f->first;
                  string const &value= f->second;

                  ++f;

                  map<string, string>::iterator where = tmpCopy.find(name);

                  if (where != tmpCopy.end())
                  {
                       where->second = value;   // replace existing
                  }
                  else
                  {
                      tmpCopy[name] = value;    // add new
                  }
             }

             typedef map<string,string>::value_type value_type;

             CXXTLS_FOREACH(value_type const &var, tmpCopy)
             {
                 string value;

                 value = var.first + equalSign + string(var.second) + nl;

                 workingContext.executionContext()->fileHandles_->writeBytes(1, value);
             }

             // now print the function names

             cxxtls::StringExpr::ExecutionVariables::FunctionsMap::const_iterator firstFunc = env.functions().begin(),
                                                                                  lastFunc  = env.functions().end();
             while(firstFunc != lastFunc)
             {
                 workingContext.executionContext()->fileHandles_->writeBytes(1, 
                                                                             firstFunc->first + funcParmsString + tab 
                                                                             + colon + space + firstFunc->second.file_
                                                                             + colon + lexical_cast<string>(firstFunc->second.line_)
                                                                             + nl
                                                                            );

                 ++firstFunc;
             }


         }


         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Set the command line arguments:  set [-- | -x | +x] [arg1 ...]; -x enables logging, +x turns it off";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// UNSET command
//


struct    BrashBuiltinUnsetExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         int count=0;
         CXXTLS_FOREACH(string const &cur, arguments.args_)
         {
            ++count;

            if(count != 1)
               workingContext.executionContext()->variables().removeVariable(cur);
         }


         workingContext.variables().setErrorCode(0);

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Delete named variables or functions.";
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// SHIFT command
//


static string tooManyOptions("Too many command line options");

struct    BrashBuiltinShiftExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         int shiftCount=1;

         if(arguments.args_.size() > 2)
         {
            FileName tmp(commandLine->sourceFile_);
            tmp.convert_to_absolute_path();

            workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                        tooManyOptions + nl);

            return ScriptFunctions::SRC_Success;
         }
         else
         if(arguments.args_.size() == 2)
         {
             if (arguments.args_[1].empty())
                 shiftCount=-1;
             else
             {
                 // shiftCount = lexical_cast<int>(arguments.args_[1]);

                 string::const_iterator f = arguments.args_[1].begin(),
                                        l = arguments.args_[1].end();
                 
                 std::auto_ptr<StringExpr::Arithmetic::Assignment> assignment(StringExpr::Arithmetic::parseAssignment(f,l));
                 
                 if(!assignment.get())
                 {
                    FileName tmp(commandLine->sourceFile_);
                 
                    tmp.convert_to_absolute_path();
                 
                    workingContext.fileHandles()->writeBytes(1, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) 
                                                                + string(":  shift Error, invalid expression syntax, '") + space + arguments.args_[1] + apostrophe + nl );

                    workingContext.variables().setErrorCode(1);

                    return ScriptFunctions::SRC_Success;

                 }
                 else
                 {
                 
                     shiftCount = assignment->execute(&workingContext.executionContext()->variables());
                 }

             }



             if(shiftCount < 1)
             {
                FileName tmp(commandLine->sourceFile_);
                tmp.convert_to_absolute_path();

                 workingContext.fileHandles()->writeBytes(2,   tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                               string(": shift:  error:  Invalid shift count: ") + arguments.args_[1] + nl);
                 return ScriptFunctions::SRC_Success;
             }
         }

         workingContext.executionContext()->variables().shiftArguments(shiftCount);



         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Remove the first 1 or more arguments from the command line;  shift [count];  The default count is 1";
     }

};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// LET command  (aka (( ... )) )
//


struct    BrashBuiltinLetExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);


         int exitCode=0;

         int count=0;

         CXXTLS_FOREACH(string const &cur, arguments.args_)
         {
            if(count == 0)
            {
               ++count;
               continue;
            }

            string::const_iterator f = cur.begin(), l = cur.end();

            std::auto_ptr<StringExpr::Arithmetic::Assignment> assignment(StringExpr::Arithmetic::parseAssignment(f,l));

            if(!assignment.get())
            {
               exitCode = 1;

               FileName tmp(commandLine->sourceFile_);

               tmp.convert_to_absolute_path();

               workingContext.fileHandles()->writeBytes(1, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) 
                                                           + string(":  Error, invalid expression syntax, '") + space + cur + apostrophe + nl );
               break;
            }

            long value = assignment->execute(&workingContext.executionContext()->variables());

            if(value ==0)
               exitCode=1;
            else
               exitCode=0;
         }

         
         workingContext.variables().setErrorCode(exitCode);

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Evaluate integer expression:  let   'a=b+c' 'x=9' 'f=2*r/13%7';  Don't use $ for variable names";
     }

};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .WC command  
//

#define SMALL_BUF_SIZE (1024 * 1024)

#define STDIN_FILENO 0

#ifdef S_IREG
#undef S_ISREG
#endif

#define S_ISREG(x)  1 // this is a lie, of course, but FileRepresentation won't ever give you directories
                      // or devices of interest.


struct FileRepresentation
   /// keeps a pointer to the file's name, so you can't use this as a general purpose
   /// object.
{
    int fd_;
    BrashBuiltinCommandExecutor::WorkingHandlesAndEnvironmentPointer *workingContext_;
    char const *name_;

    std::string mostRecentError_;

    FileRepresentation(BrashBuiltinCommandExecutor::WorkingHandlesAndEnvironmentPointer &wc)
    : fd_(-1) 
    , workingContext_(&wc)
    , name_("stdin") 
    {
        mostRecentError_.resize(0);
    }

    FileRepresentation(char const *filename)
    {
       mostRecentError_.resize(0);
       workingContext_ = 0;

       fd_ = open(filename, O_RDONLY | O_BINARY, 0);


       if(fd_ < 0)
       {
           mostRecentError_ = cxxtls::lastErrno();
       }

       name_ = filename;

    }

    bool ok() const
    {
        if(fd_ < 0 && workingContext_ == 0)
           return false;          // if both reps are bad, then its bad

        return true;
    }

    int readBytes(unsigned char *buffer, size_t length)
    {
        mostRecentError_.resize(0);

        if(workingContext_ == 0)
        {
           int rv = ::read(fd_, buffer, length);

           if(rv < 0)
           {
               mostRecentError_ = cxxtls::lastErrno();
           }

           return rv;
        }

        size_t bytesRead;

        ScriptThread::ManagedFileHandles::IO_ERROR rc = workingContext_->fileHandles()->readBytes(0, length, (char*)buffer, &bytesRead);

        if(rc)
        {
            mostRecentError_ = cxxtls::lastErrno();
            return 0;  // call all errors "end of file" when reading from stdin
        }

        return bytesRead;

        


    }


    char const *name() const 
    {
       return name_;
    }

    int stat(struct stat *sb)
        // this is meant to be called only by word count, which does not care about times
        // and file types.  open will fail on directories, so we really are only 
        // interested in the size.  For normal file handles, we get everything for free
        // but for windows handles, we are only ever going to be called with h_ == stdin.
        //
        // so don't bother populating the other fields -- though we can if we need to.
    {
       // return 0 if succes, -1 if fail

       mostRecentError_.resize(0);

       if(workingContext_ == 0)
       {
           int rv = fstat(fd_, sb);

           if(rv)
           {
              mostRecentError_ = cxxtls::lastErrno();
           }

           return rv;
       }

       BY_HANDLE_FILE_INFORMATION info;


       if(!GetFileInformationByHandle(workingContext_->fileHandles()->getOSHandle(0), &info))
       {
          mostRecentError_ = cxxtls::lastErrno();
          return -1;
       }

       sb->st_mode = 0;  // basically, a regular file
       sb->st_size = ((unsigned long long)info.nFileSizeHigh ) << 32 + (unsigned long long) info.nFileSizeLow;

        return 0;

    }


    void close()
    {
       if(ok())
       {
          if(workingContext_)
          {
             // we should only get here if we are being invoked on stdin and we don't want to close it
             workingContext_ = 0;
          }
          else
          {
              ::close(fd_);
              fd_ = -1;
          }
       }
    }

    char const *mostRecentError()
    {
         return mostRecentError_.c_str();
    }


};




static int wordCount(  FileRepresentation &rep,
                       long &tlinect,
                       long &twordct,
                       long &tcharct,
                       int &doline,
                       int &doword,
                       int &dochar,
                       int &domulti,
                       BrashBuiltinCommandExecutor::WorkingHandlesAndEnvironmentPointer & workingContext)
{
        FileRepresentation fd = rep;

        struct stat sb;
        typedef unsigned char u_char;

        uintmax_t linect, wordct, charct;
        int len, warned;
        int stat_ret;
        size_t clen;
        short gotsp;
        u_char *p;
        static u_char small_buf[SMALL_BUF_SIZE];
        static u_char *buf = small_buf;
        static off_t buf_size = SMALL_BUF_SIZE;
        wchar_t wch;
        mbstate_t mbs;

        linect = wordct = charct = 0;
        
        if (!fd.ok())
        {
                {
                    char buffer[2000];
                    snprintf(buffer, sizeof(buffer), "%s: error opening file\r\n", fd.name());
                    workingContext.fileHandles()->writeBytes(2, buffer);
        
                }
                return (1);
        }
        

        if (doword || (domulti && MB_CUR_MAX != 1))
                goto word;
        /*
         * Line counting is split out because it's a lot faster to get
         * lines than to get words, since the word count requires some
         * logic.
         */
        if (doline) {
                while ((len = fd.readBytes(buf, buf_size))) {
                        if (len == -1) {
                                char buffer[2000];

                                snprintf(buffer, sizeof(buffer), "%s: read error %s\r\n", fd.name(), fd.mostRecentError());
                                workingContext.fileHandles()->writeBytes(2, buffer);

                                
                                fd.close();
                                return (1);
                        }
                        charct += len;
                        for (p = buf; len--; ++p)
                                if (*p == '\n')
                                        ++linect;
                }
                tlinect += linect;

                char buffer[40];

                snprintf(buffer, sizeof(buffer), " %7lu", linect);

                workingContext.fileHandles()->writeBytes(1, buffer);

                if (dochar) {
                        tcharct += charct;
                        {
                           snprintf(buffer, sizeof(buffer), " %7lu", charct);
                           workingContext.fileHandles()->writeBytes(1, buffer);
                        }
                }
                fd.close();
                return (0);
        }
        /*
         * If all we need is the number of characters and it's a
         * regular file, just stat the puppy.
         */
        if (dochar || domulti) {

                char buffer[2000];

                if (fd.stat(&sb)) {
                        {

                           snprintf(buffer, sizeof(buffer), "%s: fstat error", fd.name());
                           workingContext.fileHandles()->writeBytes(2, buffer);
                        }
                        fd.close();
                        return (1);
                }
                if (S_ISREG(sb.st_mode)) {
                        {
                            snprintf(buffer, sizeof(buffer), " %7lld", (long long)sb.st_size);
                            workingContext.fileHandles()->writeBytes(1, buffer);

                        }
                        tcharct += sb.st_size;
                        fd.close();
                        return (0);
                }
        }

        /* Do it the hard way... */
word:   gotsp = 1;
        warned = 0;
        memset(&mbs, 0, sizeof(mbs));
        while ((len = fd.readBytes(buf, buf_size)) != 0) {
                if (len == -1) {
                        {
                            char buffer[2000];
                            snprintf(buffer, sizeof(buffer), "%s: read error %s\r\n", fd.name(), fd.mostRecentError());
                            workingContext.fileHandles()->writeBytes(2, buffer);
                        }
                        fd.close();
                        return (1);
                }
                p = buf;
                while (len > 0) {
                        if (!domulti || MB_CUR_MAX == 1) {
                                clen = 1;
                                wch = (unsigned char)*p;
                        } else if ((clen = mbrtowc(&wch, (char const *)p, len, &mbs)) ==
                            (size_t)-1) {
                                if (!warned) {
                                        errno = EILSEQ;
                                        {
                                            char buffer[2000];
                                            snprintf(buffer, sizeof(buffer), "%s - multibyte sequence error", fd.name());
                                            workingContext.fileHandles()->writeBytes(2, buffer);
                                        }
                                        warned = 1;
                                }
                                memset(&mbs, 0, sizeof(mbs));
                                clen = 1;
                                wch = (unsigned char)*p;
                        } else if (clen == (size_t)-2)
                                break;
                        else if (clen == 0)
                                clen = 1;
                        charct++;
                        len -= clen;
                        p += clen;
                        if (wch == L'\n')
                                ++linect;
                        if (iswspace(wch))
                                gotsp = 1;
                        else if (gotsp) {
                                gotsp = 0;
                                ++wordct;
                        }
                }
        }
        if (domulti && MB_CUR_MAX > 1)
                if (mbrtowc(NULL, NULL, 0, &mbs) == (size_t)-1 && !warned)
                {
                        char buffer[2000];
                        snprintf(buffer, sizeof(buffer), "%s -- multibyte sequence issue", fd.name());
                        workingContext.fileHandles()->writeBytes(2, buffer);
                }
        if (doline) {
                tlinect += linect;

                char buffer[40];
                snprintf(buffer, sizeof(buffer)," %7lu", linect);
                workingContext.fileHandles()->writeBytes(1, buffer);
        }
        if (doword) {
                twordct += wordct;

                char buffer[40];
                snprintf(buffer, sizeof(buffer)," %7lu", wordct);
                workingContext.fileHandles()->writeBytes(1, buffer);

        }
        if (dochar || domulti) {
                tcharct += charct;

                char buffer[40];
                snprintf(buffer, sizeof(buffer), " %7lu", charct);
                workingContext.fileHandles()->writeBytes(1, buffer);

        }
        fd.close();
        return (0);
}


struct    BrashBuiltinWCExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);


         using namespace CommandLine;

         static OptionDeclaration allowedOptions[]=
         {
             OptionDeclaration("l",    0, 0, "print only line count"),
             OptionDeclaration("c",    0, 0, "print only character count"),
             OptionDeclaration("w",    0, 0, "print only word count"),
             OptionDeclaration("m",    0, 0, "print only character count -- but handle multi-byte characters"),
             OptionDeclaration("h",    0, 0, "get help"),
             OptionDeclaration("help", 0, 1, "get help"),
         };


         try 
         {
             CommandLine::Options options(arguments.args_.begin(),
                                          arguments.args_.end(),
                                          allowedOptions,
                                          sizeof(allowedOptions) / sizeof(allowedOptions[0])
                                         );

             long tlinect=0, twordct=0, tcharct=0, total=0;

             int doline=options.isSet("l");
             int doword=options.isSet("w");
             int dochar=options.isSet("c");
             
             int domulti = 0;

             if(dochar)
                  domulti=false;

             
             if(options.isSet("m"))
             {
                domulti = 1;
                dochar =  0;
             }

             if(options.isSet("h") || options.isSet("help") )
             {
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\n<brash builtin> .wc:  word count\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\nUsage:\r\n\r\n    .wc [options] [file ...]\r\n\r\n"));

                 stringstream  tmp;

                 options.printHelp(tmp);

                 string helpText = tmp.str();

                 workingContext.fileHandles()->writeBytes(2, helpText + nl);

             }
             else
             {

                if (doline + doword + dochar + domulti == 0)
                    doline = doword = dochar = 1;

                 bool err = false;

                 if(options.size() == 1)
                 {
                    int errfnd = wordCount(  FileRepresentation(workingContext),
                                             tlinect, 
                                             twordct, 
                                             tcharct,
                                             doline,
                                             doword,
                                             dochar,
                                             domulti,
                                             workingContext
                                       );

                     if(errfnd)
                     {
                        err = true;
                     }
                     else
                     {
                         workingContext.fileHandles()->writeBytes(1, space + string("stdin" )+ nl); // put the end of line on the output from wordCount which doesnt.

                         ++total;
                     }
                 }
                 else
                 for(size_t i = 1; i < options.size(); ++i)
                 {
                    string const &cur = options.arg(i);

                    int errfnd = wordCount(  FileRepresentation(cur.c_str()),
                                             tlinect, 
                                             twordct, 
                                             tcharct,
                                             doline,
                                             doword,
                                             dochar,
                                             domulti,
                                             workingContext
                                       );

                     if(errfnd)
                     {
                        err = true;
                     }
                     else
                     {
                         workingContext.fileHandles()->writeBytes(1, space + cur + nl); // put the end of line on the output from wordCount which doesnt.

                         ++total;
                     }

                 }
         
                 if (total > 1) {

                         char buffer[40];

                         if (doline)
                         {
                              snprintf(buffer, sizeof(buffer), " %7lu", tlinect);
                              workingContext.fileHandles()->writeBytes(1, buffer);
                         }

                         if (doword)
                         {
                              snprintf(buffer, sizeof(buffer), " %7lu", twordct);
                              workingContext.fileHandles()->writeBytes(1, buffer);
                         }
                         if (dochar || domulti)
                         {
                             snprintf(buffer, sizeof(buffer), " %7lu", tcharct);
                              workingContext.fileHandles()->writeBytes(1, buffer);
                         }
                         
                         workingContext.fileHandles()->writeBytes(1, string(" total\r\n"));

                 }



                workingContext.variables().setErrorCode(err ? 1 : 0);
    
             }


         }
         catch(exception &e)
         {
             workingContext.variables().setErrorCode(1);
    
             workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .wc Error:  ") + e.what() + nl);

    
             return ScriptFunctions::SRC_Success;
             

         }
         
         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Word Count:  wc [-h || --help || -l | -c | -w | -m ] file(s);  Count lines, words, characters in named files";
     }

};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .RM command  
//


struct    BrashBuiltinDotRmExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);


         int exitCode=0;

         bool arg0 = true;

         CXXTLS_FOREACH(string const &cur, arguments.args_)
         {
            if(arg0)
            {
               arg0 = false;
               continue;
            }

            FileName tmp(cur);

            bool error;

            if(tmp.is_dir())
            {
               // whoa!  on windows, if you call rmdir it will take out the directory recursively,
               // we do not want that here!

               FileName::sorted_names_t  matches; // matching file names
                                
#ifdef _MSC_VER
               FileName::find_matching(tmp + "\\*", &matches, 8);   
#else
               FileName::find_matching(tmp + "/*", &matches, 8);   
#endif


               size_t member_count = matches.size();

               if(member_count != 0)
               {
                   exitCode = 1;
    
                   string directoryNotEmpty("ERROR_DIR_NOT_EMPTY");

                   workingContext.fileHandles()->writeBytes(1, string(".rm:  Error:  cannot remove ") + space + cur + dueTo + directoryNotEmpty + nl );
    
                   break;
                  
               }
               else
                   error = tmp.rmdir();  // remove directories.
            }
            else
            {
               error = tmp.remove();  // remove normal files
            }


            if(error)
            {
               string lastErrorMessage = lastErrno();

               exitCode = 1;

               workingContext.fileHandles()->writeBytes(1, string(".rm:  Error:  cannot remove ") + space + cur + dueTo + lastErrorMessage + nl );

               break;
               
            }

         }

         
         workingContext.variables().setErrorCode(exitCode);

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Delete files and directories (non-recurively):  .rm name [...]";
     }

};

///
///
/// BOOL WINAPI MoveFile(
///   _In_ LPCTSTR lpExistingFileName,
///   _In_ LPCTSTR lpNewFileName
/// );
///

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .rename command  
//


struct    BrashBuiltinDotRenameExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         if(arguments.argc != 3)
         {
               workingContext.fileHandles()->writeBytes(2, string("<builtin> .rename error:  expected 2 parameters") + nl);
         
               workingContext.variables().setErrorCode(1);

               return ScriptFunctions::SRC_Success;
         
         }

         int ok = MoveFile(arguments.args_[1].c_str(), 
                           arguments.args_[2].c_str()
                          );

         if(!ok)
         {
               workingContext.fileHandles()->writeBytes(2, string("<builtin> .rename error:  could not rename file/dir") + nl);
               workingContext.fileHandles()->writeBytes(2, tab + string("From: ") + arguments.args_[1] + nl);
               workingContext.fileHandles()->writeBytes(2, tab + string("To  : ") + arguments.args_[2] + nl);
         
               workingContext.variables().setErrorCode(1);

               return ScriptFunctions::SRC_Success;
         
         }




         workingContext.variables().setErrorCode(0);

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Rename files or directories (recursively):  .rename filea fileb;  filea must exist and fileb must NOT exist";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// EXPORT command
//


struct    BrashBuiltinExportExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         bool exportControl = true;
         bool printValue    = false;

         int count=0;
         CXXTLS_FOREACH(string const &cur, arguments.args_)
         {
            ++count;

            if(count == 1)
               continue;  // skip the command name

            if(!cur.empty() && cur[0] == '-')
            {

                if(cur == "-n")
                   exportControl = false;
                else
                if(cur == "-p")
                   printValue = true;
                else
                if(cur == "-np" || cur == "-pn")
                {
                   printValue = true;
                   exportControl = false;
                }

                continue;
            }

            size_t equalOffset = cur.find_first_of('=');

            if(equalOffset >= cur.size())
               workingContext.executionContext()->variables().setExportFlag(cur,exportControl);
            else
            {
                string name, value;

                name = cur.substr(0,equalOffset);
                value= cur.substr(equalOffset+1);

                workingContext.executionContext()->variables().assign(name,value);
                workingContext.executionContext()->variables().setExportFlag(name,exportControl);
            }
         }

         if(printValue || arguments.args_.size() == 1)
         {
            StringExpr::ExecutionVariables &vars = const_cast<StringExpr::ExecutionVariables &>(workingContext.executionContext()->variables().variables());

            typedef StringExpr::ExecutionVariables::VariablesMap::iterator iterator;


            for(iterator f = vars.begin(), l = vars.end(); f != l; ++f)
            {

               string line;

               if(printValue)
               {
                   
                   if(f->second.exportFlag())
                      line += "export ";
                   else
                      line += "       ";
                   
                   line += f->first;
                   
                   line += "='";
                   line += f->second;
                   line += "'";
               }
               else
                   line += f->first;

               line += nl;

               workingContext.fileHandles()->writeBytes(1, line);
                

            }
         }



         workingContext.variables().setErrorCode(0);


         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Mark environment variables for export:  export [-np] name[=value] ...;  -n means no export, -p means print exports";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// HASH command
//


struct    BrashBuiltinHashExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);


         GetOpts optProc(arguments.argc, arguments.argv,"lrpdth");    // options are recognized but not supported

         int c;

         while( (c = optProc.getNextOption() ) != -1 )
         {
             switch(c)
             {
                  case '?':
                    if (optProc.optopt == 'p')
                    {
                        string message = "hash:  Option -";

                        message += char(optProc.optopt);

                        message += " requires an argument";

                        message += nl;

                        workingContext.fileHandles()->writeBytes(2, message);
                     }
                     else if (isprint (optProc.optopt))
                     {
                        string message = "hash:  unknown option -";
                        message += char(optProc.optopt);
                        message += nl;
                     }
                    else
                    {
                        string message = "hash:  unknown option -(binary character)";
                        message += nl;
                    }
                    workingContext.variables().setErrorCode(1);

                    return ScriptFunctions::SRC_Success;

                 case 'r':  
                    workingContext.executionContext()->clearExecHash(); 
                    break;

                 case 'h':  
                    {
                        string message=  "hash:  [-rdhtlp]\r\n       List or erase the executable path hash table\r\n";
                        workingContext.fileHandles()->writeBytes(1, message);
                    }
                    break;

                 case 'p':
                 case 'l':
                 case 'd':
                 case 't':
                   {
                       string message = "hash:  option ";
                       message += char(c);
                       message += " is not yet implemented, sorry.";
                       message += nl;
                       workingContext.fileHandles()->writeBytes(2, message);
                   }
                   workingContext.variables().setErrorCode(1);

                   return ScriptFunctions::SRC_Success;
                   break;
              
                 default:
                   {
                       string message = "hash:  unknown option ";
                       message += char(c);
                       message += nl;
                       workingContext.fileHandles()->writeBytes(2, message);
                   }
                   break;
             }
         }

         std::map<std::string, std::pair<std::string,int> >const  &execHash = workingContext.executionContext()->execHash();

         typedef  std::map<std::string, std::pair<std::string,int> >::value_type value_type;

         CXXTLS_FOREACH(value_type const &cur, execHash)
         {
             FileName tmp(cur.first);

             if(!tmp.is_absolute_path())
                workingContext.fileHandles()->writeBytes(1, cur.first + equalSign + cur.second.first + nl);
         }


         workingContext.variables().setErrorCode(0);


         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Print or clear the executable lookup hash table. hash [-rdthlp]" ;
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================

struct BrashBuiltinDotRegexExecutor_Helper
    // base class from which 2 alternatives are derived.
    //
    // This class exists to define an interface for the regexHelper() function will use.  The
    // .regex command works either on stdin or on named files.  This interface defines the
    // commonalities of the two approaches but hides the implementation details.
    //
    // regexHelper() then makes calls to the read method to obtain input data and makes calls
    // to the write method to print its output text.
{
    virtual bool read(string &inputLine) = 0;
    virtual void write(string const &outputLine) = 0;

    virtual void writeError(string const &message) = 0;


};

struct BrashBuiltinDotRegexExecutor_Helper_GivenIstream
:  public BrashBuiltinDotRegexExecutor_Helper
{
     // this implementation of the interface is for named file

     istream &stream_;

     BrashBuiltinCommandExecutor::WorkingHandlesAndEnvironmentPointer &workingContext_;

     BrashBuiltinDotRegexExecutor_Helper_GivenIstream(istream &s, 
                                                       BrashBuiltinCommandExecutor::WorkingHandlesAndEnvironmentPointer &w
                                                      )
     :  stream_(s)
     ,  workingContext_(w)
     {
     }

     virtual bool read(string &inputLine)
     {
         if(std::getline(stream_, inputLine))
            return true;

         return false;
     }

     virtual void write(string const &outputLine)
     {
         workingContext_.fileHandles()->writeBytes(1, outputLine);
     }

    virtual void writeError(string const &message)
     {
         workingContext_.fileHandles()->writeBytes(2, message + nl);
     }



};

struct BrashBuiltinDotRegexExecutor_Helper_GivenContext
:  public BrashBuiltinDotRegexExecutor_Helper
{

     // this implementation of the interface is for reading from stdin

     BrashBuiltinCommandExecutor::WorkingHandlesAndEnvironmentPointer &workingContext_;

     BrashBuiltinDotRegexExecutor_Helper_GivenContext(
                                                       BrashBuiltinCommandExecutor::WorkingHandlesAndEnvironmentPointer &w
                                                      )
     :  workingContext_(w)
     {
     }

     virtual bool read(string &inputLine)
     {
         bool rv =  workingContext_.fileHandles()->readLine(0, inputLine);

         if(rv && !inputLine.empty())
         {
            size_t length = inputLine.size();

            if(inputLine[length-1] == '\n')
            {
               inputLine.erase(length-1);

#ifdef _MSC_VER

               --length;

               if(length && inputLine[length-1] == '\r')
               {
                   inputLine.erase(length-1);
               }

#endif

            }

         }

         return rv;
     }

     virtual void write(string const &outputLine)
     {
         workingContext_.fileHandles()->writeBytes(1, outputLine);
     }

    virtual void writeError(string const &message)
     {
         workingContext_.fileHandles()->writeBytes(2, message + nl);
     }


};



// handle one file's worth of processing by the .regex command, see below

static 
bool   regexHelper(BrashBuiltinDotRegexExecutor_Helper &ioObject,
                   bool printLines,
                   bool printCount,
                   bool printFilename,
                   bool performSubstitution,
                   string substitution,
                   SimpleRegex *matcher,
                   SimpleRegex *pattern,
                   string filename,
                   bool   invert    // if true, then print lines not matching the matcher pattern
                  )
{
                 ///  Return true if a line was matched or if no -m option specified

                 bool rv = false; // by default assume no matches or substitutions


                 string inputLine;

                 int lineNo = 0;

                 int matchingLines = 0;

                 while(ioObject.read(inputLine))
                 {
                    ++lineNo;

                    if(lineNo % 100 == 0)
                    {
                       if(!consoleStillRunning())
                       {
                          // handle control C

                          ioObject.writeError("^C");

                          break; 
                       }
                    }

                    if(matcher)
                    {
                        // only operate on matching line

                        if(invert)
                        {
                            if( (*matcher)(inputLine))
                                continue; // skip this line
                        }
                        else
                        {
                            if(! (*matcher)(inputLine))
                                continue; // skip this line
                        }
                    }

                    rv = true; // matched a line, and no pattern was specified

                    ++matchingLines;


                    if(printCount)
                    {
                       continue;   // we are only interested in the count of matching line
                    }

                    // at this point we have a matching line

                    string outputLine;

                    if(printFilename && !filename.empty())
                    {
                        outputLine += filename;
                        outputLine += colon;
                    }

                    if(printLines)
                    {
                        outputLine += lexical_cast<string>(lineNo);
                        outputLine += colon;
                    }

                    if(performSubstitution)
                    {
                        string tmp;

                        if( (*pattern)(inputLine) )
                        {
                            pattern->replace(tmp, inputLine, substitution);
                        }
                        else
                            tmp = inputLine;

                        outputLine += tmp;
                    }
                    else
                        outputLine += inputLine;

                    outputLine += nl;

                    ioObject.write(outputLine);
                    
                 }  // read the file and process each line

                 if(printCount)
                 {
                     string outputLine = filename;

                     if(!outputLine.empty())
                         outputLine += colon;

                     outputLine += lexical_cast<string>(matchingLines);
                     outputLine += nl;
                     ioObject.write(outputLine);

                     
                 }

                 return rv;
}


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .DATE command
//

struct    BrashBuiltinDateExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
         // the alias command is a parse time activity, not an execution time.
     
         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         try
         {
             // as always, convert the string expressions that define the command line into a command line arguments
             // object for later parsing.
             
             CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);
             
             using namespace CommandLine;

             // now parse the command line arguments 
             
             static OptionDeclaration allowedOptions[]=
             {
                OptionDeclaration("help", 0, 1, "print help"),

                OptionDeclaration("h", 0, 0, "print help"),
                OptionDeclaration("f", 1, 0, "Set the <format> string.  See below."),
                OptionDeclaration("s", 1, 0, "Specify the <time> to display."),
             };
             
             CommandLine::Options  parsed(arguments.args_.begin(), arguments.args_.end(), allowedOptions, sizeof(allowedOptions)/sizeof(allowedOptions[0]));

             if(parsed.size() > 1)
             {
                FileName tmp(commandLine->sourceFile_);
                tmp.convert_to_absolute_path();

                workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                            string(":  Error:  .date:  no non-flag parameters are allowed\r\n"));
     
                workingContext.variables().setErrorCode(1);   // bad command line arguments is an error

                return ScriptFunctions::SRC_Success;  // this basically means :  return without "returning from the function you are in" or throwing an exceptio
                                                      // or crashing...

                
             }


             if(parsed.isSet("h") || parsed.isSet("help"))
             {


                   string help1 =
                                   "\r\n\r\n<brash builtin> .date:  print the date using a format\r\n"
                                   "\r\nUsage\r\n"
                                   "\r\n    .date [options]\r\n"
                                   ;

                   workingContext.fileHandles()->writeBytes(1, help1);


                   stringstream  tmp;
                   
                   parsed.printHelp(tmp);
                   
                   string helpText = tmp.str();
                   
                   workingContext.fileHandles()->writeBytes(1, nl + helpText + nl);

                   string help2=
                                   "\r\nSpecifying a time\r\n\r\n"
                                   "    Normally, the current time is printed by the .date command, however if the\r\n"
                                   "    -s option is used, a different time may be printed.  To specify a different\r\n"
                                   "    time, use the number of seconds since jan-1-1970 as the time value.  The\r\n"
                                   "    current time in this format may be obtained using the $SECONDS environment variable\r\n"
                                   "\r\nFormat strings\r\n"
                                   "\r\n    Format strings can contain normal text as well as the following substitutions:\r\n"
                                   "        %%      a literal percent sign (%).\r\n"
                                   "        %a      The abbreviated weekday name (e.g., Sun).\r\n"
                                   "        %A      The full weekday name (e.g., Sunday).\r\n"
                                   "        %b      The abbreviated month name (e.g., Jan).\r\n"
                                   "        %B      locale's full month name (e.g., January).\r\n"
                                   "        %c      The date and time (e.g., Thu Mar 3 23:05:25 2005).\r\n"
                                   "        %C      The current century; like %Y, except omit last two digits (e.g., 20).\r\n"
                                   "        %d      day of month (e.g., 01).\r\n"
                                   "        %D      date; same as %m/%d/%y.\r\n"
                                   "        %e      day of month, space padded; same as %_d.\r\n"
                                   "        %F      full date; same as %Y-%m-%d.\r\n"
                                   "        %g      (not supported)\r\n"
                                   "        %G      (not supported)\r\n"
                                   "        %h      same as %b.\r\n"
                                   "        %H      hour (00..23).\r\n"
                                   "        %I      hour (01..12).\r\n"
                                   "        %j      day of year (001..366).\r\n"
                                   "        %k      hour, space padded ( 0..23); same as %_H.\r\n"
                                   "        %l      hour, space padded ( 1..12); same as %_I.\r\n"
                                   "        %m      month (01..12).\r\n"
                                   "        %M      minute (00..59).\r\n"
                                   "        %n      a newline.\r\n"
                                   "        %N      (not supported).\r\n"
                                   "        %p      locale's equivalent of either AM or PM; blank if not known.\r\n"
                                   "        %P      like %p, but lower case.\r\n"
                                   "        %r      locale's 12-hour clock time (e.g., 11:11:04 PM).\r\n"
                                   "        %R      24-hour hour and minute; same as %H:%M.\r\n"
                                   "        %s      seconds since 1970-01-01 00:00:00 UTC.\r\n"
                                   "        %S      second (00..60).\r\n"
                                   "        %t      a tab.\r\n"
                                   "        %T      time; same as %H:%M:%S.\r\n"
                                   "        %u      day of week (1..7); 1 is Monday.\r\n"
                                   "        %U      (not supported)\r\n"
                                   "        %V      (not supported)\r\n"
                                   "        %w      day of week (0..6); 0 is Sunday.\r\n"
                                   "        %W      (not supported)\r\n"
                                   "        %x      locale's date representation (e.g., 12/31/99).\r\n"
                                   "        %X      locale's time representation (e.g., 23:13:48).\r\n"
                                   "        %y      last two digits of year (00..99).\r\n"
                                   "        %Y      year\r\n"
                                   "        %z      (not supported)\r\n"
                   ;

                 workingContext.fileHandles()->writeBytes(1, help2);
                 workingContext.variables().setErrorCode(1);   // not really an error, but returning success makes no sense


             }
             else
             {
                 string formatString;

                 if(parsed.isSet("f"))
                     formatString = parsed.flagValue("f")[0];
                 else
                     formatString="%c"; // print it all

                 DateTools::DateFormatter  formatter(formatString);

                 string formattedOutput;
                 
                 if(!parsed.isSet("s"))
                     formattedOutput = formatter();  // get the current time and formatted using the formatString
                 else
                 {
                     time_t specified = atoi(parsed.flagValue("s")[0].c_str());

                     formattedOutput = formatter(specified);  // format the specified time

                 }

                 workingContext.fileHandles()->writeBytes(1, formattedOutput + nl);

                 workingContext.variables().setErrorCode(0);   // s'all good


             }

         }
         catch(std::exception const &e)
         {
            FileName tmp(commandLine->sourceFile_);
            tmp.convert_to_absolute_path();

            workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                        string(":  Error:  .date: ") + e.what() + nl);

            workingContext.variables().setErrorCode(1);   // bad command line arguments is an error
         }

         return ScriptFunctions::SRC_Success;  // this basically means :  return without "returning from the functio you are in" or throwing an exceptio
                                               // or crashing...

     } // execute

     string abstract() const
     {
        return "Print the date according user specified format.  Use -h for help.";
        ; 
     }


};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .REGEX command
//

struct    BrashBuiltinDotRegexExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
         // the alias command is a parse time activity, not an execution time.
     
         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         try
         {
             // as always, convert the string expressions that define the command line into a command line arguments
             // object for later parsing.
             
             CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);
             
             using namespace CommandLine;

             // now parse the command line arguments 
             
             static OptionDeclaration allowedOptions[]=
             {
                OptionDeclaration("help", 0, 1, "print help"),

                OptionDeclaration("h", 0, 0, "print help"),
                OptionDeclaration("c", 0, 0, "print only count of matches when used with -m"),
                OptionDeclaration("f", 0, 0, "include matching filenames in the output"),
                OptionDeclaration("l", 0, 0, "include line numbers in the output"),
                OptionDeclaration("p", 1, 0, "regex pattern for string replacement (SED style)"),
                OptionDeclaration("r", 1, 0, "substution value for string replacement.  can contain \\0-\\9"),
                OptionDeclaration("m", 1, 0, "output only lines matching regex, unless -v used, which inverts the logic"),
                OptionDeclaration("v", 0, 0, "invert the logic of -m, only print lines not matching the regex"),
                OptionDeclaration("i", 0, 0, "use case insensitive searches"),
                OptionDeclaration("e", 0, 0, "use extended regular expression syntax"),
             };
             
             CommandLine::Options  parsed(arguments.args_.begin(), arguments.args_.end(), allowedOptions, sizeof(allowedOptions)/sizeof(allowedOptions[0]));
    
             //
             //   see if we are just supposed to print the help message and quit
             //
             
             if(parsed.isSet("h") || parsed.isSet("help") )
             {
                 vector<OptionDeclaration> const &info = parsed.info();
             
                 workingContext.fileHandles()->writeBytes(1, string("\r\n<brash builtin>:  .regex:  find/change text using regular expressions\r\n"));
                 workingContext.fileHandles()->writeBytes(1, string("\r\nUsage\r\n    .regex [options] [files]\r\n"));

                 workingContext.fileHandles()->writeBytes(1, string("\r\nDescription\r\n"));
                 
                 workingContext.fileHandles()->writeBytes(1, string("   Find and optionally transform lines from the input files or stdin\r\n"));
                 workingContext.fileHandles()->writeBytes(1, string("   using SED like string substitutions.  See -p and -r, below.\r\n\r\n"));

                 workingContext.fileHandles()->writeBytes(1, string("\r\n   Error level set to 0 if a match was found.\r\n\r\n"));

                 stringstream ss;

                 parsed.printHelp(ss);

                 workingContext.fileHandles()->writeBytes(1, ss.str() + nl);

                 workingContext.fileHandles()->writeBytes(1, string("\r\nExamples\r\n"));

                 workingContext.fileHandles()->writeBytes(1, string("\r\n    Numbering Lines:\r\n"));
                 workingContext.fileHandles()->writeBytes(1, string("\r\n        .regex -fl [file ....]\r\n"));

                 workingContext.fileHandles()->writeBytes(1, string("\r\n    Simulating GREP:\r\n"));
                 workingContext.fileHandles()->writeBytes(1, string("\r\n        .regex -fl -m pattern [file ....]\r\n"));

                 workingContext.fileHandles()->writeBytes(1, string("\r\n    SED-like string replacements:\r\n"));
                 workingContext.fileHandles()->writeBytes(1, string("\r\n        .regex -p pattern -r replacement [file ...]\r\n"));
                 workingContext.fileHandles()->writeBytes(1, string("\r\n        NOTE:  The pattern is just a regex, no leading / is required."));
                 workingContext.fileHandles()->writeBytes(1, string("\r\n               The replace can contain \\0-\\9, and & just like SED\r\n"));

                 workingContext.fileHandles()->writeBytes(1, string("\r\n    Combining:  GREP | SED:\r\n"));
                 workingContext.fileHandles()->writeBytes(1, string("\r\n        .regex -fl -m filterPattern -p pattern -r replacement [file ....]\r\n\r\n"));



                 workingContext.variables().setErrorCode(1);  // treat a help request as an error

                 return ScriptFunctions::SRC_Success;


             }

             //
             //  Now perform the actual task
             //

             workingContext.variables().setErrorCode(1);   // assume fail unless an explicit match was found


             auto_ptr<SimpleRegex> matcher(0);

             bool printCount = parsed.isSet("c");

             bool printLines = parsed.isSet("l");

             bool printFilename = parsed.isSet("f");

             string regexoptions;

             if(parsed.isSet("i"))
             {
                 regexoptions += "i";
             }

             if(parsed.isSet("e"))
             {
                 regexoptions += "e";
             }


             bool extended    = parsed.isSet("e");

             bool substitute = parsed.isSet("p") && parsed.isSet("r");

             bool match = parsed.isSet("m");
             bool invert= parsed.isSet("v");

             string substitution;



             auto_ptr<SimpleRegex> pattern(0);

             if(match)
             {
                 CXXTLS_FOREACH(string const &cur, parsed.flagValue("m"))      // should never me more than 1, but use only the last if there is
                     matcher.reset( new SimpleRegex(cur, regexoptions) );
             }
             else
             if(invert)
             {
                FileName tmp(commandLine->sourceFile_);
                tmp.convert_to_absolute_path();

                 workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                             string(": Error:  .regex: -v used by -m was not.\r\n"));

                 workingContext.variables().setErrorCode(1);   // bad command line arguments is an error

                 return ScriptFunctions::SRC_Success;
             }

             if(substitute)
             {
                 CXXTLS_FOREACH(string const &cur, parsed.flagValue("r"))     // should never me more than 1, but use only the last if there is
                 {
                     string &tmp = substitution;

                     tmp.erase( tmp.begin(), tmp.end());

                     tmp.reserve( cur.size() );

                     string::const_iterator f = cur.begin(),
                                            l = cur.end();

                     while(f != l)
                     {
                         char c = *f++;

                         if(c  != '\\')
                            tmp += c;
                         else
                         switch(c = *f++)
                         {
                             case 'r':  tmp += '\r'; break;
                             case 'n':  tmp += '\n'; break;
                             case 't':  tmp += '\t'; break;
                             case 's':  tmp += ' ' ; break;
                             case 'b':  tmp += 0x08; break; // backspace
                             case 'f':  tmp += 0x0c; break; // page down
                             case 'e':  tmp += 0x1a; break; // esc
                             default:
                                 tmp += '\\';   // we can't convert octal strings here because \0 - \9 are reserved
                                                // for substitutions of the parenthetical matching segments.  See
                                                // sed's documentation on \1, etc.
                                 tmp += c;
                                 break;
                         }
                     }


                 }
             }

             if(parsed.isSet("p") != parsed.isSet("r"))
             {
                FileName tmp(commandLine->sourceFile_);
                tmp.convert_to_absolute_path();

                 workingContext.fileHandles()->writeBytes(2,  tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                              string(": Error:  .regex: -p and -r mismatch.  Either both or neither must be used.  \r\n"));

                 workingContext.variables().setErrorCode(1);   // bad command line arguments is an error

                 return ScriptFunctions::SRC_Success;
             }

             CXXTLS_FOREACH(string const &cur, parsed.flagValue("p"))     // should never me more than 1, but use only the last if there is
                  pattern.reset( new SimpleRegex(cur, regexoptions) );


             if(printCount && (!substitution.empty() || printFilename || printLines) )
             {
                FileName tmp(commandLine->sourceFile_);
                tmp.convert_to_absolute_path();

                 workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                             string(": Error:  .regex: incompatible options, -c cannot be combined with -f, -l, or -s \r\n"));

                 workingContext.variables().setErrorCode(1);   // bad command line arguments is an error

                 return ScriptFunctions::SRC_Success;

                
             }


             if(parsed.size() < 2)
             {
                // no non-flag options were specified -- so read from stdin

               string inputLine;

               int lineNo=0;

                 BrashBuiltinDotRegexExecutor_Helper_GivenContext  helper(workingContext);

                 bool matched = 
                 regexHelper(helper,
                             printLines,
                             printCount,
                             printFilename,
                             substitute,
                             substitution,
                             matcher.get(),
                             pattern.get(),
                             "",
                             invert
                            );

                 workingContext.variables().setErrorCode(matched ? 0 : 1);



             }
             else
             for(size_t i = 1; i < parsed.size(); ++i)
             {
                 FileName cur(parsed.arg(i));

                 // read from each file in turn

                 ifstream file(cur.c_str());

                 if(!file.good())
                 {
                    FileName tmp(commandLine->sourceFile_);
                    tmp.convert_to_absolute_path();

                     workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                                 string(".regex: ERROR:  cannot open file '") + cur + apostrophe + nl);

                     workingContext.variables().setErrorCode(1);   // bad command line arguments is an error

                     // don't return here, other files may not be errors...

                     
                 }

                 BrashBuiltinDotRegexExecutor_Helper_GivenIstream   helper(file, workingContext);

                 bool matched = 
                 regexHelper(helper,
                             printLines,
                             printCount,
                             printFilename,
                             substitute,
                             substitution,
                             matcher.get(),
                             pattern.get(),
                             cur,
                             invert
                            );

                 workingContext.variables().setErrorCode(matched ? 0 : 1);


                 if(!consoleStillRunning())
                    break;


             }


         }
         catch (ScriptThread::PipeWriteFailure const &e)
         {
             // silently ignore like good litle pipe membeers do.  Normally, PipeMember::run() ignores this for us
             // but since we are catching exceptions ere, ourselves, we have to also ignore it.
         }
         catch(std::exception const &e)
         {
            FileName tmp(commandLine->sourceFile_);
            tmp.convert_to_absolute_path();

            workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                        string(":  Error: .regex: ") + e.what() + nl);

            workingContext.variables().setErrorCode(1);   // bad command line arguments is an error
         }

     
     
         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Filter/Transform input using regexes:  .regex [options]  [files];  Use -h for help"
        ; 
     }
   
};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// ALIAS command
//


struct    BrashBuiltinFakeAliasExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
         // the alias command is a parse time activity, not an execution time.

         executionContext_in->fileHandles_->writeBytes(2, string("Brash Bug:  FakeAliasExecutor invoked\n"));

         executionContext_in->variables().setErrorCode(1);

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Create command synonym:  alias name='prog options'; You can add more options at invocation";  
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// (( ... )) command
//


struct    BrashBuiltinFakeArithmeticExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
         // the alias command is a parse time activity, not an execution time.

         executionContext_in->fileHandles_->writeBytes(2, string("Brash Bug:  FakeArithmeticExecutor invoked\n"));

         executionContext_in->variables().setErrorCode(1);

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "(( expr )):  execute an arithmetic expression, as if you had done this:  let \"expr\"";
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// TYPE command
//


struct    BrashBuiltinTypeExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
         // the type command is a parse time activity, not an execution time.

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         bool exportControl = true;
         bool printValue    = false;

         int count=0;
         CXXTLS_FOREACH(string const &aliasName, arguments.args_)
         {

             ++count;

             if(count == 1)
                continue; //skip the command name

             
             if (aliasName.empty())
             {
                FileName tmp(commandLine->sourceFile_);
                tmp.convert_to_absolute_path();

                 workingContext.fileHandles()->writeBytes(2, tmp + colon + lexical_cast<string>(commandLine->sourceLine_) +
                                                             string(": Error:  type: empty argument to the type command, argument is #") + lexical_cast<string>(count) + nl);
                 continue;
             }
             
             workingContext.fileHandles()->writeBytes(1, aliasName);
             workingContext.fileHandles()->writeBytes(1, is);
             
             
             if(workingContext.executionContext()->variables().aliases().find(aliasName) != workingContext.executionContext()->variables().aliases().end())
             {
                 workingContext.fileHandles()->writeBytes(1, alias);
             }
             else
             if(workingContext.executionContext()->variables().variables().functions().find(aliasName) != workingContext.executionContext()->variables().variables().functions().end())
             {
                 workingContext.fileHandles()->writeBytes(1, functionText);
             }
             else
             if(isBrashBuiltin(aliasName))
             {
                 workingContext.fileHandles()->writeBytes(1, builtin);
             }
             else
             {
                 pair<string,int> pathname= workingContext.executionContext()->variables().variables().findInExecHash(aliasName,0);  // looking for native executable
             
                 if(pathname.first.empty())
                 {
                     std::pair<std::string, bool> location = findBrashExecutableInPath(aliasName, workingContext.executionContext()->variables().find("PATH"));

                     if(!location.first.empty())
                     {
                         workingContext.fileHandles()->writeBytes(1, location.first);
                     }
                     else
                         workingContext.fileHandles()->writeBytes(1, unknown);
                 }
                 else
                 {
                     workingContext.fileHandles()->writeBytes(1, pathname.first);
                 }
             }
             
             workingContext.fileHandles()->writeBytes(1, nl);
         }

         workingContext.executionContext()->variables().setErrorCode(0);

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Describe a command's type:  type word ... ; Types are: 'builtin', 'alias', 'function', <external file pathname>";
     }

};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .STAT command
//


struct    BrashBuiltinStatExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         for(size_t i = 1 ; i < arguments.argc;   ++i)
         {
            char const *filename = arguments.argv[i];

            if(i == 1 && filename[0] == '-' && dashH == filename || dashdashH == filename)
            {
               // user is asking for help.

               string message=
                   "\r\n<brash builtin> .stat:  print file attributes for use by the eval command\r\n"
                   "\r\n"
                   "Usage:\r\n"
                   "\r\n    .stat [-h]  [file ...]\r\n"
                   "\r\nDescription\r\n"
                   "    Print the attributes of each named file in a format that the eval\r\n"
                   "    statement can then read into environment variables.  For example\r\n"
                   "\r\n"
                   "      eval $(.stat fred.cxx)\r\n"
                   "\r\n"
                   "    This invocation invokes .stat on file fred.cxx, then sets environment\r\n"
                   "    variables with the file's attributes.\r\n"
                   "\r\n"
                   "Attribute values presented\r\n"
                   "    The following lines are produced for each file:\r\n"
                   "\r\n"
                   "    sfName=...    The file name as entered on the command line\r\n"
                   "    sfFull=...    The full pathname of the file\r\n"
                   "    sfPath=...    The full directory name\r\n"
                   "    sfBase=...    The nodename and extension of the file\r\n"
                   "    sfNode=...    Leaving out the path and the extension\r\n"
                   "    sfExt=...     Just the extension\r\n"
                   "    sfMTime=...   The time in seconds of the last modification\r\n"
                   "    sfCTime=...   The time in seconds of the last attribute change\r\n"
                   "    sfATime=...   Supposedly the time in seconds of the last access.\r\n"
                   "    sfUID=...     The user id number\r\n"
                   "    sfGID=...     The group id number\r\n"
                   "    sfMODE=...    The octal value of the file mode\r\n"
                   "    sfMODS=...    The string value of the file mode\r\n"
                   "\r\n"
                   "Notes: Multiple files can be specified on the command line, but eval won't\r\n"
                   "       know how to separate them out, so in general, only one parameter\r\n"
                   "       makes sense.\r\n"
                   "\r\n"
                   "       See the .date command for formatting seconds into human readable\r\n"
                   "       strings.\r\n"
                   "\r\n"
                   ;

                workingContext.fileHandles()->writeBytes(1, message);

                workingContext.executionContext()->variables().setErrorCode(1);   // bad command line arguments is an error
                
                return ScriptFunctions::SRC_Success;


            }


            struct stat fileinfo;

            if(0 != stat(filename,&fileinfo))
            {
                workingContext.fileHandles()->writeBytes(2, string(": Error:  .stat file, '") + filename + string("', cannot be found\r\n"));

                workingContext.executionContext()->variables().setErrorCode(1);   // bad command line arguments is an error
                
                return ScriptFunctions::SRC_Success;
            }

            string message;

            FileName tmp(filename);

            tmp.convert_to_absolute_path();

#ifdef _MSC_VER
            string tmp2;

            CXXTLS_FOREACH(char c, tmp)
            {
               if(c == '/')
                  c = '\\';

               tmp2 += c;
            }

            tmp = tmp2;
#endif

            string node = tmp.basename();

            size_t lastDot = node.find_last_of('.');

            if(lastDot < node.size())
               node = node.substr(0,lastDot);
            
            message += "sfName='";   message += filename;        message +="'\r\n";
            message += "sfFull='";   message += tmp;             message +="'\r\n";
            message += "sfPath='";   message += tmp.dirname();   message +="'\r\n";
            message += "sfBase='";   message += tmp.basename();  message +="'\r\n";
            message += "sfNode='";   message += node;            message +="'\r\n";
            message += "sfExt='";    message += tmp.extension(); message +="'\r\n";

            message += "sfSize='";    message += lexical_cast<string>(fileinfo.st_size);  message +="'\r\n";

            message += "sfMTime='"; message += lexical_cast<string>(fileinfo.st_mtime);  message +="'\r\n";
            message += "sfCTime='"; message += lexical_cast<string>(fileinfo.st_ctime);  message +="'\r\n";
            message += "sfATime='"; message += lexical_cast<string>(fileinfo.st_atime);  message +="'\r\n";

            message += "sfUID='";   message += lexical_cast<string>(fileinfo.st_uid);  message +="'\r\n";
            message += "sfGID='";   message += lexical_cast<string>(fileinfo.st_gid);  message +="'\r\n";

            message += "sfMODE='";  message += lexical_cast<string>(fileinfo.st_mode,'o'); message += "'\r\n";
            message += "sfMODS='";  message += string(FileMode(fileinfo.st_mode));         message += "'\r\n";

            workingContext.fileHandles()->writeBytes(1, message);

         }

         workingContext.executionContext()->variables().setErrorCode(0);

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Print file attributes in a manner suitable for the eval command to execute:  .stat [file...]";
     }

};



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// systemCommandExecutor


DECLARE_CXXTLS_EXCEPTION(EmptyThreadContextForSystemCommandExecutor);


static std::string systemCommandExecutor(std::string const &command, void *threadContext )
   ///   Execute a command in the specified threadContext.  
   /// 
   ///   @param command          The brash language command to execute
   ///   @param threadContext    The thread context in which to execute the command
{
    // Make the command be part of a pipe that executes the command then stores its piped output in a string
    // then return the string.
    
    cxxtls::ScriptThread::ThreadContext* executionContext_in = reinterpret_cast<cxxtls::ScriptThread::ThreadContext*>(threadContext);

    if(executionContext_in == 0)
       throw EmptyThreadContextForSystemCommandExecutor();


    if(executionContext_in->loggingExecution() )
    {
        string logMessage("+ $( ");

        logMessage += command;

        logMessage += " )\n";

        executionContext_in->fileHandles_->writeBytes(2, logMessage);
    }

    int jobId = BrashScriptCommandHandler::newJobId();

    string resultvar = "BRASH_PIPE_RESULTS_";
           resultvar += lexical_cast<string>(jobId);

    executionContext_in->variables().removeVariable(resultvar);

    string pipe_command =   string(" { ")  + command + string(" ; }  | read -f ") + resultvar;

    // note:  the logic of this function requires that the final phase of a pipe be executed in the current
    //        shell context.  This is the default for brash but is not for bash.  In bash none of the pipe
    //        is executed in the current context.  If a bash compatible mode is ever added, then this code will
    //        have to insure that that mode is not turned on for the above pipe_command execution.

    stringstream stream(pipe_command);

    ScriptLineStream  command_stream(stream);

    string filename; // really just a fragment of the expression

    if(command.size() > 30)
        filename = command.substr(0,30);
    else
        filename = command;

    command_stream.setFilename(filename);
 
    cxxtls::StringExpr::ExecutionContext::ArgsVector args;
 
    args.push_back("$()");

    int count=0;

    CXXTLS_FOREACH(string const &curArg, executionContext_in->variables().arguments())
    {
       if(0 == count++)
          continue; // skip first arg

       args.push_back(curArg);
    }


    auto_ptr<BrashStatementList>  bsl( new BrashStatementList(brashGlobalScriptCommandHandler,
                                                                      ScriptThread::ThreadContext::
                                                                              create("$()",
                                                                                     args,
                                                                                     &executionContext_in->variables().variables(),
                                                                                     executionContext_in->fileHandles_->clone()
                                                                                    ),
                                                                      new cxxtls::ScriptFunctions::StatementList,
                                                                      jobId
                                                             )
                                     );


    cxxtls::ScriptThread::ThreadContext* savedChildContext = bsl->context_;

    if(parseBrash(command_stream, *bsl, *executionContext_in->fileHandles_, executionContext_in, true))  // parse the entire contents of the stream as 1 line
    {




       BrashScriptCommandHandler::registerJob(JobDefinition(jobId, filename ) ) ;   
       
       {
           // for exception saftey, keep these together
           brashGlobalScriptCommandHandler->push(bsl.get());   
           bsl.release();
       }           
       
       BrashThreadExitType exitInfo = BrashScriptCommandHandler::waitForJob(jobId);
       
       executionContext_in->variables().setErrorCode( exitInfo.exitCode_);



       // now return the results of the pipe operation string.

       string pipe_results = executionContext_in->variables().find(resultvar);

       if(!pipe_results.empty())
       {
           size_t first_offset = pipe_results.find_first_not_of("\r\n");
           size_t last_offset = pipe_results.find_last_not_of("\r\n");

           if (first_offset >= pipe_results.size())
           {
               pipe_results.resize(0);
           }
           else
           if (last_offset >= pipe_results.size())
           {
               // how can we get here ?
               pipe_results = pipe_results.substr(first_offset, pipe_results.size() - first_offset);
           }
           else
           {
           
                size_t usable_width = last_offset - first_offset + 1;

                string tmp = pipe_results.substr(first_offset, usable_width);

                pipe_results = tmp;
           }

       }

       return pipe_results;

    }
    else
    {
       // ignore the command, assuming someone else has printed the error messages

       executionContext_in->fileHandles_->writeBytes(2, string("Syntax Error in string expression:  ") + pipe_command + nl);
       executionContext_in->variables().setErrorCode(1);
    }


    return emptyString;

}

// now install the system command executor as part of the string expression mechanism -- replacing the default "popen" variant

static int install_syscom()
{
    StringExpr::ExecutionContext::setSystemCommandExecutor( &systemCommandExecutor );
    return 0;
}

namespace something_no_one_will_ever_use_I_assume
{
int syscom_installed = install_syscom();
}

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// EVAL command
//


struct    BrashBuiltinEvalExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {
         // the type command is a parse time activity, not an execution time.

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0, false, false);

         int count=0;
         CXXTLS_FOREACH(string const &expression, arguments.args_)
         {

             ++count;

             if(count == 1)
                continue; //skip the command name

             stringstream stream(expression);

             ScriptLineStream  command_stream(stream);

             string filename; // really just a fragment of the expression

             if(expression.size() > 30)
                 filename = expression.substr(0,30);
             else
                 filename = expression;

             command_stream.setFilename(filename);


             int jobId = BrashScriptCommandHandler::newJobId();
          
             cxxtls::StringExpr::ExecutionContext::ArgsVector args;
          
             args.push_back("eval");

             int count=0;


             CXXTLS_FOREACH(string const &curArg, workingContext.executionContext()->variables().arguments())
             {
                if(0 == count++)
                   continue; // skip first arg

                args.push_back(curArg);
             }


             auto_ptr<BrashStatementList>  bsl( new BrashStatementList(brashGlobalScriptCommandHandler,
                                                                               ScriptThread::ThreadContext::
                                                                                       create("eval",
                                                                                              args,
                                                                                              &workingContext.variables().variables(),
                                                                                              workingContext.fileHandles()->clone()
                                                                                             ),
                                                                               new cxxtls::ScriptFunctions::StatementList,
                                                                               jobId
                                                                      )
                                              );

             bsl->setSuppressExit(true); // don't let the exit command run inside eval statements

             if(parseBrash(command_stream, *bsl, *workingContext.fileHandles(), workingContext.executionContext()))
             {
                // execute it

                BrashScriptCommandHandler::registerJob(JobDefinition(jobId, filename ) ) ;   
                
                {
                    // for exception saftey, keep these together
                    brashGlobalScriptCommandHandler->push(bsl.get());   
                    bsl.release();
                }           
                
                BrashThreadExitType exitInfo = BrashScriptCommandHandler::waitForJob(jobId);
                
                workingContext.variables().setErrorCode( exitInfo.exitCode_);

             }
             else
             {
                // ignore the command, assuming someone else has printed the error messages

                workingContext.fileHandles()->writeBytes(2, string("source:  Error:  syntax error in 'eval string'\r\n"));
                workingContext.executionContext()->variables().setErrorCode(1);
                return ScriptFunctions::SRC_Success;
             }

             
         }


         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Evaluate strings as if they were commands:  eval 'statement1' ...";
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================

StatementReturnCode
brashReadAndExecuteInContext(std::string const &fileName_in,
                             cxxtls::ScriptThread::ThreadContext                *executionContext_in,
                             cxxtls::StringExpr::ExecutionContext::ArgsVector   &args,
                             bool                                                suppressExit
                            )
{

              int jobId = BrashScriptCommandHandler::newJobId();

              FileName fileName = fileName_in;

              if(!fileName.is_absolute_path())
              {
                 string curdir = executionContext_in->fileHandles_->currentDirectory();

                 fileName = curdir + slash + fileName_in;
              }

              auto_ptr<BrashStatementList>  statements(new BrashStatementList( brashGlobalScriptCommandHandler, // owner
                                                                               ScriptThread::ThreadContext::
                                                                                       create(fileName,
                                                                                              args,
                                                                                              &executionContext_in->variables().variables(),
                                                                                              executionContext_in->fileHandles_->clone()
                                                                                             ),
                                                                               new cxxtls::ScriptFunctions::StatementList,
                                                                               jobId
                                                                             )
                                                      );

              if(suppressExit)
              {
                 statements->setSuppressExit(true);
              }
              
              ScriptLineStream fileStream(fileName.c_str(), "brash>");
          
              if(fileStream.error())
              {
                  executionContext_in->fileHandles_->writeBytes(2, string("source:  Error:  unable to open file '") + fileName + singleQuote + nl);
                  executionContext_in->variables().setErrorCode(1);
                  return ScriptFunctions::SRC_Success;
              }


              bool parseGood = parseBrash(fileStream,
                                          *statements,
                                          *executionContext_in->fileHandles_,
                                          executionContext_in
                                         );

              if(!parseGood)
              {
                  executionContext_in->fileHandles_->writeBytes(2, fileStream.filename() + colon + lexical_cast<string>(fileStream.lineNumber()) +  "  Error, invalid command syntax\r\n");
                  executionContext_in->variables().setErrorCode(1);
                  return ScriptFunctions::SRC_Success;
              }

              
              BrashScriptCommandHandler::registerJob(JobDefinition(jobId, fileName ) ) ;   

              {
                  // for exception saftey, keep these together
                  brashGlobalScriptCommandHandler->push(statements.get());   
                  statements.release();
              }           
              
              BrashThreadExitType exitInfo = BrashScriptCommandHandler::waitForJob(jobId);

              executionContext_in->variables().setErrorCode( exitInfo.exitCode_);
              
              return ScriptFunctions::SRC_Success;
}

//
// SOURCE command (aka . )
//

// statically initalize a set of well known non-script file

static set<string> knownNonScriptExtensions;

static int initializeknownNonScriptExtensions()
{
    knownNonScriptExtensions.insert(".exe");
    knownNonScriptExtensions.insert(".com");
    knownNonScriptExtensions.insert(".bin");
    knownNonScriptExtensions.insert(".obj");
    knownNonScriptExtensions.insert(".lib");
    knownNonScriptExtensions.insert(".dll");
    knownNonScriptExtensions.insert(".so");
    knownNonScriptExtensions.insert(".gif");
    knownNonScriptExtensions.insert(".jpg");
    knownNonScriptExtensions.insert(".jpeg");
    knownNonScriptExtensions.insert(".doc");
    knownNonScriptExtensions.insert(".docx");
    knownNonScriptExtensions.insert(".ppt");
    knownNonScriptExtensions.insert(".pptx");
    knownNonScriptExtensions.insert(".pdf");
    knownNonScriptExtensions.insert(".pdb");
    knownNonScriptExtensions.insert(".ilk");
    knownNonScriptExtensions.insert(".suo");
    knownNonScriptExtensions.insert(".tagpp");   //cxxtls specific
    knownNonScriptExtensions.insert(".mui");
    knownNonScriptExtensions.insert(".cxx");
    knownNonScriptExtensions.insert(".h");
    knownNonScriptExtensions.insert(".xml");
    knownNonScriptExtensions.insert(".htm");
    knownNonScriptExtensions.insert(".html");
    knownNonScriptExtensions.insert(".bat");
    knownNonScriptExtensions.insert(".csh");

    return 0;
}

static int callInitializer = initializeknownNonScriptExtensions();


struct    BrashBuiltinSourceExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         cxxtls::StringExpr::ExecutionContext::ArgsVector    savedArgs, newArgs;

         string myScriptName;


         if(arguments.argc < 2)
         {
            workingContext.fileHandles()->writeBytes(2, string("source:  missing script name\n"));
            executionContext_in->variables().setErrorCode( 1 );
            
         }
         else
         {
             for(size_t i = 1; i < arguments.args_.size(); ++i)
                newArgs.push_back(arguments.args_[i]);

             std::pair<std::string,bool> location;

             FileName tmp(arguments.args_[1]);  // the name of the script to run

             if(tmp.exists())
             {
                string ext = StrTool::lowercase(tmp.extension());

                if(knownNonScriptExtensions.find(ext) != knownNonScriptExtensions.end())
                {
                    executionContext_in->variables().setErrorCode( 1 );
                    workingContext.fileHandles()->writeBytes(2, string("Error:  File, ") + newArgs[0] + string(", is not a script\r\n"));
                    return ScriptFunctions::SRC_Success;

                }

                location.first = tmp;
                location.second= true;  // pretend its a script, anyway

             }
             

             if(location.first.empty())
                 location = findBrashExecutableInPath(newArgs[0],  workingContext.executionContext()->variables().find("PATH"));

             if(location.first.empty() || ! location.second)
             {
                workingContext.fileHandles()->writeBytes(2, string("Error:  File, ") + newArgs[0] + string(", is not a script\r\n"));
                executionContext_in->variables().setErrorCode( 1 );
             }
             else
             {
                 newArgs[0] = location.first;

                 workingContext.executionContext()->variables().swapArguments(savedArgs, myScriptName);
                 
                 brashReadAndExecuteInContext(arguments.args_[1], workingContext.executionContext(), newArgs, false);
                 
                 workingContext.executionContext()->variables().swapArguments(savedArgs, myScriptName);
             }
         }



         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Read and execute a  script in the current shell context.";
     }

};


//==================================================================================================
//==================================================================================================
//==================================================================================================

extern char const *brashVersionLog[];   // null terminate list of character strings.


// VER command

struct    BrashBuiltinVerExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         char const **scan = brashVersionLog;

         while(*scan)
         {
             executionContext_in->fileHandles_->writeBytes(1, string(*scan));
             ++scan;
         }

         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Descripe the current version of brash";
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================

static bool headLine(HANDLE in, string &output)
   // return true if a line can be read from the input file handle, in.  The line will always end in nl even if the
   // end of file is found before a nl is read
{

   output.resize(0);


   for(;;)
   {
       char c;
       
       DWORD bytesRead = 0;
       
       if(ReadFile(in, &c, 1, &bytesRead, 0))
       {
         if(bytesRead == 0)
            c=0x1a;  // call this end of file since that is how it usually happens

         if(c == 0x1a)
         {
             if (output.empty())
                 return false;   // when the first char is ^Z, call that end of file
             else
             {
                 output += '\n';
                 return true;
             }
         }
         
          


         output += c; 

         if(c == '\n')
            return true;
       }
       else
       {
           // end of file

           if(output.empty())
              return false;

           if(output[ output.size() - 1 ] != '\n')
           {
              output += nl;
           }

           return true;

       }

   }
    
}


static void headStdin(size_t lines, HANDLE in, HANDLE out, HANDLE err)
    // read the specified count of lines from the input stream, in, and write them to out.
    //
    // do not read any more bytes than absolutely necessary.
{


    size_t count = 0;

    string tmp;


    while(count < lines && headLine(in, tmp))
    {
       if(tmp.size() == 0)
          continue;  // this should not happen, but if it does, don't access invalid memory

       ++count;

       DWORD bytesWritten;

       if(tmp[ tmp.size() - 1] != '\n')
          tmp += nl;


       WriteFile(out, tmp.data(), tmp.size(), &bytesWritten, 0);

       if (bytesWritten != tmp.size())
       {
           string message = string("<brash builtin> .head error:  cannot copy data from stdin to output stream:  ") + lexical_cast<string>((long)out) + nl; 

           DWORD bytesWritten;

           WriteFile(err, message.data(), message.size(), &bytesWritten, 0);
           return;
          
       }

    }

    // out of data or out of lines to write
     
}

static void headFilename(size_t lines, string const &filename, HANDLE out, HANDLE err)
{
    std::ifstream file(filename.c_str(), std::fstream::in | std::fstream::binary);


    if(!file.good())
    {
        string message = "<brash builtin> .head error:  cannot open file " + filename + nl;

        DWORD bytesWritten;

        WriteFile(err, message.data(), message.size(), &bytesWritten, 0);
        return;
    }

    size_t count = 0;

    string tmp;


    while(count < lines && ::getline(file, tmp))
    {
       if(tmp.size() == 0)
          continue;  // this should not happen, but if it does, don't access invalid memory

       ++count;

       DWORD bytesWritten;

       if(tmp[ tmp.size() - 1] != '\n')
          tmp += nl;


       WriteFile(out, tmp.data(), tmp.size(), &bytesWritten, 0);

       if (bytesWritten != tmp.size())
       {
           string message = "<brash builtin> .head error:  cannot copy data from " + filename + string(" to output stream:  ") + lexical_cast<string>((long)out) + nl; 

           DWORD bytesWritten;

           WriteFile(err, message.data(), message.size(), &bytesWritten, 0);
           return;
          
       }

    }

    // out of data or out of lines to write


}



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .HEAD command
//

struct    BrashBuiltinHeadExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         // redirectedFileHandles, if any will auto destroy/close

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         static CommandLine::OptionDeclaration allowedOptions[]=
         {
             CommandLine::OptionDeclaration("l",    1, 0, "specify line count"),
             CommandLine::OptionDeclaration("h",    0, 0, "get help"),
             CommandLine::OptionDeclaration("help", 0, 1, "get help"),
         };


         try 
         {
             CommandLine::Options options(arguments.args_.begin(),
                                          arguments.args_.end(),
                                          allowedOptions,
                                          sizeof(allowedOptions) / sizeof(allowedOptions[0])
                                         );

             size_t lines = 0;
             
             
             if(options.isSet("l"))
             {
                 lines = lexical_cast<size_t>(options.flagValue("l")[0]); // 1 options is required, so it will exist or we won't get here
                                                                          // due to an exception throw.
             }

             if(options.isSet("h") || options.isSet("help") )
             {
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\n<brash builtin> .head:  print some lines from files to stdout.\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\nUsage:\r\n\r\n    .head [options] [file ...]\r\n\r\n"));

                 stringstream  tmp;

                 options.printHelp(tmp);

                 string helpText = tmp.str();

                 workingContext.fileHandles()->writeBytes(2, helpText + nl);

                 workingContext.fileHandles()->writeBytes(2, string("Note:  If '-' is used as a file name, stdin will be read.\r\n\r\n"));

                 workingContext.variables().setErrorCode(1);

                 return ScriptFunctions::SRC_Success;

             }
             

             if(lines == 0)
             {
                // nothing to do

                workingContext.variables().setErrorCode(0);


                return ScriptFunctions::SRC_Success;

             }


             if(options.size() < 2)
             {
                 headStdin(lines, workingContext.fileHandles()->getOSHandle(0), workingContext.fileHandles()->getOSHandle(1), workingContext.fileHandles()->getOSHandle(2));
             }
             else
             {
                  bool stdinHasBeenRead = false;

                  for(size_t i = 1; i < options.size(); ++i)
                  {
                     string const &filename = options.arg(i);
                     
                     if(filename == "-")
                     {
                         if(!stdinHasBeenRead)
                         {
                             headStdin(lines, workingContext.fileHandles()->getOSHandle(0), workingContext.fileHandles()->getOSHandle(1), workingContext.fileHandles()->getOSHandle(2));
                             stdinHasBeenRead = true;
                         }
                     }
                     else
                     {
                         headFilename(lines, filename, workingContext.fileHandles()->getOSHandle(1), workingContext.fileHandles()->getOSHandle(2));
                     }
                  }
             }




         }
         catch(exception &e)
         {
             workingContext.variables().setErrorCode(1);
    
             workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .head Error:  ") + e.what() + nl);
    
             return ScriptFunctions::SRC_Success;

         }
         


         workingContext.variables().setErrorCode(0);


         return ScriptFunctions::SRC_Success;
     }

     string abstract() const
     {
        return "Copy some lines from files to stdout:  .head [-h | --help | -l lineCount  [file ...] ];\n\t\t    If no files specified or if file is '-', then stdin will be used.";
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .SOLVE command
//
struct    BrashBuiltinSolveExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         // redirectedFileHandles, if any will auto destroy/close

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         static CommandLine::OptionDeclaration allowedOptions[]=
         {
             CommandLine::OptionDeclaration("h",    0, 0, "get help"),
             CommandLine::OptionDeclaration("help", 0, 1, "get help"),
         };


         try 
         {
             CommandLine::Options options(arguments.args_.begin(),
                                          arguments.args_.end(),
                                          allowedOptions,
                                          sizeof(allowedOptions) / sizeof(allowedOptions[0])
                                         )
                                         ;
             
             if(options.isSet("h") || options.isSet("help") )
             {

                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\n<brash builtin> .solve:  Solve a set of simultaneous linear equations.\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("\r\nUsage:\r\n\r\n     .solve N <file \r\n\r\n"));

                 stringstream  tmp;

                 options.printHelp(tmp);

                 string helpText = tmp.str();

                 workingContext.fileHandles()->writeBytes(2, helpText + nl);



                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\nSimultaneous equations are defined by N rows of N + 1 columns of numbers.\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("The final number on each line is the right hand side of each equation.\r\n"));

                 workingContext.fileHandles()->writeBytes(2, string("If the set can be solved, the equation parameters will be printed stdout.\r\n"));

                 workingContext.fileHandles()->writeBytes(2, string("Otherwise errors will be written to stderr.  Error level will be set accordingly.\r\n\r\n"));

                 workingContext.fileHandles()->writeBytes(2, string("A max of 10,000 equations is allowed.  Complexity is NP-Complete.\r\n\r\n"));


                 workingContext.fileHandles()->writeBytes(2, string("Example:  suppose you have 2 equations and 2 unknowns, but suppose we\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("know that both variables are actually 1.  We can create an input file,\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("named 's', that contains these 2 equations:\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    1 1 2\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    2 1 3\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("Then we can run .solve like this:\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    .solve 2 <s\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("And we will get the following results:\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    1 1\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("But note that you must use TRULY UNIQUE equations on each line.  You can't use\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("the same equations on more than one line and you can't use simple MULTIPLES of any\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("existing equation.  This WILL NOT WORK:\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    1 1 2    -- two equations are equal\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    1 1 2\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("Nor will this:\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    1 1 2\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    2 2 4   -- this is a multiple of the first\r\n"));

                 workingContext.variables().setErrorCode(1);
                 
                 return ScriptFunctions::SRC_Success;
             }


             if(options.size() != 2)
             {
                 workingContext.fileHandles()->writeBytes(2, string("\r\n<brash builtin> .solve:  Error expected exactly 1 number on the command line\r\n\r\n"));
                 
                 workingContext.variables().setErrorCode(1);
                 
                 return ScriptFunctions::SRC_Success;
                
             }

             int count = lexical_cast<int>(options.arg(1));  // may throw exception to be handled below

             if(count < 1)
             {
                 workingContext.fileHandles()->writeBytes(2, string("\r\n<brash builtin> .solve:  Error N to be greater than 0\r\n\r\n"));
                 
                 workingContext.variables().setErrorCode(1);
                 
                 return ScriptFunctions::SRC_Success;
             }

             cxxtls::GaussianLpSolver solver(count);

             int row;

             for(row = 0; row < count; ++row)
             {
                size_t col;
          
                double d;

                string line;

                if(!headLine(workingContext.fileHandles()->getOSHandle(0), line))
                {
                        workingContext.fileHandles()->writeBytes(2, string("\r\n<brash builtin> .solve: Error:  Out of Input data.\r\n\r\n"));
                        
                        workingContext.variables().setErrorCode(1);
                        
                        return ScriptFunctions::SRC_Success;
                }

                StrTool::stringlist_t input;

                StrTool::parse_words(line, &input, 10000);

                StrTool::stringlist_t::iterator fcol = input.begin();
          
                for(col = 0; col < count ; ++col, ++fcol)
                {

                    try
                    {
                        d = lexical_cast<double>(*fcol);
                    }
                    catch(...)
                    {
                        string errorMessage("\r\n\r\n<brash builtin> .solve:  error invalid number at row ");

                        errorMessage += lexical_cast<string>(row + 1);
                        errorMessage += ", column " + lexical_cast<string>(col + 1);


                        workingContext.fileHandles()->writeBytes(2, errorMessage + nl);
                        
                        workingContext.variables().setErrorCode(1);
                        
                        return ScriptFunctions::SRC_Success;
                    }
                    
                    solver.setParameter(col, row, d);
                }
          
                try
                {
                    d = lexical_cast<double>(*fcol);
                }
                catch(...)
                {
                        string errorMessage("\r\n<brash builtin> .solve:  error invvalid right hand side number at row ");

                        errorMessage += lexical_cast<string>(row + 1);

                        workingContext.fileHandles()->writeBytes(2, errorMessage + nl);
                        
                        workingContext.variables().setErrorCode(1);
                        
                        return ScriptFunctions::SRC_Success;
                }
          
                solver.setRHS(row, d);
          
             }
          
             if(solver.solve())
             {
                for(row = 0; row < count; ++row)
                {
                   double d = solver.result(row);

                   if(isnan(d)  || isinf(d))
                   {
                        workingContext.fileHandles()->writeBytes(2, string("r\n<brash builtin> .solve: Error:  Unable to solve equation set.\r\n\r\n"));
                        
                        workingContext.variables().setErrorCode(1);
                        
                        return ScriptFunctions::SRC_Success;
                   }
                }
          
                for(row = 0; row < count; ++row)
                {
                     workingContext.fileHandles()->writeBytes(1, lexical_cast<string>(solver.result(row)) + space);
                }

                 workingContext.fileHandles()->writeBytes(1, nl);

          
             }
             else
             {
                workingContext.fileHandles()->writeBytes(2, string("\r\n<brash builtin> .solve: Error:  Unable to solve equation set.\r\n\r\n"));
                
                workingContext.variables().setErrorCode(1);
                
                return ScriptFunctions::SRC_Success;
             }



             workingContext.variables().setErrorCode(0);
             
             
             return ScriptFunctions::SRC_Success;

         }
         catch(exception &e)
         {
             workingContext.variables().setErrorCode(1);
    
             workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .solve Error:  ") + e.what() + nl);
    
             return ScriptFunctions::SRC_Success;

         }


     }

     string abstract() const
     {
        return "Solve simultaneous linear equations:  .solve [-h | --help | lineCount <stdin];\n\t\t    Reads N rows of N+1 numbers and solves for X[1], X[2] .. X[N].";
     }


}; // class BrashBuilintSolveExecutor


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .linReg command -- linear regression
//
struct    BrashBuiltinLinRegExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         // redirectedFileHandles, if any will auto destroy/close

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);

         static CommandLine::OptionDeclaration allowedOptions[]=
         {
             CommandLine::OptionDeclaration("i",    0, 0, "interpolate given X, count, X-Step as parameters"),
             CommandLine::OptionDeclaration("h",    0, 0, "get help"),
             CommandLine::OptionDeclaration("help", 0, 1, "get help"),
         };


         try 
         {
             CommandLine::Options options(arguments.args_.begin(),
                                          arguments.args_.end(),
                                          allowedOptions,
                                          sizeof(allowedOptions) / sizeof(allowedOptions[0])
                                         )
                                         ;
             
             if(options.isSet("h") || options.isSet("help") )
             {

                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\n<brash builtin> .linreg:  Compute slope and intercept given points near a line\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("\r\nUsage:\r\n\r\n     .linreg<file \r\n\r\n"));

                 stringstream  tmp;

                 options.printHelp(tmp);

                 string helpText = tmp.str();

                 workingContext.fileHandles()->writeBytes(2, helpText + nl);


                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\nPoints defining the line consist of 2 or more x y pairs\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("which must each appear on its own line in the stdin file.\r\n"));

                 workingContext.fileHandles()->writeBytes(2, string("\r\nIf interpolation is requested using -i, the slope / intercept will be omitted.\r\n"));
                 
                 workingContext.fileHandles()->writeBytes(2, string("\r\nFor example, if file 's' contains:\r\n"));

                 workingContext.fileHandles()->writeBytes(2, string("    0 1\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    1 2\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    3 6\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    4 8\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    5 10\r\n"));

                 workingContext.fileHandles()->writeBytes(2, string("\r\nExecuting .linreg like this:  '.linreg <s' will yield:\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("\r\n    slope = 1.85714, intercept = 0.52381\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("\r\nTo get interpolation, use '.linreg -i 10 10 0.5 <s' and see:\r\n")); 
    
                 workingContext.fileHandles()->writeBytes(2, string("    10 19.0952\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    10.5 20.0238\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    11 20.9524\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    11.5 21.881\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    12 22.8095\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    12.5 23.7381\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    13 24.6667\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    13.5 25.5952\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    14 26.5238\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("    14.5 27.4524\r\n"));

                 workingContext.variables().setErrorCode(1);

                 
                 return ScriptFunctions::SRC_Success;
             }

             if(   (options.size() != 1 && options.size() != 4)
                || (options.size() == 4 && !options.isSet("i") )
                || (options.isSet("i") && options.size() == 1)
               )
             {
                 workingContext.fileHandles()->writeBytes(2, string("\r\n<brash builtin> .linreg:  Error invalid command line\r\n\r\n"));
                 
                 workingContext.variables().setErrorCode(1);
                 
                 return ScriptFunctions::SRC_Success;
                
             }

             int iterations = -1;
             double startX  = -1e31;
             double deltaX  = -1e31;

             if(options.isSet("i"))
             {
                 sscanf(options.arg(1).c_str(), "%lf", &startX);
                 sscanf(options.arg(2).c_str(), "%d",  &iterations);
                 sscanf(options.arg(3).c_str(), "%lf", &deltaX);

                 int suppliedParms = 0;

                 if(iterations > 0)
                    ++ suppliedParms;

                 if(startX != -1e31)
                    ++ suppliedParms;

                 if(deltaX != -1e31)
                    ++suppliedParms;

                 if(suppliedParms != 0  
                    & suppliedParms != 3
                   )
                 {
                     workingContext.fileHandles()->writeBytes(2, string("\r\n<brash builtin> .linreg:  -i expects START_X, ITERATIONS, DELTA_X\r\n\r\n"));
                     
                     workingContext.variables().setErrorCode(1);
                     
                     return ScriptFunctions::SRC_Success;
                 }

             }




             std::vector< std::pair<double, double> > data;

             int row;

             for(row=0;;++row)
             {

                string line;

                if(!headLine(workingContext.fileHandles()->getOSHandle(0), line))
                {
                    break;
                }

                double x = -1e31;
                double y = -1e31;

                sscanf(line.c_str(), "%lf %lf", &x, &y);

                if(x == -1e31 || y == -1e31)
                {
                        workingContext.fileHandles()->writeBytes(2, string("r\n<brash builtin> .linreg: Error:  bad data on row ") + lexical_cast<string>(row+1)
                                                                    + nl + nl);
                        
                        workingContext.variables().setErrorCode(1);
                        
                        return ScriptFunctions::SRC_Success;
                    
                }

                data.push_back( std::pair<double,double>(x,y) );

             }
          
             try
             {
                std::pair<double,double> equation = cxxtls::linearRegression(data);

                if(iterations <= 0)
                workingContext.fileHandles()->writeBytes(1, string("slope = ")         + lexical_cast<string>(equation.first) + 
                                                            string(", intercept = ")   + lexical_cast<string>(equation.second) +
                                                            nl
                                                        );

                if(iterations > 0)
                {
                    for(int i = 0; i < iterations; ++i)
                    {
                        double x =(startX + i * deltaX);
                        double y = equation.first * x + equation.second;

                        workingContext.fileHandles()->writeBytes(1, lexical_cast<string>(x) + space + lexical_cast<string>(y) +nl);

                    }
                }


             }
             catch(...)
             {
                workingContext.fileHandles()->writeBytes(2, string("\r\n<brash builtin> .linreg: Error:  Unable to solve for line equation.\r\n\r\n"));
                
                workingContext.variables().setErrorCode(1);
                
                return ScriptFunctions::SRC_Success;
             }



             workingContext.variables().setErrorCode(0);
             
             
             return ScriptFunctions::SRC_Success;

         }
         catch(exception &e)
         {
             workingContext.variables().setErrorCode(1);
    
             workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .linreg Error:  ") + e.what() + nl);
    
             return ScriptFunctions::SRC_Success;

         }


     }

     string abstract() const
     {
        return "Linear Regression -- compute the slope/intercept for a line:  .linreg [-h | --help ] <stdin];\n\t\tRead x y pairs each on a line and print slope/intercept";
     }


}; // class BrashBuiltinLinRegExecutor



//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .CUT command
//

static void removeNewline(string &line)
{
   if(line.empty())
      return;

   if(line[ line.size() - 1] == '\n')
   {
       line.erase(line.size()-1);

#ifdef _MSC_VER
       if(line[ line.size() - 1] == '\r')
       {
           line.erase(line.size()-1);
       }
#endif

   }
}



struct    BrashBuiltinCutExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         // redirectedFileHandles, if any will auto destroy/close

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);


         static CommandLine::OptionDeclaration allowedOptions[]=
         {
             CommandLine::OptionDeclaration("c",    2, 0, "specify desired character ranges"),
             CommandLine::OptionDeclaration("f",    2, 0, "specify desired field ranges.  See -d."),
             CommandLine::OptionDeclaration("d",    1, 0, "specify delimiter for -f.  Space is default"),
             CommandLine::OptionDeclaration("w",    0, 0, "use whitespace as the delimiter for -f"),
             CommandLine::OptionDeclaration("h",    0, 0, "get help"),
             CommandLine::OptionDeclaration("help", 0, 1, "get help"),
         };

         try 
         {
             CommandLine::Options options(arguments.args_.begin(),
                                          arguments.args_.end(),
                                          allowedOptions,
                                          sizeof(allowedOptions) / sizeof(allowedOptions[0])
                                         )
                                         ;
             
             if(options.isSet("h") || options.isSet("help") )
             {
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\n<brash builtin> .cut:  Copy character ranges from lines in files.\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\nUsage:\r\n\r\n     .cut [options] [file ...]\r\n\r\n"));

                 stringstream  tmp;

                 options.printHelp(tmp);

                 string helpText = tmp.str();

                 workingContext.fileHandles()->writeBytes(2, helpText + nl);

                 workingContext.fileHandles()->writeBytes(2, string("Note:  If '-' is used as a file name, stdin will be read.\r\n\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("       -c and -f may not be used together.\r\n\r\n"));

                 workingContext.fileHandles()->writeBytes(2, string("A 'range' is a comma separated list of of numbers or number-sub-ranges.\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("Here are some examples\r\n\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("   1,2,3,4,9,13,128\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("   10,4-12,20-30,2-5,9,8,7\r\n\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("Ranges are used with the -c and the -f options.\r\n\r\n"));

                 workingContext.variables().setErrorCode(1);

                 return ScriptFunctions::SRC_Success;

             }
             
             string range;

             char delim = ' ';

             if((options.isSet("c") || options.isSet("d")) && options.isSet("w"))
             {
                     workingContext.variables().setErrorCode(1);
                 
                     workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .cut Error:  option -w is not compatible with -d or -c")  + nl);
                 
                     return ScriptFunctions::SRC_Success;
             }

             if(options.isSet("w") && !options.isSet("f"))
             {
                     workingContext.variables().setErrorCode(1);
                 
                     workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .cut Error:  option -w requires -f")  + nl);
                 
                     return ScriptFunctions::SRC_Success;
             }


             if(options.isSet("d"))
             {
                 string delimString = options.flagValue("d")[0];  // won't get here if at least 1 value is not found

                 if(delimString.empty())
                 {
                     workingContext.variables().setErrorCode(1);
                 
                     workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .cut Error -- invalid -d option")  + nl);
                 
                     return ScriptFunctions::SRC_Success;
                 }
                 else
                     delim = delimString[0];
             }




             if(options.isSet("c") || options.isSet("f"))
             {

                 if(options.isSet("c") && options.isSet("f"))
                 {
                     workingContext.variables().setErrorCode(1);
                 
                     workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .cut Error -- illegal mix of -c and -f options")  + nl);
                 
                     return ScriptFunctions::SRC_Success;
                 }

                 vector<string> const *subRanges = 0;
                 
                 if(options.isSet("c"))
                     subRanges = &options.flagValue("c");
                 else
                     subRanges = &options.flagValue("f");

                 bool first = true;

                 CXXTLS_FOREACH(string cur, *subRanges)
                 {
                    if(first)
                       first = false;
                    else
                       range += ",";

                    range += cur;

                 }

                 bool splitOnDelim = options.isSet("f");

                 // create a local class so we can let it interpreted the various
                 // option combinations to choose the right member function of the CutSet
                 // to call

                 struct caller
                 {
                    CutSet actor_;

                    bool fields_;

                    bool whiteSpace_;

                    char delim_;

                    caller(string const &range,
                           bool useFields,
                           bool useWhitespace,
                           char del
                          )
                    :  actor_(range)
                    ,  fields_(useFields)
                    ,  whiteSpace_(useWhitespace)
                    ,  delim_(del)
                    {
                    }

                    string operator() (string line)
                    {
                       if(!fields_)
                          return actor_.fetchCharacters(line);
                       else
                       if(whiteSpace_)
                          return actor_.fetchWords(line);
                       else
                          return actor_.fetchDelimitedFields(line, delim_);
                    }


                 };

                 caller  helper(range, splitOnDelim, options.isSet("w"), delim);


                 // workingContext.fileHandles()->writeBytes(2, string("\r\n\r\n<brash builtin> .cut:  ranges are:  ") + range + nl);


                 if(options.size() > 1)
                 {
                     // process named files from the command line -- including '-' which means "use standard in"

                     bool stdinHasBeenHandled = false;

                     for(size_t i = 1; i < options.size(); ++i)
                     {
                        string filename = options.arg(i);

                        if(filename != "-")
                        {
                            ifstream file(filename);
                            
                            if(!file.good())
                            {
                                workingContext.variables().setErrorCode(1);
                            
                                workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .cut Error -- could not open file: ") + filename + nl);
                            
                                return ScriptFunctions::SRC_Success;
                               
                            }
                            
                            string line;
                            
                            while(::getline(file, line))
                            {
                                removeNewline(line);

                                string value = helper(line);
                            
                                workingContext.fileHandles()->writeBytes(1, value + nl);
                            
                            }
                        }
                        else
                        if(!stdinHasBeenHandled)
                        {
                            // stdin can only be used once

                            stdinHasBeenHandled = true;

                            string line;
                            
                            while(headLine(workingContext.fileHandles()->getOSHandle(0), line))
                            {
                                removeNewline(line);

                                string value = helper(line);
                            
                                workingContext.fileHandles()->writeBytes(1, value + nl);
                            
                            }

                        }
                     
                     }

                 }
                 else
                 {
                    // read from stdin if no file names are provided

                    string line;
                    
                    while(headLine(workingContext.fileHandles()->getOSHandle(0), line))
                    {
                        removeNewline(line);

                        string value = helper(line);
                    
                        workingContext.fileHandles()->writeBytes(1, value + nl);
                    
                    }
                     

                 }


             }
             else
             {
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\n<brash builtin> .cut:  Error:  No -c options specified.\r\n"));

                 workingContext.variables().setErrorCode(1);

                 return ScriptFunctions::SRC_Success;

             }



         }
         catch(exception &e)
         {
             workingContext.variables().setErrorCode(1);
    
             workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .cut Error:  ") + e.what() + nl);
    
             return ScriptFunctions::SRC_Success;

         }



         workingContext.variables().setErrorCode(0);


         return ScriptFunctions::SRC_Success;

     }

     string abstract() const
     {
        return "Select character ranges from lines in files:  .cut [-h | --help | {-c range}  [file ...] ];\n\t\t    Read the files and extract data in the specified ranges";
     }


};   // BrashBuiltinCutExecutor

//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .EXPAND command
//
//  Expand or Unexpand tabs based on the -u option


static std::string handleTabs(std::string const &line, bool action)
    ///  If action is true, expand tabs rather than compressing them.    
{
    if(action)
        return StrTool::expand_tabs(line);

    return StrTool::pack_tabs(line);
}


struct    BrashBuiltinExpandExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         // redirectedFileHandles, if any will auto destroy/close

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);


         bool expandAction = true; // by default, expand tabs


         static CommandLine::OptionDeclaration allowedOptions[]=
         {
             CommandLine::OptionDeclaration("h",    0, 0, "get help"),
             CommandLine::OptionDeclaration("help", 0, 1, "get help"),
             CommandLine::OptionDeclaration("u",    0, 0, "converts repetitions of 8 leading spaces to tabs"),
         };

         try 
         {
             CommandLine::Options options(arguments.args_.begin(),
                                          arguments.args_.end(),
                                          allowedOptions,
                                          sizeof(allowedOptions) / sizeof(allowedOptions[0])
                                         )
                                         ;
             
             if(options.isSet("h") || options.isSet("help") )
             {
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\n<brash builtin> .expand:  Expand tabs\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\nUsage:\r\n\r\n     .expand [-h|--help|-u] [file ...]\r\n\r\n"));

                 stringstream  tmp;

                 options.printHelp(tmp);

                 string helpText = tmp.str();

                 workingContext.fileHandles()->writeBytes(2, helpText + nl);

                 workingContext.fileHandles()->writeBytes(2, string("Note:  If '-' is used as a file name, stdin will be read.\r\n\r\n"));

                 workingContext.variables().setErrorCode(1);

                 return ScriptFunctions::SRC_Success;

             }

             if(options.isSet("u"))
             {
                 expandAction = false;
             }
             
             {


                 if(options.size() > 1)
                 {
                     // process named files from the command line -- including '-' which means "use standard in"

                     bool stdinHasBeenHandled = false;

                     for(size_t i = 1; i < options.size(); ++i)
                     {
                        string filename = options.arg(i);

                        if(filename != "-")
                        {
                            ifstream file(filename);
                            
                            if(!file.good())
                            {
                                workingContext.variables().setErrorCode(1);
                            
                                workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .cut Error -- could not open file: ") + filename + nl);
                            
                                return ScriptFunctions::SRC_Success;
                               
                            }
                            
                            string line;
                            
                            while(::getline(file, line))
                            {
                                removeNewline(line);

                                string value = handleTabs(line,expandAction);
                            
                                workingContext.fileHandles()->writeBytes(1, value + nl);
                            
                            }
                        }
                        else
                        if(!stdinHasBeenHandled)
                        {
                            // stdin can only be used once

                            stdinHasBeenHandled = true;

                            string line;
                            
                            while(headLine(workingContext.fileHandles()->getOSHandle(0), line))
                            {
                                removeNewline(line);

                                string value = handleTabs(line,expandAction);
                            
                                workingContext.fileHandles()->writeBytes(1, value + nl);
                            
                            }

                        }
                     
                     }

                 }
                 else
                 {
                    // read from stdin if no file names are provided

                    string line;
                    
                    while(headLine(workingContext.fileHandles()->getOSHandle(0), line))
                    {
                        removeNewline(line);

                        string value = handleTabs(line, expandAction);
                    
                        workingContext.fileHandles()->writeBytes(1, value + nl);
                    
                    }
                     

                 }


             }



         }
         catch(exception &e)
         {
             workingContext.variables().setErrorCode(1);
    
             workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .expand Error:  ") + e.what() + nl);
    
             return ScriptFunctions::SRC_Success;

         }



         workingContext.variables().setErrorCode(0);


         return ScriptFunctions::SRC_Success;

     }

     string abstract() const
     {
        return "Expand or unexpand Tabs:  .expand [-h | --help | -u ]    [file ...];\r\n\t\t    -u replaces leading spaces with tabs rather than converting tabs to spaces.";
     }


};   // BrashBuiltinExpandExecutor


//==================================================================================================
//==================================================================================================
//==================================================================================================
//
// .DIFF command
//
//  Compare two files


    static std::string stripWhitespace(std::string const &s)
        // totally bolixes up the return value, though
        //
        // used to compare lines without regard to the whitespace in them
    {
       std::string rv;

       rv.reserve(s.size());

       char const *next = s.data();
       char const *end  = next + s.size();

       while(next != end)
       {
           char c = *next++;

           if(isspace(c))
              continue;

           if(c == '\r' || c == '\n')
              continue;

           rv += c;
       }

       return rv;

    }


    class BrashDiffString
        /// The representation of strings in a file that is actually used by the diff command
        /// as opposed to just the actual string itself.
    {
        std::string primary_;
        std::string secondary_;

        unsigned int hash_;

    public:

        operator std::string const &() const
        {
           return primary_;  // when converting to a string, use the actual string,
                             // not the "space compressed version" stored in secondary_
        }

        BrashDiffString()
        :  hash_(0)
        {
        }

        void init(std::string const &prim, std::string const &sec)
        {
            primary_ = prim;
            secondary_ = sec;

            hash_ = 0;

            unsigned char const * next = (unsigned char const*)sec.data();
            unsigned char const * end  = next + sec.size();

            while(next != end)
            {
               hash_ *= 101;
               hash_ += *next++;
            }
        }

        BrashDiffString(std::string const &prim, std::string const &sec)
        {
            init(prim,sec);
        }

        BrashDiffString &operator=(BrashDiffString const &rhs)
        {
            primary_ = rhs.primary_;
            secondary_ = rhs.secondary_;
            hash_ = rhs.hash_;
            return *this;
        }

        BrashDiffString &operator=(std::string const &rhs)
        {
            init(rhs, rhs);

            return *this;
        }


        BrashDiffString(BrashDiffString const &rhs)
        {
            this->operator=(rhs);
        }

        BrashDiffString(std::string const &rhs)
        {
            this->operator=(rhs);
        }

        bool operator== (BrashDiffString const &rhs) const
        {
            //
            //  if the hash codes aren't the same, then they aren't
            //  equal
            //

            if(hash_ != rhs.hash_)
               return false;      

            // if the hashes are the same, then compare the
            // secondaries, not the primaries!

            return secondary_ == rhs.secondary_;
        }

    };   // class BrashDiffString


    struct BrashDiffAccumulator
        /// a diff accumulator that works like unix diff command does -- kinda
    {
        HANDLE handle_;

        size_t xStartRow_;
        size_t yStartRow_;

        std::vector<std::string> buffer_;

        char type_;

        bool foundDiff_;

        bool allChanges_;

        bool print_;  // yes, do in fact print the differences

        size_t totalDeleted_;
        size_t totalAdded_;


        BrashDiffAccumulator(HANDLE h)
        :  handle_(h)
        ,  xStartRow_(0)
        ,  yStartRow_(0)
        ,  type_(0)
        ,  foundDiff_(false)
        ,  allChanges_(false)
        ,  print_(true)
        ,  totalDeleted_(0)
        ,  totalAdded_(0)
        {
        }

        void flushBuffer(size_t xRow, size_t yRow)
        {
            if(type_ == 0 || !print_ || buffer_.empty())
            {
               type_ = ' ';
               return;
            }

            char const *direction = "    ";


            if(type_ == '>')
            {
               direction = "add ";
            }
            else
            {
               direction = "del ";
            }

            char tmp[100];

            if(type_ == '<')
                snprintf(tmp, sizeof(tmp), "%s@%ld,%ld/%ld\n", direction, xStartRow_, buffer_.size(), yStartRow_);
            else
                snprintf(tmp, sizeof(tmp), "%s@%ld/%ld,%ld\n", direction, xStartRow_+1, yStartRow_, buffer_.size());

            DWORD bytesWritten;

            WriteFile(handle_, tmp, strlen(tmp), &bytesWritten, 0);

            tmp[0] = '\t';
            tmp[1] = 0;


            for(size_t i = 0; i < buffer_.size(); ++i)
            {
               WriteFile(handle_, tmp, 1, &bytesWritten, 0);

               WriteFile(handle_, buffer_[i].data(), buffer_[i].size(), &bytesWritten, 0);

               WriteFile(handle_, nl.data(), nl.size(), &bytesWritten, 0);

            }

            buffer_.clear();

            xStartRow_ = xRow;
            yStartRow_ = yRow;

            type_ = ' ';
        }


        void operator() (char     type,  // minus, plus, or blank.  blank means not changed.  plus means deleted
                         BrashDiffString &item_in,
                         size_t   i,    // first sequence array index
                         size_t   j     // second sequence array index
                        )
        {

            std::string const &item = item_in;

            if(allChanges_)
            {
                if(type != ' ')
                {
                    foundDiff_ = true;
                }

                if(!print_)
                   return; // // should not be getting here, because the option combinations make no sense, but...


                DWORD bytesWritten;

                char tmp[2];

                tmp[0]= type;
                tmp[1]= ' ';

                WriteFile(handle_, &tmp, 2, &bytesWritten, 0);
                
                WriteFile(handle_, item.data(), item.size(), &bytesWritten, 0);
                
                WriteFile(handle_, nl.data(), nl.size(), &bytesWritten, 0);




                return;


            }


            switch(type)
            {
               case ' ':
                   if(type_ != ' ')
                       flushBuffer(i,j);
                   type_ = ' ';
                   break;
                   
               case '>':
               case '<':


                    foundDiff_ = true;

                    if(type == '>')
                        ++totalAdded_;
                    else
                        ++totalDeleted_;


                   if(type != type_)
                   {
                      if(type_ != ' ' && type_ != 0)
                         flushBuffer(i,j);
                      else
                      if (type_ == ' ' || type_ == 0)
                      {
                          xStartRow_ = i;
                          yStartRow_ = j;
                      }
                   }

                   type_ = type;
                   
                   buffer_.push_back(item);

                   break;
            }
        }
         

    };



struct    BrashBuiltinDiffExecutor
:  public BrashBuiltinCommandExecutor
{
    virtual StatementReturnCode 
    execute(  BrashCommandLine*                             commandLine,
              cxxtls::ScriptThread::ThreadContext          *executionContext_in,
              cxxtls::ScriptFunctions::FunctionScope       *functionScope,
              cxxtls::ScriptFunctions::JobController       *jobController,
              bool                                          skipArg0            // for use by the command "builtin"
           ) const 
     {

         // WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object

         WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);

         // redirectedFileHandles, if any will auto destroy/close

         CommandLineArguments arguments(commandLine, workingContext.executionContext(), skipArg0);


         static CommandLine::OptionDeclaration allowedOptions[]=
         {
             CommandLine::OptionDeclaration("A",    0, 0, "list all lines indicating diffs with <>"),
             CommandLine::OptionDeclaration("w",    0, 0, "ignore whitespace"),
             CommandLine::OptionDeclaration("q",    0, 0, "quiet!  set the exit status but produce no output lines"),

             CommandLine::OptionDeclaration("h",    0, 0, "get help"),
             CommandLine::OptionDeclaration("help", 0, 1, "get help"),
         };

         try 
         {
             CommandLine::Options options(arguments.args_.begin(),
                                          arguments.args_.end(),
                                          allowedOptions,
                                          sizeof(allowedOptions) / sizeof(allowedOptions[0])
                                         )
                                         ;
             
             if(options.isSet("h") || options.isSet("help") )
             {
                 workingContext.fileHandles()->writeBytes(2, string("\r\n\r\n<brash builtin> .diff:  Compare 2 Files\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("\r\nUsage:\r\n\r\n     .diff [options] fileA fileB\r\n\r\n"));

                 stringstream  tmp;

                 options.printHelp(tmp);

                 string helpText = tmp.str();

                 workingContext.fileHandles()->writeBytes(2, helpText + nl + nl);

                 workingContext.fileHandles()->writeBytes(2, string("This version of diff produces output that explains how to convert\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("the text in the first file into the second.  This is slightly\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("different than the output of the standard unix diff, but the same lines\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("should show up in both.  This version of diff produces no 'changed'\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("indicators -- only adds and deletes.\r\n\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("Output lines are formatted like this:\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("   del @12,3/23\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("         deleted row 1\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("         deleted row 2\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("         deleted row 3\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("   add @100/23,3\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("         added row 1\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("         added row 2\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("         added row 3\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("   ...\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("Where the add/del lines define the rows from the two files where\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("the change occurred.  In the above, ',3' refers to the number\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("of affected lines.  When deleting it goes with the first file, when\r\n"));
                 workingContext.fileHandles()->writeBytes(2, string("adding, it goes with the second\r\n\r\n"));

                 workingContext.variables().setErrorCode(1);

                 return ScriptFunctions::SRC_Success;

             }

             if(options.size() != 3)  // parm 0 is .diff
             {

                 workingContext.fileHandles()->writeBytes(2, string(".diff Error:  expected exactly 2 filenames.\r\n\r\n"));

                 workingContext.variables().setErrorCode(1);

                 return ScriptFunctions::SRC_Success;
             }
             
             FileName fileA = options.arg(1); // note:  arguments.args_ still has the - options in it but options.arg() does not
             FileName fileB = options.arg(2);

             if(!fileA.exists())
             {
                 workingContext.fileHandles()->writeBytes(2, string(".diff Error:  could not find file '") + fileA + apostrophe + nl + nl);

                 workingContext.variables().setErrorCode(1);

                 return ScriptFunctions::SRC_Success;
             }

             if(!fileB.exists())
             {
                 workingContext.fileHandles()->writeBytes(2, string(".diff Error:  could not find file '") + fileB + apostrophe + nl + nl);

                 workingContext.variables().setErrorCode(1);

                 return ScriptFunctions::SRC_Success;
             }


             std::list<std::string> fileAcontents;
             std::list<std::string> fileBcontents;


             string error;

             error = fileA.read_lines(FileName::Accum<std::list<std::string> > (fileAcontents), 0, true);  // remove the crlf

             if(!error.empty())
             {
                 workingContext.fileHandles()->writeBytes(2, string(".diff Error:  reading file '") + error + apostrophe + nl + nl);

                 workingContext.variables().setErrorCode(1);

                 return ScriptFunctions::SRC_Success;
             }

             error = fileB.read_lines(FileName::Accum<std::list<std::string> >(fileBcontents), 0, true);  // remove the crlf

             if(!error.empty())
             {
                 workingContext.fileHandles()->writeBytes(2, string(".diff Error:  reading file '") + error + apostrophe + nl + nl);

                 workingContext.variables().setErrorCode(1);

                 return ScriptFunctions::SRC_Success;
             }

             std::vector<BrashDiffString> aVec;
             std::vector<BrashDiffString> bVec;

             aVec.reserve(fileAcontents.size());
             bVec.reserve(fileBcontents.size());

             if(options.isSet("w"))
             {
                 CXXTLS_FOREACH(string &cur, fileAcontents)
                     aVec.push_back(BrashDiffString(cur, stripWhitespace(cur)));
                 
                 CXXTLS_FOREACH(string &cur, fileBcontents)
                     bVec.push_back(BrashDiffString(cur,stripWhitespace(cur)));
             }
             else
             {
                 CXXTLS_FOREACH(string &cur, fileAcontents)
                     aVec.push_back(cur);
                 
                 CXXTLS_FOREACH(string &cur, fileBcontents)
                     bVec.push_back(cur);
             }

             {
                 BrashDiffAccumulator accum(workingContext.fileHandles()->getOSHandle(1));
         
                 if(options.isSet("A"))
                    accum.allChanges_ = true;

                 if(options.isSet("q"))
                    accum.print_ = false;


                 LongestCommonSubsequence::
                 accumulateRawDiffs(
                                    &aVec[0], aVec.size(),
                                    &bVec[0], bVec.size(),
                                    accum
                                   );
         
                 if(accum.foundDiff_)
                     accum.flushBuffer(aVec.size(), bVec.size());


                 workingContext.variables().setErrorCode( accum.foundDiff_ );

                 
             }



         }
         catch(exception &e)
         {
             workingContext.variables().setErrorCode(1);
    
             workingContext.fileHandles()->writeBytes(1, string("<brash builtin> .diff Error:  ") + e.what() + nl);
    
             return ScriptFunctions::SRC_Success;

         }


         
         return ScriptFunctions::SRC_Success;



     }

     string abstract() const
     {
        return "Compare two ASCII files:  .diff [-h | --help ] fileA fileB;  Similar to unix diff but not identical";   
     }

};

//==================================================================================================
//==================================================================================================
//==================================================================================================



struct BuiltinsInitializer
{

    BuiltinsInitializer()
    {
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(":",          new BrashBuiltinNOOPExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("cd",         new BrashBuiltinCdExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("pushd",      new BrashBuiltinPushDExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("dirs",       new BrashBuiltinDirsExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("popd",       new BrashBuiltinPopDExecutor) );

        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("echo",       new BrashBuiltinEchoExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("basename",   new BrashBuiltinBasenameExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("dirname",    new BrashBuiltinDirnameExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".attrib",    new BrashBuiltinAttribExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".ls",        new BrashBuiltinLSExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".cat",       new BrashBuiltinCatExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".cut",       new BrashBuiltinCutExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".head",      new BrashBuiltinHeadExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".expand",    new BrashBuiltinExpandExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".diff",      new BrashBuiltinDiffExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("exit",       new BrashBuiltinExitExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("sleep",      new BrashBuiltinSleepExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("env",        new BrashBuiltinEnvExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("builtins",   new BrashListBuiltinsExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("jobs",       new BrashBuiltinJobsExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("true",       new BrashBuiltinTrueExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("false",      new BrashBuiltinFalseExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("pwd",        new BrashBuiltinPwdExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("[",          new BrashBuiltinTestExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("read",       new BrashBuiltinReadExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("set",        new BrashBuiltinSetExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("shift",      new BrashBuiltinShiftExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("let",        new BrashBuiltinLetExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".wc",        new BrashBuiltinWCExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("export",     new BrashBuiltinExportExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("hash",       new BrashBuiltinHashExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("unset",      new BrashBuiltinUnsetExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("type" ,      new BrashBuiltinTypeExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("source",     new BrashBuiltinSourceExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("eval",       new BrashBuiltinEvalExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".",          new BrashBuiltinSourceExecutor) );
    
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("break",      new BrashBuiltinBreakExecutor   ) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("continue",   new BrashBuiltinContinueExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("return",     new BrashBuiltinReturnExecutor  ) );

        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("((",         new BrashBuiltinFakeArithmeticExecutor) );

        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("getopts",    new BrashBuiltinGetOptsExecutor  ) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>("builtin",    new BrashBuiltinBuiltinExecutor  ) );

        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".rename",    new BrashBuiltinDotRenameExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".rm",        new BrashBuiltinDotRmExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".regex",     new BrashBuiltinDotRegexExecutor) );

        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".date",      new BrashBuiltinDateExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".stat",      new BrashBuiltinStatExecutor) );

        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".ver",       new BrashBuiltinVerExecutor) );


        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".solve",     new BrashBuiltinSolveExecutor) );
        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".linreg",    new BrashBuiltinLinRegExecutor) );

        builtinCommands.insert(   pair<string, BrashBuiltinCommandExecutor*>(".stopwatch",    new BrashBuiltinStopwatchExecutor) );


    }


    ~BuiltinsInitializer()
    {
       typedef map<string, BrashBuiltinCommandExecutor*>::value_type  VALUE_TYPE;

       CXXTLS_FOREACH( VALUE_TYPE &cur, builtinCommands)
           delete cur.second;
    }

};

static BuiltinsInitializer  initializer;

bool isBrashBuiltin(std::string const &name)
{
   return builtinCommands.find(name) != builtinCommands.end();
}

static string findDirInPath(string const &dirname, string const &dirpath, string const &startCurdir)
{
   if(dirname.empty())
      return dirname;

   if(dirname == "." || dirname == "..")
      return dirname;



#ifdef _MSC_VER
    char sepChar=';';
    char const *dirChar="\\";
#else
    char sepChar=':';
    char const *dirChar="/";
#endif

    vector<string> members;

    StrTool::splitOnDelimiter(dirpath, members, StrTool::Is_Delim(sepChar));


    CXXTLS_FOREACH(string const &cur, members)
    {
        FileName tmp( cur  );

        if(!tmp.exists())
           continue;
        
        
        if(!tmp.is_absolute_path() )
        {
            tmp = startCurdir + dirChar + cur + dirChar + dirname;
        }
        else
        {
            tmp = cur + dirChar + dirname;
        }




        if(tmp.is_dir())
        {   return tmp;
        }
    }

    return dirname;


}


